/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/



/*   Code for computing bigram frequencies.
     Margaret M. Fleck, 2005  */


/* These functions are based on the notions of an "utterance", which
   is a list of tokens (symbols), and a "section" which is a list of
   utterances.  

   An utterance is meant to be a linguistic unit such as an utterance,
   phrase, sentence, paragraph, or turn.  The practical significance
   of the utterance divisions is that bigrams do not span utterance
   boundaries.  Bigrams at the start/end of utterances start/end
   with the special symbol END.

   A section is meant to be some chunk of corpus data which is a
   convenient size to process in memory, e.g. a news story or a single
   Switchboard conversation.  

   Bigram counting could be implemented using a single Earray
   containing 4-byte (long) count values.  Instead, for didactic
   reasons, this code illustrates how to save some space by storing
   only 1-byte counts for most bigrams, with the few large counts in a
   separate hash table.  In the real world, this optimization might be
   useful if your code is storing other information in addition to the
   ngram count table and especially if you are counting higher-order
   ngrams.  
*/

#include "bigram.h"

/* Print ngram counts to stdout.
   Also prints a summary total to stderr. */
void print_bigram_counts (Earray *counttable, Hashtable *bigcounts)
{
  Earrayptr *iter;
  Item nextngram;
  unsigned long position, count, total = 0, num = 0;
  unsigned char *ptr;

  /* Initialize iterator over global ngram table */
  iter = earraystart(counttable);
  while ((ptr=earraynext(iter, &position))){
    nextngram = code2ngram(position);
    if (length(ngram2list(nextngram)) == 2){        /* The ngram table also contains 1-grams and one 0-gram */
      count = *ptr;
    if (*ptr == 255)
      count = item2long(hashget(bigcounts,nextngram)); 
    total += count;
    num ++;
    printf("%lu %s\n", count , printitem(nextngram)); 
    }
  }
  fflush(stdout);
  fprintf(stderr, "\nTotal number of bigrams counted:  %lu types, %lu tokens\n", num, total);
}

/* Increment count for a single ngram. */
void ngram_count (Item ngram, Earray *counttable, Hashtable *bigcounts)
{
  unsigned char *ptr, oldval;

  ptr = earray_ref(counttable,ngram2code(ngram));
  oldval = *ptr;
  if (oldval == 255)                    /* Value has already overflowed into bigcounts */
    hashincrement(bigcounts,ngram, 1);
  else if (oldval == 254){              /* Value needs to overflow into bigcounts. */
    hashincrement(bigcounts,ngram,255);
    *ptr = 255;
  }
  else                                  /* Value still fits in a single byte. */
    *ptr = oldval + 1;
}

/* Adds bigram counts from input utterance into the count table.
   Each utterance should be a list of symbols (tokenized
   versions of input words).  Works by side-effect. */  
void utt_count_bigrams (Item utterance, Earray *counttable, Hashtable *bigcounts)
{
  Item previous, current;

  /*  Pad end of utterance to handle ngrams at beginning and end. */
  utterance = append(utterance, list(1,END));
  previous = END;
  while (!null(utterance)){
    current = car(utterance);
    if (present(previous) || present(current))
      ngram_count(list2ngram(list(2, previous, current)), counttable, bigcounts); 
    utterance = cdr(utterance);
    previous = current;
  }
}  


/* Count all the bigrams in a "section", i.e. a list of utterances. */
void section_count_bigrams (Item section, Earray *counttable, Hashtable *bigcounts)
{

  while (!null(section)){
    utt_count_bigrams(car(section),counttable, bigcounts);
    section = cdr(section);
  }
}

