/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/



/* Sample customized interpreter

   This interpreter is required to run the demo Schwa scripts.   It also
   illustrates the variety of ways that C functions can be linked into
   a Schwa interpreter. 

   Your own interpreter may link in a large number of C functions or only
   a few.   These functions may give users full access to the options in
   the underlying C functions or, alternatively, present them with only
   a simplified set of choices.   The correct approach depends on your
   application.

*/

#include "tools_wrappers.h"
#include "bigram.h"

/* This interpreter uses a number of functions defined in the Schwa
   tools directory.   The call to define_schwa_tools (in main)
   loads these functions into the interpreter.   

   Our demo interpreter also requires a few custom functions, defined below. */


/* ================================================================ */
/* New utility functions                                            */
/* ================================================================ */

/* A user-level function that slightly modifies the behavior of
   a Schwa library function. */
Item token2lower(Item insymbol)
{
  return string2symbol(string2lower(symbol2string(insymbol)));
}

/* Sometimes a shell command is the easiest way to get the job done. */
Item remove_file(Item filename)
{
  system(gcprintf("rm -f %s", item2string(filename)));
  return MISSING;
}

/* Extract all entries for a word, by grepping a dictionary file
   such as the CMU dictionary.   This illustrates the fact that
   you can link any random piece of C hackery into your interpreter,
   not just well-thought-out library functions. */
Item extract_entry (Item filename, Item word, Item outfile)
{
  char *command;

  command = gcprintf("grep -is \"^%s \" %s >> %s", 
		    symbol2string(word), 
		    item2string(filename), 
		    item2string(outfile));
  system(command);
  return MISSING;
}

/* Count frequencies of all tokens in file, adding them into the
   counts hashtable. */
Item count_words (Item counts, Item filename)
{
  FILE *fp;
  Item words, word;
  Hashtable *rawcounts;

  rawcounts = item2ptr(counts,string2symbol("HASHTABLE"));
  fp = fopen(item2string(filename), "r");
  if (!fp)
    error("Can't open file %s\n", item2string(filename));

  while (!at_end(words=tokenize_next_line(fp, tokfix_english))){
    words = flatten_list(words);   /* remove any list structure within the utterance. */
    while (!null(words)){
      word = car(words);
      words = cdr(words);
      if (item2type(word) != TYPESYMBOL)
	error("count_words:  token %d is not a symbol\n", printitem(word));
      hashincrement(rawcounts,word, 1);
    }
  }
  fclose(fp);
  return MISSING;
}



/* ================================================================ */
/* Wrappers for bigram frequency functions                          */
/* ================================================================ */

/* These functions illustrate how you can store an arbitrary C pointer
   in a special object.  In this case, we're storing a struct that bundles
   together two inputs to the underlying C functions. */

struct smartcounts {
  Earray *smallcounts;     /* An earray containing a 1-byte count for every bigram. */
  Hashtable *bigcounts;    /* A hashtable containing counts longer than one byte. */
};

Item new_counttable_wp (void)
{
  struct smartcounts *mycounts;
  mycounts = safe_malloc(sizeof(struct smartcounts));
  mycounts->smallcounts = new_earray(sizeof(unsigned char));
  mycounts->bigcounts = new_hashtable();
 return ptr2item(mycounts, string2symbol("SMARTCOUNTS"));
}


Item print_bigrams (Item counttable)
{
  struct smartcounts *mycounts;
  mycounts = item2ptr(counttable, string2symbol("SMARTCOUNTS"));

  print_bigram_counts(mycounts->smallcounts, mycounts->bigcounts);
  return MISSING;
}

/* Wrapper for top-level counting function to be linked into the interpreter. */
Item count_bigrams_wp (Item input, Item counttable)
{
  struct smartcounts *mycounts;
  mycounts = item2ptr(counttable, string2symbol("SMARTCOUNTS"));

  section_count_bigrams(input,mycounts->smallcounts, mycounts->bigcounts);
  return MISSING;
}



/* ================================================================ */
/* Top-level code                                                   */
/* ================================================================ */

/* Add extra corpus functions to standard interpeter environment. */
void add_corpus_functions(Schwa_env *environment)
{
  defun("token2lower", token2lower, 1, environment);
  defun("extract_entry", extract_entry, 3, environment);
  defun("count_words", count_words, 2, environment);
  defun("remove_file", remove_file, 1, environment);
  defun("count_bigrams", count_bigrams_wp, 2, environment);
  defun("print_bigrams", print_bigrams, 1, environment);
  defun("new_counttable", new_counttable_wp, 0, environment);
}

int main(int argc, char *argv[])
{

  Schwa_env *environment;

  environment = standard_env();       /* Load basic Schwa functions. */
  define_schwa_tools(environment);    /* Load functions from the Schwa tools directory. */
  add_corpus_functions(environment);  /* Load our own task specific functions */

  invoke_interpreter(argc, argv, environment);

  return 0;
}


