/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* 
   Expandable (chunky sparse) arrays.

   Margaret M. Fleck, 2005 

   Earrays are arrays that expand as needed to accomodate references.

   They are intended for situations in which you expect to store data
   in large sets of consecutive locations, skipping only a moderate
   number of locations in the middle.   If your data has lots of holes
   (very sparse), a hash table may be more appropriate.

   Earrays allocate storage in chunks of modest size.  A hash table
   allows the earray to approximate the set of locations actually in
   use.   In particular, excessive space is not wasted when there are large
   (>> BLOCKSIZE) gaps in the set of locations.

   Once allocated, the underlying storage locations stay put, so you
   don't have to worry about references becoming invalid as the Earray
   expands itself.

   These functions operate like malloc:   it is the caller's responsibility
   to specify the size of objects to be stored and coerce the raw (void *)
   output pointers to an appropriate type.

   Locations whose value has not been explicitly set are guaranteed to
   have the value 0 (which appears as MISSING if coerced to an Item).
   Setting a location's value to all zeros bytes (e.g. MISSING) makes
   it disappear logically, but its space remains allocated.

*/

#include "utils.h"
#include "earray.h"

/* ================================================================ */
/* Constants                                                        */
/* ================================================================ */

/* The size of blocks controls how chunky the space allocation is. 
   Much larger blocksizes interact poorly with the garbage collector
   and waste space if there are gaps in the set of data locations.
   For much smaller blocksizes, the top-level hash table becomes
   a significant overhead.  */
#define BLOCKSIZE 256
#define BLOCKMASK 255
#define BLOCKBITS 8

static int setup = 0;  /* Have the constants been initialized yet? */
static Item S_BLOCK;
static Item S_PTR;

/* Set up values for the symbolic constants.  */
static void symsetup(void) 
{
  S_BLOCK = string2symbol("BLOCK");
  S_PTR = string2symbol("POINTER");
}

/* ================================================================ */
/* Functions                                                        */
/* ================================================================ */

/* The width input is the number of bytes required to store each of your objects
   (e.g. as computed by sizeof).   */
Earray *new_earray(unsigned int width)
{
  Earray *output;

  output = safe_malloc(sizeof(Earray));

  output->table = new_hashtable();
  output->width = width;

  return output;
}

/* Return a pointer to this location in the Earray. */
void *earray_ref(Earray *table, unsigned long position)
{
  unsigned int blocknum, blockoffset, width;
  Item dataptr;
  char *data;

  static int count = 0;

  if (!setup) symsetup();

  /* Compute the block we want and the offset within it */
  blocknum = position >> BLOCKBITS;
  blockoffset = position & BLOCKMASK;
  width = table->width;

  /* Get the data block from the hash table, creating it if necessary. */
  dataptr = hashget(table->table, long2item(blocknum));
  if (!present(dataptr)){
    /*     if (count%1000 == 0)
	   fprintf(stderr, "Allocating %dth new block\n", count); */
    count++;
    data = safe_malloc(width*BLOCKSIZE);
    hashset(table->table, long2item(blocknum), ptr2item(data,S_BLOCK));
  }
  else
    data = item2ptr(dataptr, S_BLOCK);
  
  return data + (width*blockoffset);
}

/* Is the cell at this position and width entirely filled with zeros? 
   Returns 1 if all zeros, 0 otherwise. */
static int all_zero(void *array, unsigned int position, unsigned int width)
{
  unsigned int i;
  unsigned char *ptr;

  ptr = (unsigned char *)array + (width*position);
  for (i = 0; i < width; i++){
    if (*(ptr+i) != 0)
      return 0;
  }
  return 1;
}

/* Is this position set to a non-zero value?   Returns 0 or 1.  
   Can be used to probe the Earray for non-zero values without causing new space to be allocated. */
int earray_isset(Earray *table, unsigned long position)
{
  unsigned int blocknum, blockoffset, width;
  Item dataptr;
  char *data;

  if (!setup) symsetup();

  /* Compute the block we want and the offset within it */
  blocknum = position >> BLOCKBITS;
  blockoffset = position & BLOCKMASK;
  width = table->width;

  /* Get the data block from the hash table */
  dataptr = hashget(table->table, long2item(blocknum));
  if (!present(dataptr))
    return 0;

  data = item2ptr(dataptr, S_BLOCK);
  
  return all_zero(data, width*blockoffset, width);
}

/* ================================================================ */
/* Iterating through the non-zero values in an Earray               */
/* ================================================================ */


/* Return an iterator pointing at the start of the Earray. */
Earrayptr *earraystart(Earray *table)
{
  Earrayptr *output;

  output = (Earrayptr *)safe_malloc(sizeof(Earrayptr));
  
  output->iter = hashstart(table->table);
  output->currentblock = 0;
  output->currentpos = 0;
  output->currentbase = 0;
  output->width = table->width;
  return output;
}


/* Returns the next position in block with non-zero value.  Returns BLOCKSIZE
   when there are no more non-zero values left in block. */
static int earray_next_helper(char *block, unsigned int currentpos, unsigned int width)
{

  while (currentpos < BLOCKSIZE){
    if (!all_zero(block, currentpos, width))
      return currentpos;
    currentpos++;
  }
  return currentpos;
}

/* Returns a pointer to the next position in the Earray with non-zero value.
   The variable position is reset to the coordinate for this location.
   Returns 0 when there are no more values. 
   WARNING:  positions are enumerated in storage order, which often differs
   from standard numerical order. */
void *earraynext(Earrayptr *ptr, unsigned long *position)
{
  char *block;
  Item dataptr, pair;
  unsigned int blocknum, width;
  int newpos;

  if (!setup) symsetup();

  width = ptr->width;

  while (1){
    /* Ensure that we have a block of data to work with. */
    if (ptr->currentblock == 0 || ptr->currentpos >= BLOCKSIZE){
      ptr->currentpos = 0;
      pair = hashnext(ptr->iter);
      if (at_end(pair))
	return 0;
      ptr->currentbase = item2long(car(pair));
      dataptr = cadr(pair);
      ptr->currentblock = item2ptr(dataptr, S_BLOCK);
    }

    /* Extract the block details from our datastructure. */
    blocknum = ptr->currentbase;
    block = ptr->currentblock;  

    newpos = earray_next_helper(block, ptr->currentpos, width);
    ptr->currentpos = newpos+1;
    if (newpos < BLOCKSIZE){
      *position = newpos + (blocknum << BLOCKBITS);
      ptr->currentpos = newpos+1;
      return block+(newpos*width);
    }
  }
}


