/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/



/* Read-eval-print loop.
   Margaret M. Fleck, 2005 */

/* Notice that the values returned from the last three forms
   evaluated are bound to the magic variables @, @@, and @@@.  
   (Using *,**,*** would clash with multiplication.) */

/* Top-level functions are
      invoke_interpreter
      defun, defun_vararg, defun_sf
      setsymbol, evalsymbol
      eval, evalsequence, evalargs
*/     

#include "schwa.h"

/* ================================================================ */
/* Forward declarations                                             */
/* ================================================================ */

/* These functions are private to the interpreter. */

static Item parse_shell_args(int argc, char *argv[], Schwa_env *environment);
static void read_eval_print(Schwa_env *environment, FILE *instream, 
		     FILE *outstream, int errorrestart);
static void script_kludge(FILE *fp);

/* ================================================================ */
/* Constants                                                        */
/* ================================================================ */

/* The maximum function arity supported by this system.
   Functions with arity = MAXARITY+1 take a variable number of arguments. 
   Functions with arity = MAXARITY+2 are special forms, which will
      evaluate their own arguments.
*/
#define MAXARITY 8

/* ================================================================ */
/* Globals                                                          */
/* ================================================================ */

/* openfiles is a list of all file handles currently open for code loading
   operations, so we can close them in case of an error break.
   Make sure this initializer is the same as that for NIL (above). */

static Item openloads = {TAGCOMPLEX};

/* ================================================================ */
/* Functions useful in creating your own extensions to environment  */
/* ================================================================ */

/* Constant representing false, which in this case is the number 0.
   Also a representative true value, which happens to be the number 1. 
   These are the Item versions of 0 and 1, i.e. what you need when
   creating a value that will be given to eval.   */
const Item FALSE = {TAGNUMBER};
const Item TRUE =  {4 + TAGNUMBER};

/* Add a new function to the environment, bound to the input <name>.
   The input function should be a pointer to a C function which
   takes <arity> inputs of type Item and returns a value of type Item. 
   This may require writing a "wrapper" function if your existing C function
   takes/returns arguments of other types.  See stdenv.c for examples.  */
void defun(char *name, void *function, int arity, Schwa_env *environment)
{
  if (arity < 0 || arity > MAXARITY)
    error("defun: number of args %d for function %s not in range [0,%d]\n", 
	  arity, name, MAXARITY);
  setsymbol(string2symbol(name),function2item(function,arity), environment);
}

/* Similar to defun, but the C function should take exactly one input
   argument.   The interpreter will pass to the function a LIST of
   all input arguments. */
void defun_vararg(char *name, void *function, Schwa_env *environment)
{
  setsymbol(string2symbol(name),function2item(function,MAXARITY+1), environment);
}

/* Similar to defun, but the C function should take exactly two
   input arguments:  a list of all input arguments (unevaluated)
   and the environment.   This is used to define special forms. */
void defun_sf(char *name, void *function, Schwa_env *environment)
{
  setsymbol(string2symbol(name),function2item(function,MAXARITY+2), environment);
}

/* ================================================================ */
/* The read-eval-print loop                                         */
/* ================================================================ */

/* Catch signal and invoke an error break. */
static void signal2error(int signum)
{
  error("Interrupt\n");
}

/* Internal read-eval-print loop helper.
   Commands are read from the (open) file stream 
   and values printed to outstream (if present). 
   Commands are echoed if instream isn't stdin.  */
static void rev_loop(Schwa_env *environment, FILE *instream, FILE *outstream)
{
  Item input, value;
  static int allsetup = 0;
  static Item lastname, llastname, lllastname;
  FILE *output;

  if (!allsetup){
    lastname = string2symbol("@");
    llastname = string2symbol("@@");
    lllastname = string2symbol("@@@");
    allsetup = 1;
  }

  if (outstream)
    fprintf(outstream, "%s", environment->prompt);
  fflush(outstream);
  while (!at_end(input = read_sexp(instream))){
    output = outstream;
    if (!output && environment->force_echo)
      output = stderr;

    if (output && instream != stdin) 
      fprintf(output, "%s\n", printitem(input));
    value = eval(input,environment);
    setsymbol(lllastname, evalsymbol(llastname,environment), environment);
    setsymbol(llastname, evalsymbol(lastname,environment), environment);
    setsymbol(lastname,value, environment);
    if (output)
      fprintf(output, "%s\n%s ",printitem(value), environment->prompt);
    fflush(outstream);
  }
}

/* Close any open files associated with loads that were terminated by
   an error break. */
static void closeloads (void)
{
  Item ptr;
  FILE *fp;

  ptr = openloads;
  while (!null(ptr)){
    fp = item2ptr(car(ptr),string2symbol("FILE_HANDLE"));
    fclose(fp);
    ptr = cdr(ptr);
  }
  openloads = NIL;
}


/* Read-eval-print loop.   
   Commands are read from the (open) file instream 
   and output printed to outstream (if present). 
   Commands are echoed if instream isn't stdin. 
   Set up an environment with standard_env (below), then add
   additional functions using defun.  
   If errorrestart is set, restarts loop automatically if error break.
   Maintains a list (backtrace) of currently-active function calls, which
   is dumped into the variable *backtrace* whenever there is an error break.
*/
static void read_eval_print(Schwa_env *environment, FILE *instream, 
		     FILE *outstream, int errorrestart)
{
  int rv;

  /* Set up jump point for error recovery.*/
  
  errorjump = 1; 
  rv = sigsetjmp(errorenv, 1);
  if (rv == 0){
    /* Code to execute */
    environment->localbindings = NIL;
    environment->dynamicbindings = NIL;
    environment->breakisset = 0;
    environment->backtrace = NIL;
    environment->errorrestart = errorrestart;
    rev_loop(environment, instream, outstream);
  }
  else {
    /* This code is run after an error break. 
       We always clean up, because caller may restart interpreter
       by hand even if automatic restart (errorrestart) isn't enabled. */
    environment->localbindings = NIL;
    environment->dynamicbindings = NIL;
    environment->breakisset = 0;
    setsymbol(string2symbol("*backtrace*"), 
	      environment->backtrace, environment);
    environment->backtrace = NIL;
    closeloads();
    if (environment->errorrestart)
      rev_loop(environment, instream, outstream);
  }
}

/* Run top-level interpreter taking input from a file rather than from
   stdin, e.g. we're executing a shell script.   We can count on
   read_eval_print to catch any errors and, thus, allow us to clean
   up by closing the file at the end. */
static void filetoplevel(char *filename, Schwa_env *environment)
{
  FILE *fp;

  fp = fopen(filename, "r");
  if (!fp)
    error("Can't open file %s\n", filename);
  script_kludge(fp);
  read_eval_print(environment, fp, 0, 0); 
  fclose(fp);
}


/* Top-level function for starting interpreter (e.g. from your main
   function).
   Used to create programs that can be invoked either directly (for
   interactive session) or from a shell script.   
   Input your own customized environment, e.g. created using standard_env 
   plus defun/defun_vararg.   
   Your main function should pass argc and argv directly to this function. */
void invoke_interpreter(int argc, char *argv[], Schwa_env *environment)
{
  struct sigaction sigact;

  /* Set up control-C handling */
  sigact.sa_handler = signal2error;
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags = 0;
  sigaction(SIGINT,&sigact, 0);

  if (argc > 1){   
    /* Interpreter was invoked from a shell script. 
       First argument is the name of the shell script file. */
    parse_shell_args(argc, argv, environment);
    filetoplevel(argv[1], environment);
    /* Optionally, start an interactive shell after script file finishes. */
    if (environment->leavelive)
      read_eval_print(environment, stdin, stdout, 1);
  }
  else            
    /* Interpreter was invoked from the command line. 
       Start an interactive session. */
    read_eval_print(environment, stdin, stdout, 1);
}


/* ================================================================ */
/* Function calling                                                 */
/* ================================================================ */

/* Applies a C function to args.  The input function should be a
   function (not e.g. a symbol bound to it).  Caller should have
   verified that it's really a function and checked that number of
   args is consistent with arity.  The input args should already have
   been passed through eval. */
static Item applycfunc (Item function, Item args)
{
  int arity, i;
  Item nargs[8];
  FUNC0 fp0;
  FUNC1 fp1;
  FUNC2 fp2;
  FUNC3 fp3;
  FUNC4 fp4;
  FUNC5 fp5;
  FUNC6 fp6;
  FUNC7 fp7;
  FUNC8 fp8;

  arity = item2arity(function);
  if (arity > MAXARITY)
    error("Vararg function or special form passed to applycfunc.");

  /*   if (length(args) != arity)
    error("Function %s called with %d args\n", printitem(function),
    length(args)); */

  /* Ok, so our function takes a specific number of inputs, between 0 and MAXARITY. */
  for (i=0; i < arity; i++){
    nargs[i] = car(args);
    args = cdr(args);
  }
  switch (arity){
  case 0:
    fp0 = (FUNC0)item2function(function);
    return (*fp0)();
    break;
  case 1:
    fp1 = (FUNC1)item2function(function);
    return (*fp1)(nargs[0]);
    break;
  case 2:
    fp2 = (FUNC2)item2function(function);
    return (*fp2)(nargs[0], nargs[1]);
    break;
  case 3:
    fp3 = (FUNC3)item2function(function);
    return (*fp3)(nargs[0], nargs[1], nargs[2]);
    break;
  case 4:
    fp4 = (FUNC4)item2function(function);
    return (*fp4)(nargs[0], nargs[1], nargs[2], nargs[3]);
    break;
  case 5:
    fp5 = (FUNC5)item2function(function);
    return (*fp5)(nargs[0], nargs[1], nargs[2], nargs[3], nargs[4]);
    break;
  case 6:
    fp6= (FUNC6)item2function(function);
    return (*fp6)(nargs[0], nargs[1], nargs[2], nargs[3], nargs[4],
		  nargs[5]);
    break;
  case 7:
    fp7= (FUNC7)item2function(function);
    return (*fp7)(nargs[0], nargs[1], nargs[2], nargs[3], nargs[4],
		  nargs[5], nargs[6]);
    break;
  case 8:
    fp8= (FUNC8)item2function(function);
    return (*fp8)(nargs[0], nargs[1], nargs[2], nargs[3], nargs[4],
		  nargs[5], nargs[6], nargs[7]);
    break;
  default:
    error("Applycfunc can't handle this case\n");
  }

  /* The compiler can't see that the following is unreacheable. */
  return MISSING;
}

/* Evaluate closure on the input args.  The input name is used to
   create more informative error messages.  The input args should
   already have been passed through eval. */
static Item applyclosure(Item name, Item closure, Item args, Schwa_env *environment)
{
  Item value, oldbindings, formals, symbol;
  Closure *rawclosure;

  /* Save current set of bindings. */
  oldbindings = environment->localbindings;

  /* Restore definition environment. */
  rawclosure = item2closure(closure);
  environment->localbindings = rawclosure->bindings;

  /* Check arity. */
  formals = rawclosure->formals;
  if (length(formals) != length(args))
    error("Closure %s called with wrong number of args %s\n",
	  printitem(name), printitem(args));

  /* Add bindings for inputs. */
  while (!null(formals)){
    symbol = car(formals);
    value = car(args);
    formals = cdr(formals);
    args = cdr(args);
    environment->localbindings = cons(list(2,symbol,value), environment->localbindings);
  }
  
  /* Evaluate body of closure and restore old set of bindings. */
  value = evalsequence(rawclosure->body, environment);

  environment->localbindings = oldbindings; 
  
  return value;
}


/* Apply a function (C function or closure) to input args.
   Assumes that neither the head nor the args has been evaluated yet. */
Item apply(Item head, Item args, Schwa_env *environment)
{
  FUNC1 fp1;
  FUNCSF fpsf;
  Item function, newargs, name;
  int arity = 0; 
  unsigned int type;

  function = eval(head, environment);
  type = item2type(function);

  /* Most informative way to refer to function in any error messages. */
  if (item2type(head) == TYPESYMBOL)
    name = head;
  else 
    name = function;

  if (type != TYPEFUNCTION && type != TYPECLOSURE){
    if (item2type(head) == TYPESYMBOL && !present(function))
      error("Symbol %s is unbound\n", printitem(head));
    else
      error("Expression %s doesn't evaluate to a function or closure\n",
	    printitem(head));
  }


  if (type == TYPEFUNCTION)
    arity = item2arity(function);

  /* Evaluate the input arguments, but first check this isn't
     a special form (whose args shouldn't be evaluated). */
  if (arity < MAXARITY+2)
    newargs = evalargs(args, environment);
  else 
    newargs = args;

  /* This is a closure */
  if (type == TYPECLOSURE)
    return applyclosure(name, function, newargs, environment);

  if (arity < MAXARITY && length(args) != arity)
    error("Function %s called with %d args when it requires %d: %s \n", 
	  printitem(name),
	  length(args), arity, printitem(cons(name, args)));

  if (arity == MAXARITY+2){
    /* This is a special form.  Invoke it, passing the unevaluated 
       arglist plus the current environment.  */
    fpsf = (FUNCSF)item2function(function);
    return (*fpsf)(newargs,environment);
  }

  if (arity > MAXARITY){
    /* This function takes a variable number of inputs.  
       Pass them to it as a list. */
    fp1 = (FUNC1)item2function(function);
    return (*fp1)(newargs);
  }

  return applycfunc(function,newargs);
}

/* Evaluate lambda expression, returning a closure */
static Item evallambda(Item args, Schwa_env *environment)
{
  Item formals, ptr;
  Closure *output;

  output = (Closure *)safe_malloc(sizeof(Closure)); 
  
  formals = car(args);
  
  /* Sanity check formals */
  if (item2type(formals) != TYPELIST)
    error("Non-list %s as first arg in lambda\n", printitem(formals));
  ptr = formals;
  while (!null(ptr)){
    if (item2type(car(ptr)) != TYPESYMBOL)
      error("Non-symbol %s in list of formals for lambda expression\n", 
	    printitem(car(ptr)));
    ptr = cdr(ptr);
  }

  /* Should formals, body be copied here, rather than just re-using list structure? */
  output->formals = formals;
  output->body = cdr(args);
  output->bindings = environment->localbindings;

  return(closure2item(output));
}


/* Evaluate a list of expressions in this environment, 
   returning list of evaluated results. 
   Caller should ensure that exprs is really a list. */
Item evalargs(Item exprs, Schwa_env *environment)
{
  Item output = NIL, value;

  while (!null(exprs)){
    value = eval(car(exprs), environment);
    output = cons(value,output);
    exprs = cdr(exprs);
  }
  return nreverse(output);
}


/* ================================================================ */
/* Symbol bindings                                                  */
/* ================================================================ */

/* See if symbol is eq to first element of any pair in list.
   If so, return the first such pair.  Otherwise returns MISSING. */
static Item findbinding(Item symbol, Item list)
{
  Item pair;

  while (!null(list)){
    pair = car(list);
    if (eq(car(pair), symbol))
      return pair;
    list = cdr(list);
  }
  return MISSING;
}

/* To evaluate a symbol, first check if there is a local binding, then
   check dynamic bindings, then access global environment. */
Item evalsymbol(Item symbol, Schwa_env *environment)
{
  Item binding;
  
  /* The following special symbols evaluate to themselves. */
  if (eq(symbol,END) || eq(symbol,ANY) || eq(symbol,MISSING))
    return symbol;
  binding = findbinding(symbol,environment->localbindings); 
  if (present(binding))
    return cadr(binding);
  else {
    binding = findbinding(symbol,environment->dynamicbindings); 
    if (present(binding))
      return cadr(binding);
    return hashget(environment->globalbindings, symbol);
  }
}

/* If there's a local binding, reset value for that binding.
   Otherwise check for a dynamic bindings.
   Otherwise reset value in global environment. */
void setsymbol(Item symbol, Item value, Schwa_env *environment)
{
  Item binding;
  
  if (eq(symbol,END) || eq(symbol,ANY) || eq(symbol,MISSING))
    error("Attempt to rebind %s\n", printitem(symbol));
  binding = findbinding(symbol,environment->localbindings); 
  if (present(binding))
    set_car(cdr(binding),value);
  else {
    binding = findbinding(symbol,environment->dynamicbindings); 
    if (present(binding))
      set_car(cdr(binding),value);
    hashset(environment->globalbindings, symbol, value);
  }
}

/* ================================================================ */
/* Non-local exits                                                  */
/* ================================================================ */

/* This is a very, very simple sort of non-local exit facility,
   which allows you to return to the most recent enclosing block.
   The break function passes a return value up to this block.
   scope is purely dynamic. */

/* Set up an enclosing block from which returns are possible. */
static Item evalblock(Item args, Schwa_env *environment)
{
  int rv, issaved = 0;
  sigjmp_buf *saved = 0;
  Item output;
  Item localbindings, dynamicbindings;

  /* Save any pre-existing jump point */
  if (environment->breakisset){
    issaved = 1;
    saved = environment->breakpoint;
  }

  /* Save local and dynamic bindings */
  localbindings = environment->localbindings;
  dynamicbindings = environment->dynamicbindings;

  /* Set up jump point */
  environment->breakisset = 1; 
  environment->breakpoint = safe_malloc(sizeof(sigjmp_buf));
  rv = sigsetjmp(*(environment->breakpoint), 1);
  if (rv == 0)
    /* evaluate the code inside this expression */
    output = evalsequence(args,environment);
  else {
    /* we are returning from a breakpoint inside the code */
    output = environment->breakvalue;
  }

  /* Restore any pre-existing jump point */
  if (issaved)
    environment->breakpoint = saved;
  else
    environment->breakisset = 0;

  /* Restore local and dynamic bindings */
  environment->localbindings = localbindings;
  environment->dynamicbindings = dynamicbindings;

  return output;
}

/* Return to most recent enclosing block */
static Item evalreturn(Item args, Schwa_env *environment)
{

  if (!(environment->breakisset))
    error("Return was called outside the scope of any block\n");
  if (length(args) > 1)
    error("Too many inputs to return:  %s\n", printitem(args));
  if (null(args))
    environment->breakvalue = MISSING;
  else
    environment->breakvalue = eval(car(args), environment);
  siglongjmp(*(environment->breakpoint), 1);
}


/* ================================================================ */
/* Helpers for evaluating different types of expressions            */
/* ================================================================ */

/* Skip shell script header line (starting with #!) at the top of the file, 
   if there is one.  Leaves file pointer at the start of the next line (if
   there was a script line) or at the start of the file (if there wasn't). */
static void script_kludge(FILE *fp)
{
  int inchar;

  inchar = fgetc(fp);
  if (inchar != '#'){
    ungetc(inchar,fp);
    return;
  }
  inchar = fgetc(fp);
  if (inchar != '!'){
    rewind(fp);
    return;
  }
  /* Ok, we've found a script line.   Now read through next newline. */
  while ((inchar = fgetc(fp)) != EOF && inchar != '\n');
}


/* Read forms one-by-one from a file, evaluating each one.  If
   interactive is non-zero, does echoing appropriate for an
   interactive session.  Otherwise, assumes we're evaluating a shell
   script and executes silently.  In any case, stops reading file if
   it encounters an error, since failure of one step typically results
   in garbage from later ones.  */
void evalfile(char *filename, Schwa_env *environment, int interactive)
{
  FILE *fp;
  Item oldloads;

  fp = fopen(filename, "r");
  oldloads = openloads;
  openloads = cons(ptr2item(fp, string2symbol("FILE_HANDLE")), openloads);
  if (!fp)
    error("Can't open file %s\n", filename);
  script_kludge(fp);
  if (interactive)
    rev_loop(environment, fp, stdout);
  else
    rev_loop(environment, fp, 0);
  fclose(fp);
  openloads = oldloads;
}


/* Evaluate a list of expressions in this environment, 
   in order.  Returns value of last expression. 
   Returns MISSING if input contains no expressions.
   Caller should ensure that exprs is really a list. */
Item evalsequence(Item exprs, Schwa_env *environment)
{
  Item ptr, value = MISSING;
  ptr = exprs;
  while (!null(ptr)){
    value = eval(car(ptr), environment);
    ptr = cdr(ptr);
  }
  return value;
}

/* Evaluate a let expression.  */
static Item evallet(Item args, Schwa_env *environment)
{
  Item newbindings, expr, value, symbol;
  Item oldbindings, pairs, pair;

  if (null(args))
    error("No args to while\n");

  /* Save current set of bindings. */
  oldbindings = environment->localbindings;

  /* Add new bindings for local variables declared by let. */
  pairs = car(args);
  expr = cdr(args);
  if (item2type(pairs) != TYPELIST)
    error("Binding list for eval is non-list %s\n", printitem(pairs));

  newbindings = NIL;

  /* Collect up the new bindings.   Notice that initialization
     expressions are evaluated in the old environment without any
     of these new bindings in place. */
  while (!null(pairs)){
    pair = car(pairs);
    pairs = cdr(pairs);
    if (item2type(pair) != TYPELIST || length(pair) < 1 || length(pair) > 2)
      error("Malformed binding %s\n", printitem(pair));
    symbol = car(pair);
    if (length(pair) < 2)
      value = MISSING;
    else
      value = eval(cadr(pair), environment);
    newbindings = cons(list(2,symbol,value), newbindings);
  }

  /* Install the new bindings. */
  environment->localbindings = append(newbindings, environment->localbindings);

  /* Evaluate expression and restore old set of bindings. */
  value = evalsequence(expr, environment);
  environment->localbindings = oldbindings; 
  
  return value;
}


static Item evalset(Item args, Schwa_env *environment)
{
  Item value, symbol;
  if (length(args) != 2)
    error("Wrong # args to set:  %s", printitem(args));
  symbol = car(args);
  assert_input_type(symbol, TYPESYMBOL);
  value = eval(nth(1, args), environment);
  setsymbol(symbol, value, environment);
  return value;
}

/* The guts of evallist */
static Item evallistx(Item form, Schwa_env *environment)
{
  Item head, args; 
  static int allsetup = 0;
  static Item settag, quotetag, lambdatag;
  static Item begintag, lettag, livetag, debugontag, debugofftag;
  static Item returntag, blocktag;

  if (!allsetup){
    settag = string2symbol("set");
    lambdatag = string2symbol("lambda");
    quotetag = string2symbol("quote");
    begintag = string2symbol("begin");
    lettag = string2symbol("let");
    debugontag = string2symbol("debug_on");
    debugofftag = string2symbol("debug_off");
    livetag = string2symbol("leave_live");
    returntag = string2symbol("return");
    blocktag = string2symbol("block");
    allsetup = 1;
  }

  head = car(form);
  args = cdr(form);

  /* The following forms are handled explicitly by the interpreter,
     either because they require dangerous access to internals or
     because they are so basic that they shouldn't be redefined. */
  if (eq(head,settag))
    return evalset(args, environment);
  else if (eq(head,lambdatag))
    return evallambda(args, environment);
  else if (eq(head,begintag))
    return evalsequence(args, environment);
  else if (eq(head,quotetag))
    return car(args);
  else if (eq(head,lettag))
    return evallet(args, environment);
  else if (eq(head,blocktag))
    return evalblock(args, environment);
  else if (eq(head,returntag))
    return evalreturn(args, environment);
  else if (eq(head,livetag)){             
    environment->leavelive = 1;           
    return MISSING;
  }
  else if (eq(head,debugontag)){
    environment->force_echo = 1;
    return MISSING;
  }
  else if (eq(head,debugofftag)){
    environment->force_echo = 0;
    return MISSING;
  }
  else 
    return apply(head, args, environment);

  /* The compiler can't see that the following is unreacheable. */
  return MISSING;
}

/* Evaluate input list in the input environment.
   Caller must ensure that list is really a list. */
static Item evallist(Item form, Schwa_env *environment)
{
  Item rv, oldtrace;

  /* Push this form onto the list of currently active function calls,
     so we have backtrace information in case of an error break. */
  oldtrace = environment->backtrace;
  environment->backtrace = cons(form, oldtrace);
  rv = evallistx(form, environment);
  environment->backtrace = oldtrace;
  return rv;
}



/* ================================================================ */
/* Argument processing for shell scripts                            */
/* ================================================================ */


/* Call this when interpreter is invoked via a shell script.
   Moves command-line arguments into the environment variable *args*. */
static Item parse_shell_args(int argc, char *argv[], Schwa_env *environment)
{
  int i;
  Item args = NIL;
   
  if (argc < 1)
    error("Parse_shell_args called with no input args.\n");

  for (i = 2; i < argc; i++)
    args = cons(string2item(argv[i]), args);
  setsymbol(string2symbol("*args*"), nreverse(args), environment);
  return MISSING;
}


/* ================================================================ */
/* Top-level eval function                                          */
/* ================================================================ */

/* Evaluate the input expression in the input environment,
   possibly modifying the environment. */
Item eval(Item expression, Schwa_env *environment)
{
  int type;
  type = item2type(expression);
  switch (type){
  case TYPELIST:
    if (null(expression))   /* So you don't have to put quote before #NIL */
      return expression;
    else
      return evallist(expression, environment);
    break;
  case TYPESYMBOL:
    return evalsymbol(expression,environment);
    break;
  default:
    /* When in doubt, the expression evaluates to itself. 
       This is the standard right answer for objects like numbers,
       a plausible answer for things like bundles, and a sensible 
       failsafe for odd things like special objects. */
    return expression;
    break;
  }
}



