/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/



/* EQ hash tables

   Margaret M. Fleck, 2005

   Keys should be Items that are symbols, integers, or ngrams.  The
   code will accept other types of Items, but the resulting behavior
   may not be what you wanted, because keys are compared using eq.
   Numerical precision issues could also arise with floats.

   To use other sorts of keys, you should either
      -- ensure that you don't have two keys that look alike but aren't
         eq, e.g. use a lookup table to map floating point numbers to
         a small set of canonical float Items. 
      -- or, convert these values into symbols, integers, or ngrams
         before using them as keys, e.g. feed strings through string2symbol

   Retrieval functions return MISSING when a key has no corresponding value.
   Keys that lack values are not distinguished from keys whose value has been
   explicitly set to MISSING.   Keys with value MISSING are not returned by
   the iterator (hashstart and hashnext) and the function (hashkeys) listing 
   all keys 

   The table arrays are segmented into blocks, so that they can be allocated
   incrementally in small pieces.   The garbage collector will not be happy
   if we keep allocating and deallocating really large arrays.   Smaller
   blocks are used for small hash tables, to avoid wasted space.

   Hash tables may not be the best approach if you plan to store values
   for a VERY large number of items, e.g. big n-gram tables.   It
   may be more efficient to extract type-specific codes (e.g. symbol2code,
   ngram2code) and create an array mapping these numbers directly to values.
*/

#include "lists.h"


/* Table sizes for this algorithm are powers of 2.   Initially the table size is
   STARTSIZE = 2^STARTBITS  */
#define STARTBITS 2
#define STARTSIZE 4

#define MAXOCCUPANCY 4.0  /* Table is considered full when average number of entries
			   per bucket reaches this value. */
/* Constants used for dividing big tables into reasonable-sized blocks.   
   Stay well clear of GC's chunk size of 4K */
#define BLOCKSIZE 512
#define BLOCKSHIFT 9
#define BLOCKBITS 511


/* Remove key from table. */
static void hashremove(Item key, Hashtable *table)
{
  unsigned long bin;
  Kvlist *pair = 0, *previous = 0, *ptr = 0;
  Kvlist **block;

  bin = hashint(key.uinteger,table->bits);

  block = table->table[bin>>BLOCKSHIFT];
  ptr = block[bin&BLOCKBITS];
  while (ptr){
    if (eq(ptr->key,key)){
      pair = ptr;
      break;
    }
    previous = ptr;
    ptr = ptr->next;
  }
  
  if (pair){
    table->numoccupants -= 1;
    if (previous)
      previous->next = pair->next;
    else  
     block[bin&BLOCKBITS] = pair->next;
  }
}


/* Allocate and return a new block of space of the specified size.
   The output is an array of blocks.  Each block is BLOCKSIZE, except
   that the last block may be smaller.  The smaller last block
   prevents major wasted space for small tables.  Each cell is set to
   zero. */
static Kvlist ***new_table(unsigned long size)
{
  Kvlist ***table;
  unsigned long i, numblocks;

  numblocks = size/BLOCKSIZE;   
  if (numblocks*BLOCKSIZE < size)
    numblocks++;

  if (numblocks == 1){
    table = safe_malloc(sizeof(Kvlist **));
    table[0] = safe_malloc(size*sizeof(Kvlist *));
  }
  else {
    table = safe_malloc(numblocks*sizeof(Kvlist **));
    for (i = 0; i < numblocks; i++){
      table[i] = safe_malloc(BLOCKSIZE*sizeof(Kvlist *));
    }
  }
  for (i = 0; i < size; i++)
    table[i>>BLOCKSHIFT][i&BLOCKBITS] = 0;

  return table;
}

Hashtable *new_hashtable (void)
{
  Hashtable *output;

  output = safe_malloc(sizeof(Hashtable));
  
  output->bits = STARTBITS;
  output->tablesize = STARTSIZE;
  output->numoccupants = 0;
  output->table = new_table(STARTSIZE);

  return output;
}


/* Expand size of table by a factor of two. */
static void expand_table(Hashtable *table)
{
  unsigned long oldsize, newsize, i;
  Kvlist ***olddata, *ptr;

  newsize = 2*(table->tablesize);
  olddata = table->table;
  oldsize = table->tablesize;
  table->bits++;
  table->tablesize = newsize;
  table->table = new_table(newsize);
  table->numoccupants = 0;


  /* rehash data in the table. */
  for (i = 0; i < oldsize; i++){
    ptr = olddata[i>>BLOCKSHIFT][i&BLOCKBITS];
    while (ptr){
      hashset(table, ptr->key, ptr->value);
      ptr = ptr->next;
    }
  }
  /*   fprintf(stderr, "Rebuilt Table has %ld occupants\n", table->numoccupants); */
}

/* Adds (key,value) pair, replacing existing value for this key (if any).  
   Set a key's value to MISSING to delete it from the table. 
*/
void hashset(Hashtable *table, Item key, Item value)
{
  unsigned long bin;
  Kvlist *ptr = 0, *pair = 0, *oldlist;
  Kvlist **block;

  if (eq(key,MISSING)){
    hashremove(key,table);
    return;
  }

  if (table->numoccupants >= MAXOCCUPANCY*table->tablesize){
    expand_table(table);
  }

  bin = hashint(key.uinteger,table->bits);

  block = table->table[bin>>BLOCKSHIFT];
  ptr = block[bin&BLOCKBITS];
  oldlist = ptr;

  while (ptr){
    if (eq(ptr->key, key)){
      pair = ptr;
      break;
    }
    ptr = ptr->next;
  }

  if (!pair){
    pair = safe_malloc(sizeof(Kvlist));
    pair->key = key;
    pair->next = oldlist;
    block[bin&BLOCKBITS] = pair;
    table->numoccupants += 1;
  }

  pair->value = value;
  return;
}


/* Returns the value associated with this key, or MISSING if no matching pair. */
Item hashget(Hashtable *table, Item key)
{
  unsigned long bin;
  Kvlist *ptr = 0;
  Kvlist **block;

  bin = hashint(key.uinteger,table->bits);

  block = table->table[bin>>BLOCKSHIFT];
  ptr = block[bin&BLOCKBITS];

  while (ptr){
    if (eq(ptr->key, key)){
      return ptr->value;
    }
    ptr = ptr->next;
  }

  return MISSING;
}


/* Adds new value to the list stored under key. */
void hashadd(Hashtable *table, Item key, Item value)
{
  Item previous;

  previous = hashget(table, key);
  if (present(previous))
    hashset(table, key, cons(value, previous));
  else
    hashset(table, key, list(1,value));
}

/* Increments integer value stored under key. */
void hashincrement(Hashtable *table, Item key, int amount)
{
  Item previous;

  previous = hashget(table, key);
  if (present(previous))
    hashset(table, key, long2item(amount+item2long(previous)));
  else
    hashset(table, key, long2item(amount));
}

/* ================================================================ */
/* Iterating through hash table                                     */
/* ================================================================ */

/* Return an iterator pointing at start of hashtable */
Hashptr *hashstart(Hashtable *table)
{
  Hashptr *output;

  output = (Hashptr *)safe_malloc(sizeof(Hashptr));
  output->table = table->table;
  output->bin = 0;
  output->tablesize = table->tablesize;
  output->remaining = (table->table)[0][0];

  return output;
}

/* Returns a (newly-allocated) key/value pair.   
   Returns END at end of table.
   Advances ptr to next location in table.  */
Item hashnext(Hashptr *ptr)
{
  Item output;
  Kvlist *remaining;
  unsigned long bin;

  remaining = ptr->remaining;
  while (!remaining){
    /* No pairs, or no more pairs, in this bin, must move onto next bin. */
    ptr->bin += 1;
    if (ptr->bin >= ptr->tablesize)
      return END;
    bin = ptr->bin;
    remaining = (ptr->table)[bin>>BLOCKSHIFT][bin&BLOCKBITS];
    ptr->remaining = remaining;
  }
  
  /* We're now at a bin containing more pairs. */
  output = list(2,remaining->key,remaining->value);
  ptr->remaining = remaining->next;
  return output;
}


/* Returns a seq containing all keys for the hash table. 
   By running the output through sortseq, you can print out the
   hashtable in a human-friendly order. */
Item *hashkeys(Hashtable *table)
{
  long length, i;
  Item *output, pair;
  Hashptr *ptr;


  length = table->numoccupants;
  output = safe_malloc((length+1)*sizeof(Item));
  output[length] = END;
  
  /* put keys into output table */
  ptr = hashstart(table);
  i = 0;
  while (!at_end(pair = hashnext(ptr))){
    output[i] = car(pair);
    i++;
  }
  return output;
}

/* ================================================================ */
/* For debugging only                                               */
/* ================================================================ */

#if 0

/* Display occupancy pattern of table. 
   Groupsize determines how many consecutive bins are added together.
   Groupsize=1 reports counts for every individual bin. */
void print_occupancy(Hashtable *table, int groupsize)
{
  unsigned long i;
  int count = 0, total = 0;
  Kvlist *ptr;

    fprintf(stderr, "Table has %ld occupants\n", table->numoccupants);

  for (i = 0; i < table->tablesize; i++){
    ptr = table->table[i>>BLOCKSHIFT][i&BLOCKBITS];
    while (ptr){
      count++;
      total++;
      /*       fprintf(stderr, " %s\n", printitem(ptr->key)); */
      ptr = ptr->next;
    }
    if ((i+1)%groupsize == 0 || i == table->tablesize){
      fprintf(stderr, "[%ld] %d \n", i, count);
      count = 0; 
    }
  }
  fprintf(stderr, "Total in cells %d\n", total);
} 
 
#endif
