/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* Data structures that can contain a diversity of elements.
   Aka stripped down mini-Scheme, with a few special features.
   Margaret M. Fleck, 2004-2005 */

#include "lists.h"

/* ================================================================ */
/* Typing                                                           */
/* ================================================================ */

/* The least-significant two bits of each item contain its first-level type:  
   integer, symbol, string, or complex.

   integer:          the most significant 30 bits contain the value
   symbol, ngrams:   the most 29 bits contains the value, the 8 bit
                     distinguishes symbols from ngrams
   string:           masking off the two type bits yields a (char *)
   complex:          masking off the two type bits yields a (Pair *)

   A Pair contains two fields, car and cdr.   The least-significant two
   bits of the cdr determine the subtype:  cons cell, feature bundle,
   or float;

                    type               subtype (if complex)
   TAGNUMBER        integer            miscelleaneous built-in types
   TAGCOMPLEX       complex            cons cell
   TAGSYMBOL        symbol/ngram       feature bundle
   TAGSTRING        string             special

   Some code depends on the correspondence between type and subtype
   tags, e.g. that the cdr of a cons cell has the same subtype tag as 
   is used to mark complex objects.  The special object MISSING is guaranteed
   to have all zero bits (and, therefore, TAGSYMBOL is 0).

   "Miscellaneous built-in types" is a catch-all for types that require
   significant support from the core interpreter and/or reader, but which
   aren't heavily-used enough to be given their own 2-bit tags.  Currently, 
   this category includes floats, functions, seqs, and closures.

   "Special" is a type that allows users to embed pointers to arbitrary
   C objects into our data structures, along with a symbolic label that
   (hopefully) means something to the user.   Special objects can be
   passed to/from functions in the interpreter/scripting interface.
   Except for a rudimentary sort of printing, support is left up to the user. 

   Effective limits on numerical precision
      integers:  30 bits (almost a long)
      floats:     32 bits (float)
      maximum number of symbols or ngrams:  2**29

*/


/* Looks only one level deep, for main type tags */
static unsigned int item2type1 (Item input)
{
  unsigned long output;

  output = input.sinteger & TAG;

  if (output == TAGSYMBOL)
    output = input.uinteger & SYMTAG;

  return (unsigned int) output;
}

/* Looks only one level deep, for subtype tags. */
static unsigned int item2type2 (Item input)
{
  unsigned long output;

  output = input.uinteger & TAG;

  if (output == TAGNUMBER)
    output  = input.uinteger & LONGTAG;

  return (unsigned int) output;
}

/* Return type of input item.   Looks at type and subtype (when appropriate).
   Return value should be one of 
     TYPEINT,  TYPESTRING,  TYPESYMBOL,  TYPELIST,  TYPEFLOAT,  TYPEBUNDLE 
     TYPEFUNCTION, TYPESEQ, TYPENGRAM, TYPESPECIAL , TYPECLOSURE */
unsigned int item2type (Item input)
{
  unsigned long output;
  unsigned long output2;
  Pair *ptr;

  output = item2type1(input);

  if (output == TAGCOMPLEX){
    ptr = (Pair *)(input.uinteger & ~TAG);
    if (!ptr)
      output2 = TAGCOMPLEX;    /* Items containing a null pointer represent 
                                  the empty list NIL. */
    else
      output2 = item2type2(ptr->cdr);
    output = (output2 << SUBTYPESHIFT) + output;
  }

  return (unsigned int) output;
}


/* ================================================================ */
/* Magic constants                                                  */
/* ================================================================ */

/* The empty list. */
const Item NIL = {TAGCOMPLEX};

/* No value is available.  This is guaranteed to have all zero bits.  */
const Item MISSING = {TAGSYMBOL}; 

/* A feature has been set, but no specific value has been assigned to it. */
const Item ANY = {TAGSYMBOL+8};

/* The end of something, such as a file or a string composed of Items. 
   This symbol has a restricted range of uses.  In particular, it cannot be
   freely embedded in input streams/files because readers tend to interpret it
   as truncating the input. */
const Item END = {TAGSYMBOL + 16};

/* The first non-reserved symbol code.  This is > 0 iff we've 
   decided to reserve some codes at the beginning.  */
/* Code 0 is reserved for the special value MISSING */
/* Code 1 is reserved for the special value ANY */
/* Code 2 is reserved for the special value END */
#define FIRSTCODE 3

/* ================================================================ */
/* Making lists                                                     */
/* ================================================================ */

/* For atomic objects, this is ==.
   For non-atomic objects, this is a simple pointer comparison. */
int eq(Item one, Item two)
{
  return(one.uinteger == two.uinteger);
}

/* Returns non-zero iff the input is not MISSING or END. */
int present(Item input)
{
  return !eq(input,MISSING) && !eq(input,END);;
}

/* Is input a special symbol? */
static int special_symbol(Item input)
{
  return (eq(input, MISSING) || eq(input, END) || eq(input,ANY));
}

/* Returns non-zero iff the input is NIL */
int null(Item input)
{
  return eq(input,NIL);
}

/* Returns non-zero iff the input is END */
int at_end(Item input)
{
  return eq(input,END);
}

/* Intended for internal use only. */
static Item pair2item(Pair *ptr)
{
  Item output;
  unsigned long tagged;

  tagged = (unsigned long) ptr;
  tagged = tagged | TAGCOMPLEX;

  output.uinteger = tagged;
  return output;
}

/* Intended for internal use only.  
   Assumes you've already check that input is a complex object. */
static Pair *item2pair(Item input)
{
  unsigned long tagged;

  tagged = input.uinteger;
  tagged = tagged & (~TAG);
  return (Pair *) tagged;
}


/* The cdr must be a list.  Dotted pairs are not supported. 
   This code depends on the fact that the same type tag TAGCOMPLEX 
   is used for both type and subtype of lists. */
Item cons (Item car, Item cdr)
{
  Pair *cell;

  assert_input_type(cdr, TYPELIST);
  
  cell = (Pair *)safe_malloc(sizeof(Pair));
  cell->cdr = cdr;
  cell->car = car;
    
  return pair2item(cell);
}

Item car (Item list)
{
  Pair *cell;

  assert_input_type(list, TYPELIST);
  if (eq(list,NIL))
    error("Tried to take the car of NIL.\n");

  cell = item2pair(list);
  return cell->car;
}

Item cdr (Item list)
{
  Pair *cell;

  assert_input_type(list, TYPELIST);
  if (eq(list,NIL))
    error("Tried to take the cdr of NIL.\n");
  
  cell = item2pair(list);
  return cell->cdr;
}

Item cadr (Item list)
{
  return car(cdr(list));
}

/* Make a list out of the input items.
   First argument is the total number of items. */
Item list(unsigned int n, ...)
{
  int i;
  Item output = NIL;
  Item next;
  va_list ap;

  va_start(ap, n);
  for (i = 0; i < n; i++){
    next = va_arg(ap, Item);
    output = cons(next, output);
  }
  va_end(ap);

  output = reverse(output);
  return output;
}

void set_car (Item list, Item value)
{
  Pair *cell;

  assert_input_type(list, TYPELIST);
  
  cell = item2pair(list);
  cell->car = value;
}

/* Value must be a list.   Dotted pairs are not supported. */
void set_cdr (Item list, Item value)
{
  Pair *cell;

  assert_input_type(list, TYPELIST);
  assert_input_type(value, TYPELIST);
  
  cell = item2pair(list);
  cell->cdr = value;
}


Item reverse (Item list)
{
  Item output = NIL;
  assert_input_type(list, TYPELIST);
  
  while (!null(list)){
    output = cons(car(list), output);
    list = cdr(list);
  }
  return output;
}

/* in-place version of reverse */
Item nreverse (Item list)
{
  Item output = NIL;
  Item next;
  assert_input_type(list, TYPELIST);
  
  while (!null(list)){
    next = list;
    list = cdr(list);
    set_cdr(next,output);
    output = next;
  }
  return output;
}

/* Make a shallow copy of list, i.e. makes new top-level list
   cells but reuses lower-level structure. */
Item copylist (Item list)
{
  return nreverse(reverse(list));
}

/* Append the two lists destructively. */
Item nconc (Item list1, Item list2)
{
  Item ptr;

  if (null(list1))
    return list2;
  ptr = list1;
  while (!null(cdr(ptr)))
    ptr = cdr(ptr);
  set_cdr(ptr,list2);
  return list1;
}

/* Append the two lists, non-destructively */
Item append (Item list1, Item list2)
{
  return nconc(copylist(list1), list2);
}

/* Is this object a member of this list?  Equality is determined by eq. */
int memq(Item object, Item list)
{
  Item ptr;
  ptr = list;
  while (!null(ptr)){
    if (eq(car(ptr), object))
      return 1;
    ptr = cdr(ptr);
  }
  return 0;
}

/* Remove all instances of this object from list (non-destructively). */
Item delq(Item object, Item list)
{
  Item ptr, output = NIL, next;
  ptr = list;

  while (!null(ptr)){
    next = car(ptr);
    if (!eq(next, object))
      output = cons(next, output);
    ptr = cdr(ptr);
  }
  return reverse(output);
}

/* Is this object a member of this list?  Equality is determined by equal. */
int member(Item object, Item list)
{
  Item ptr;
  ptr = list;
  while (!null(ptr)){
    if (equal(car(ptr), object))
      return 1;
    ptr = cdr(ptr);
  }
  return 0;
}

long length(Item list)
{
  int output = 0;
  while (!null(list)){
    output++;
    list = cdr(list);
  }
  return output;
}

Item nth(long n, Item list)
{
  Item ptr;
  int i;
  assert_input_type(list, TYPELIST);
  if (n < 0)
    error("first arg for nth was negative %ld\n", n);
  ptr = list;
  i = n;
  while (i > 0){
    if (null(ptr))
      error("nth: list %s is too short to have an element number %ld\n", printitem(list), n);
    ptr = cdr(ptr);
    i--;
  }
  if (null(ptr))
    error("nth: list %s is too short to have an element number %ld\n", printitem(list), n);
  return car(ptr);
}

/* List should be a list of pairs.
   Finds pair whose car is object and returns its cadr.
   Returns MISSING if no such pair. */   
Item assq(Item object, Item list)
{
  Item pair;
  while (!null(list)){
    pair = car(list);
    if (eq(car(pair), object))
      return (cadr(pair));
    list = cdr(list);
  }
  return MISSING;
}

/* Two items are equal if 
     -- they contain equal elements in the same order (lists, strings, seqs)
        or in the same structure (bundles)
     -- they are eq (other types)
*/
int equal(Item one, Item two)
{
  int type, i;
  Item features, *sone, *stwo;

  type = item2type(one);
  if (type != item2type(two))
    return 0;
  switch (type) {
  case TYPELIST:
    while(!null(one)){
      if (null(two) || !equal(car(one), car(two)))
	return 0;
      one = cdr(one);
      two = cdr(two);
    }
    if (!null(two))
      return 0;
    else
      return 1;
    break;
  case TYPESTRING:
    return !strcmp(item2string(one), item2string(two));
    break;
  case TYPESEQ:
    sone = item2seq(one);
    stwo = item2seq(two);
    i = 0;
    while (!at_end(sone[i])){
      if (!equal(sone[i],stwo[i]))
	return 0;
      i++;
    }
    if (!at_end(stwo[i]))
      return 0;
    else 
      return 1;
    break;
  case TYPEBUNDLE:
    if (!eq(bundlehead(one),bundlehead(two)))
      return 0;
    features = bundlefeats(one);
    if (!equal(features,bundlefeats(two)))
      return 0;
    while (!null(features)){
      if (!equal(bundleget(one,car(features)), bundleget(two,car(features))))
	return 0;
      features = cdr(features);
    }
    return 1;
    break;
  default:
    return eq(one,two);
  };
}


/* ================================================================ */
/* Bundles                                                          */
/* ================================================================ */

/* Bundles are sets of feature/value pairs.  They are specifically
   intended to be used as
      -- dynamically-typed versions of structs/records
      -- linguistic feature bundles

   The name of each feature is a symbol.

   Each bundle has an optional head (also a symbol).   The head can be 
   used to indicate the type of a dynamic structure (e.g. for debugging and
   for the benefit of functions manipulating it).  Linguistic feature
   bundles are often headless.

   All features not explicitly assigned a value are considered to have
   the value MISSING.  Setting a feature's value to MISSING deletes
   it from the bundle.

   The "signature" of a bundle is an ngram consisting of the head followed by 
   the list of features which currently have (non-MISSING) values assigned,
   in their current storage order.   The signature typically changes across
   the lifespan of the bundle.  Signatures are used internally in the
   implementation.

   The signature lists features in the order they were added to the bundle.

   Bundles are intended to be used for storing relatively short lists of
   features (use a hash table for large sets) and, specifically, for
   applications which create a large number of bundles with identical signatures.
   (Normal human behavior and coding practice mean that bundles containing
   the same set of features will tend to use only a few distinct orderings
   for these features.)

   In this implementation, bundles are stored as
         cdr field:   the signature (an ngram)
         car field:   pointer to an array of storage positions   

   The signature is stored in the cdr field because the cdr field
   determines the subtype of complex objects and we're using the fact
   that the head of a bundle is always a symbol. 

   The array is allocated at exactly the right length to contain the number of 
   features currently in use.  
*/


/* Makes a new bundle with the specified symbol as head.  
   Supply the input MISSING if you want to create a headless bundle. */
Item new_bundle(Item head)
{
  Pair *cell;
  assert_input_type(head, TYPESYMBOL);
  cell = (Pair *)safe_malloc(sizeof(Pair));
  cell->cdr = list2ngram(list(1,head));  
  (cell->car).ptr = 0;
  return pair2item(cell);
}

/* Returns the index of this feature in the input signature (represented
   as a list) or -1 if it's not there. */
static int feature_position(Item feature, Item siglist)
{
  int pos = 0;
  Item featlist;

  featlist = cdr(siglist);
  while (!null(featlist)){
    if (eq(car(featlist),feature))
      return pos;
    featlist = cdr(featlist);
    pos++;
  }
  return -1;
}

/* Gets the value associated with this feature in bundle.
   Returns MISSING if no value for the feature. */
Item bundleget(Item bundle, Item feature)
{
  Item signature, siglist;
  Pair *cell;
  Item *array;
  int pos;

  assert_input_type(feature, TYPESYMBOL);
  assert_input_type(bundle, TYPEBUNDLE);

  cell = item2pair(bundle);
  signature = cell->cdr;
  siglist = ngram2list(signature);
  pos = feature_position(feature, siglist);

  if (pos < 0)
    return MISSING;
  else {
    array = (Item *)(cell->car).ptr;
    return array[pos];
  }
}

/* Has this feature been set (i.e. assigned a value, which could be ANY) 
   in this bundle?  */
int bundleisset(Item bundle, Item feature)
{
  Item signature, siglist;
  Pair *cell;
  int pos;

  assert_input_type(feature, TYPESYMBOL);
  assert_input_type(bundle, TYPEBUNDLE);

  cell = item2pair(bundle);
  signature = cell->cdr;
  siglist = ngram2list(signature);

  pos = feature_position(feature, siglist);
  if (pos >= 0) return 1;
  else return 0;
}

/* Sets or resets the value of feature in this bundle.
   Feature must be a symbol. 
   To unset a feature (make it disappear entirely from the bundle), 
   set its value to MISSING. */
Item bundleset(Item bundle, Item feature, Item value)
{
  Item signature, siglist;
  Pair *cell;
  Item *array;

  int pos, i, numfeats;
  if (eq(feature,MISSING))
    error("bundleset:   MISSING can't be used as feature name.\n");
  assert_input_type(feature, TYPESYMBOL);
  assert_input_type(bundle, TYPEBUNDLE);
  cell = item2pair(bundle);
  signature = cell->cdr;
  siglist = ngram2list(signature);
  numfeats = length(cdr(siglist));
  array = (Item *)(cell->car).ptr;

  pos = feature_position(feature, siglist);
  if (eq(value,MISSING)){    /* user wants to delete this feature */
    if (pos >= 0){          /* and it's actually present in this bundle */
      signature = list2ngram(cons(car(siglist),delq(feature,cdr(siglist))));
      /* Update storage array to remove feature. */
      numfeats--;
      if (numfeats == 0)  /* we're removing the last feature */
	array = 0;
      else{
	for (i = pos+1; i <= numfeats; i++)
	  array[i-1] = array[i];
	array = (Item *)safe_realloc(array, numfeats*sizeof(Item));
      }}}
  else {  /* We're adding a real value for this feature. */
    if (pos < 0){   /* Feature isn't present;  create a location for it at end of storage array. */
      signature = list2ngram(cons(car(siglist),nconc(cdr(siglist),list(1,feature))));
      pos = numfeats;
      numfeats++;
      if (array == 0)
	array = (Item *)safe_malloc(sizeof(Item));
      else
	array = (Item *)safe_realloc(array,numfeats*sizeof(Item));
    }
    /* Feature is now present, set its value. */
    array[pos] = value;
  }

  /* Make sure bundle contains the latest version of everything and return value. */
  cell->cdr = signature;
  (cell->car).ptr = array;
  return value;
}

/* Returns the head symbol of the bundle. */
Item bundlehead(Item bundle)
{
  Pair *cell;
  Item signature;
  assert_input_type(bundle, TYPEBUNDLE);
  cell = item2pair(bundle);
  signature = cell->cdr;
  return car(ngram2list(signature));
}

/* Returns the list of features in the bundle.  */
Item bundlefeats(Item bundle)
{
  Pair *cell;
  Item signature;
  assert_input_type(bundle, TYPEBUNDLE);
  cell = item2pair(bundle);
  signature = cell->cdr;
  return cdr(ngram2list(signature));
}


/* ================================================================ */
/* Miscellaneous complex objects                                    */
/* ================================================================ */

/* Miscellaneous complex objects all have the same basic 2-bit subtype tag of 
   TAGNUMBER. This occupies the low-order 2 bits of their cdr fields.  The 
   next 4 bits of the cdr field contain an extension to their type tag.   The 
   rest of the cdr field contains supplementary descriptive data as follows:
   
        functions:  the function's arity
	seqs:   the length of the seq (not including the termination END)
	floats: <nothing>

   The car field of these objects contains the main data, e.g. the actual
   float value or the function pointer. */

/* When a complex object is a float, the actual numerical value is
   stored in the car field.   The cdr field contains only the type
   tag. We can't safely use both fields to store a double because
   we don't know how exactly the double is stored. */
Item float2item(float value)
{
  Pair *cell;

  cell = (Pair *)safe_malloc(sizeof(Pair));

  (cell->cdr).uinteger = TAGNUMBER + TAGFLOAT;
  (cell->car).real = value;
  return pair2item(cell);
}

float item2float(Item input)
{
  Pair *cell;
  assert_input_type(input,TYPEFLOAT);
  cell = item2pair(input);
  return (cell->car).real;
}


/* Closure Items are simply wrappers around the corresponding
   C items. */
Item closure2item(Closure *table)
{
  Pair *cell;

  cell = (Pair *)safe_malloc(sizeof(Pair));

  (cell->cdr).uinteger = TAGNUMBER + TAGCLOSURE;
  (cell->car).ptr = table;
  return pair2item(cell);
}

Closure *item2closure(Item input)
{
  Pair *cell;
  assert_input_type(input,TYPECLOSURE);
  cell = item2pair(input);
  return (Closure *)(cell->car).ptr;
}

/* The input pointer should be a pointer to a C function.
   This function's inputs and output must all be type Item,
   and it must take <arity> input values.  */
Item function2item(void *function, unsigned int arity)
{
  Pair *cell;

  cell = (Pair *)safe_malloc(sizeof(Pair));
  (cell->cdr).uinteger = TAGNUMBER + TAGFUNCTION + (arity << SUBTYPESHIFT);
  (cell->car).ptr = function;
  return pair2item(cell);
}

/* Returns the arity of the function stored in this item. */
unsigned int item2arity(Item input)
{
  Pair *cell;
  unsigned int infofield;
  assert_input_type(input,TYPEFUNCTION);
  cell = item2pair(input);
  infofield = (cell->cdr).uinteger;
  return (infofield >> SUBTYPESHIFT);
}

/* Caller will need to coerce the return value to the appropriate 
   specific C function type. */
void *item2function(Item input)
{
  Pair *cell;
  assert_input_type(input, TYPEFUNCTION);
  cell = item2pair(input);
  return (cell->car).ptr;
}



/* ================================================================ */
/* Utilities for seqs, which are string-like arrays of Items        */
/* ================================================================ */

/* A seq is an array of Items, terminated by the special Item END.
   Thus, it is like a string but composed of Items rather than chars.
   Seqs support two access methods:   sequential scanning from the start
   (stopping at END) and random access.   

   Your C code can manipulate seqs as either raw C arrays or as Items.
   Many utilities operate on the raw C arrays, because this will be
   most familiar to the C programmer.  The terminating mark END allows
   you to do many operations in C without explicit representation of
   the array's length.  However, it is your responsibility to keep
   track of its length if you wish to do random access (rather than
   sequential scanning), resize the underlying array, or move the
   position of END.

   Converting to the item representation (e.g. seq2item) gives you
   access to several useful features:
        -- nice display of the seq via the printitem function
        -- pre-computed length information, to simplify random access
	-- random access functions with automatic bounds checking
   However, it limits your ability to resize the array.
       
   Specifically, because the Item representation contains a
   pre-computed length for the seq, problems may arise if you create
   an Item and then change the length of the underlying array
   (e.g. call realloc or move the position of END).  Either avoid
   doing such things or rebuild your Item representations afterwards.

   Because seqs serve as the array primitive for interpreted code,
   they are likely to be used for destructive operations.  Therefore
   Item representations are guaranteed to contain a pointer to the
   actual input array, never a copy.   

   Notice that C string operations will not work properly on seqs, because
   they may contain internal zero bytes and the byte-wise representation
   of END is not all zeros.  */
   

/* Create an Item of type seq from the raw C array representation. */
Item seq2item(Item *data)
{
  Pair *cell;
  unsigned long length;

  cell = (Pair *)safe_malloc(sizeof(Pair));

  length = seqlen(data);
  (cell->cdr).uinteger = TAGNUMBER + TAGSEQ + (length << SUBTYPESHIFT);
  (cell->car).ptr = data;
  return pair2item(cell);
}

/* Set the value at this position in the seq */
void seqset(Item seq, long position, Item value)
{
  long length;
  Item *rawseq;

  assert_input_type(seq, TYPESEQ);
  length = seqlength(seq);
  if (position >= length)
    error("Position %ld is past the end (position %ld) of seq %s\n",
	  position, length-1, printitem(seq));
  if (position < 0)
    error("Attempt to access negative position %ld in seq %s\n",
	  position, printitem(seq));
  rawseq = item2seq(seq);
  rawseq[position] = value;
}

/* Get the value at this position in the seq */
Item seqget(Item seq, long position)
{
  long length;
  Item *rawseq;

  assert_input_type(seq, TYPESEQ);
  length = seqlength(seq);
  if (position >= length)
    error("Position %ld is past the end (position %ld) of seq %s\n",
	  position, length-1, printitem(seq));
  if (position < 0)
    error("Attempt to access negative position %ld in seq %s\n",
	  position, printitem(seq));
  rawseq = item2seq(seq);
  return rawseq[position];
}

/* Fill all positions in seq with the specified value */
void seqfill(Item seq, Item value)
{
  long length, i;
  Item *rawseq;

  assert_input_type(seq, TYPESEQ);
  length = seqlength(seq);
  rawseq = item2seq(seq);

  for (i = 0; i < length; i++)
    rawseq[i] = value;
}


/* Creates a newly allocated seq of the given length, filled with the
   value MISSING and properly terminated by END.  Whether it's more
   convenient to call this function or do the work yourself depends on
   your application. */
Item new_seq(long length)
{
  Item *rawarray;

  rawarray = (Item *)safe_malloc((length+1)*sizeof(Item));
  rawarray[length] = END;
  return seq2item(rawarray);
}

/* Returns the length of the seq stored in this Item. 
   This function differs from seqlen (below) in that its input is
   an item (not a C array) and it returns the length stored in the
   Item rather than dynamically computing length from the array 
   contents. */
unsigned long seqlength(Item input)
{
  Pair *cell;
  unsigned int infofield;
  assert_input_type(input,TYPESEQ);
  cell = item2pair(input);
  infofield = (cell->cdr).uinteger;
  return (infofield >> SUBTYPESHIFT);
}


/* The return value will be an array of Items, terminated by END. */
Item *item2seq(Item input)
{
  Pair *cell;
  assert_input_type(input,TYPESEQ);
  cell = item2pair(input);
  return (cell->car).ptr;
}


/* Make a list out of a seq. */
Item seq2list(Item *input)
{
  Item *ptr;
  Item output = NIL;

  ptr = input;

  while (!at_end(*ptr)){
    output = cons(*ptr,output);
    ptr++;
  }
  return nreverse(output);
}

/* Make a seq out of a list. */
Item *list2seq(Item input)
{
  int len;
  Item *data, *ptr;

  assert_input_type(input,TYPELIST);

  len = length(input);
  data = safe_malloc(sizeof(Item)*(len+1));
  ptr = data;
  while(!null(input)){
    *ptr = car(input);
    ptr++;
    input = cdr(input);
  }
  *ptr = END;
  return data;
}

/* The length of a sequence, not including the terminating END.
   Analogous to strlen. */
long seqlen(Item *seq)
{
  long output = 0;
  while (!at_end(*seq)){
    output++;
    seq++;
  }
  return output;
}

/* Make a dynamically-allocated copy of the input seq */
Item *copyseq(Item *input)
{
  Item *output;
  long length, i;

  length = seqlen(input);
  output = (Item *)safe_malloc(sizeof(Item)*(length+1));

  for (i = 0; i <= length; i++)
    output[i] = input[i];

  return output;
}

/* Returns the empty sequence. */
static Item *emptyseq(void)
{
  Item  *output;

  output = safe_malloc(sizeof(Item));
  *output = END;
  return output;
}


/* Like substring.   Returns a copy of the subsequence between the two 
   input positions, including both endpoints.   Returns the empty sequence
   if start < end. */
Item *subseq (Item *start, Item *end)
{
  Item *output;
  unsigned long length, i;

  if (end < start)
    return emptyseq();

  length = end - start + 1;
  output = safe_malloc(sizeof(Item)*length+1);
  for (i = 0; i < length; i++)
    output[i] = *(start+i);
  output[length] = END;
  return output;
}

/* Two seq's are similar if their corresponding elements are eq. 
   This is intended for seq's containing symbols, integers, and ngrams. 
   This is faster than testing the sign of seqcmp. */
int seqsimilar (Item *one, Item *two)
{
  int i;
  i= 0;
  while (!at_end(one[i])){
    if (!eq(one[i],two[i]))
      return 0;
    i++;
  }
  if (!at_end(two[i]))
    return 0;
  else
    return 1;
}

/* ================================================================ */
/* Sorting a seq                                                    */
/* ================================================================ */

/* These functions define a human-friendly order for Items, useful
   (for example) when printing out long tables for human inspection.

   Sorting order is first by type:
      integers, floats, 
      symbols, strings, 
      ngrams, seqs, lists, 
      bundles, functions, closures,
      everything else

   then, within each type:
       -- numerical order for integers and floats
       -- strcmp order for symbols and strings,
       -- item by item for ngrams
       -- internal storage order for lists and everything else

   This isn't a great sorting order for lists, but anything smarter presents
   the danger of looping on circular structures.

   This is intended as a sensible default order.   And, in particular, it
   is intended for sorting the symbols and ngrams common in natural language
   applications.  Feel free to define your own order if you need something 
   different.
*/


/* Convert internal type ID's into an order meaningful to humans
   and constant across implementations */
static int typeorder(int internaltype)
{
  switch (internaltype){
  case TYPEINT: 
    return 1;
    break;
  case TYPEFLOAT:
    return 2;
    break;
  case TYPESYMBOL:
    return 3;
    break;
  case TYPESTRING:
    return 4;
    break;
  case TYPENGRAM:
    return 5;
    break;
  case TYPESEQ:
    return 6;
    break;
  case TYPELIST:
    return 7;
    break;
  case TYPEBUNDLE:
    return 8;
    break;
  case TYPEFUNCTION:
    return 9;
    break;
  case TYPECLOSURE:
    return 10;
    break;  
  default:
    return 100;
    break;
  }
}

/* Compare two integers. */
static int intcmp(long one, long two)
{
  if (one < two)
    return -1;
  else if (two < one)
    return 1;
  else
    return 0;
}

int itemcmp(const Item one, const Item two);

int seqcmp(Item *one, Item *two)
{
  int rv;

  while (!at_end(*one)){
    rv = itemcmp(*one, *two);
    if (rv != 0)
      return rv;
    one++;
    two++;
  }

  if (!at_end(*two))
    return 1;

  return 0;
}


/* Which input Item goes first in alphabetical order? */
int itemcmp(const Item one, const Item two)
{
  int t1, t2, type1, type2;
  float fnum1, fnum2;

  type1 = item2type(one);
  type2 = item2type(two);
  t1 = typeorder(type1);
  t2 = typeorder(type2);
  
  if (t1 > t2)
    return 1;
  else if (t1 < t2)
    return -1;
  else {
    switch(type1){
    case TYPEINT:
      return intcmp(item2long(one), item2long(two));
      break;
    case TYPEFLOAT:
      fnum1 = item2float(one);
      fnum2 = item2float(two);
      if (fnum1 < fnum2)
	return -1;
      else if (fnum1 > fnum2)
	return 1;
      else return 0;
      break;
    case TYPESTRING:
      return strcmp(item2string(one),item2string(two));
      break;
    case TYPESYMBOL:
      if (special_symbol(one)){
	if (!special_symbol(two))
	  return -1;
	else
	  return 0;
      }
      if (special_symbol(two))
	return 1;
      return strcmp(symbol2string(one),symbol2string(two));
    case TYPENGRAM:
      return seqcmp(ngram2seq(one), ngram2seq(two));
    case TYPESEQ:
      return seqcmp(item2seq(one), item2seq(two));
    default:
      return intcmp(one.uinteger, two.uinteger);
      break;
    }
  }
}

/* Version of itemcmp with type suitable for handing
   to qsort. */
static int itemcmpx(const void *one, const void *two)
{
  Item item1, item2;
  Item *p1, *p2;
  
  p1 = (Item *)(one);
  p2 = (Item *)(two);
  item1 = *p1;
  item2 = *p2;

  return itemcmp(item1,item2);
}


/* Sorts seq in place, into a humanfriendly order.   See details
   of order above.   This function doesn't allow you to supply
   a different sorting function because you can easily write
   your own variations using qsort. */
Item *sortseq(Item *input)
{
  long length;

  length = seqlen(input);

  /* sort the keys */
  qsort(input,length,sizeof(Item),itemcmpx);
  return input;
}


/* ================================================================ */
/* Special complex objects                                          */
/* ================================================================ */

/* Special objects are an escape mechanism that allows the programmer to
   embed an arbitrary C pointer into one of our data structures.
   They are also used to implement types that do not require significant
   support from the core interpreter or reader (e.g. hash tables).

   Special objects contain a generic C pointer and a symbolic label which 
   (hopefully) means something to the programmer, e.g. tells him what
   type the pointer should be coerced to.   */

/* Notice that the interpreter uses the labels 
     FILE_HANDLE 
     HASHTABLE
     EARRAY
   for objects that your code is likely to encounter.  Therefore, you shouldn't 
   use these labels for your own purposes.  The labels FUNCTION and CLOSURE
   should also be avoided. */

/* Checks whether the input item is type SPECIAL and has the
   specified label. */
int is_special_type(Item object, Item label)
{
  if (item2type(object) != TYPESPECIAL)
    return 0;
  if (!eq(item2label(object),label))
    return 0;
  return 1;
}

/* Creates a "special" object.  */
Item ptr2item (void *ptr,Item label) 
{
  Pair *cell;

  assert_input_type(label,TYPESYMBOL);

  cell = (Pair *)safe_malloc(sizeof(Pair));
  (cell->cdr).uinteger = ((label.uinteger) & (~TAG)) | TAGSTRING;
  (cell->car).ptr = ptr;
  return pair2item(cell);
}

/* Reset the pointer in a special object.  Use with care. */
void setspecialptr(Item special, void *newptr)
{
  Pair *cell;

  assert_input_type(special, TYPESPECIAL);
  cell = item2pair(special);
  (cell->car).ptr = newptr;
}

/* Return label of a special object.  */
Item item2label(Item input)
{
  Pair *cell;
  Item output;
  unsigned int infofield;

  assert_input_type(input,TYPESPECIAL);

  cell = item2pair(input);
  infofield = (cell->cdr).uinteger;
  output.uinteger = ((infofield & (~TAG)) | TAGSYMBOL);
  return output;
}

/* Return the raw C pointer inside input Item.
   Item is first checked to make sure its type label agrees with
   the input label you have supplied.  
   This irritating little feature exists so that you can't forget
   to check the type.
   Caller will need to coerce output to the appropriate C type. */
void *item2ptr(Item input, Item label)
{
  Pair *cell;
  assert_special_input(input, label);
  cell = item2pair(input);
  return (cell->car).ptr;
}


/* ================================================================ */
/* Accessors for basic items                                        */
/* ================================================================ */


/* For efficiency, the output item often includes the original
   input string.  However, copying must occur if the input is not
   4-byte aligned.   The details may change in future (e.g. when
   schwa is moved to 64-bit).   Therefore, don't write code 
   that depends on whether copying does or doesn't occur.   

   If you need to embed strings into items and then do destructive
   operations on them, make them special objects rather than strings. 

   Do not pass stack-allocated strings directly into this function.
   Use copystring first.  (String constants are ok because they
   live on the heap.) */
Item string2item (char *string)
{
  Item output;
  unsigned long tagged;

  tagged = (unsigned long) string;
  if ((tagged&TAG) != 0){         /* input not 4-byte aligned */
    string = copystring(string);
    tagged = (unsigned long) string;
  }
  tagged = tagged | TAGSTRING;

  output.uinteger = tagged;
  return output;
}

Symboltable *symbols = 0; 

/* Is this string interned yet? */
int stringinterned (char *string)
{
  long code;
  if (!symbols) return 0;
  code = intern_string(string, symbols, 0);
  if (code < 0) return 0;
  else return 1;
}

Item string2symbol (char *string)
{
  Item output;
  unsigned long code;
  if (!symbols)
    symbols = new_symboltable(FIRSTCODE);
  code = intern_string(string, symbols, 1);
  code = (code << SYMSHIFT) | TAGSYMBOL;
  output.uinteger = code;
  return output;
}

/* Assumes you've first checked the type of input Item. */
char *symbol2string (Item symbol)
{
  unsigned long code;
  assert_input_type(symbol, TYPESYMBOL);
  if (special_symbol(symbol))
    error("Special symbol %s doesn't have an associated string\n", 
	  printitem(symbol));
  if (!symbols)   
    symbols = new_symboltable(FIRSTCODE);
  code = symbol.uinteger;
  code = code >> SYMSHIFT;
  return code2string(code, symbols);
}

/* Converts the input string into a number.  */
Item string2number(char *input)
{
  float fv;
  long iv;
  int rv;

  if (!(isdigit(*input) || *input == '.' || *input == '-'))
    error("string2number:  %s is not the printed representation of a number\n", input);
  
  if (strchr(input,'.')){
    rv = sscanf(input,"%f", &fv);
    if (rv != 1)
      error("string2number: angled float %s\n", input);
    return float2item(fv);
  }
  else {
    rv = sscanf(input,"%ld", &iv);
    if (rv != 1)
      error("string2number: mangled integer %s\n", input);
    return long2item(iv);
  }
}


Item long2item (long integer)
{
  Item output;
  long tagged;
  if (labs(integer) > MAXINTEGER)
      error("long2item:  input magnitude too large\n"); 
  tagged = (integer*INTFACTOR + TAGNUMBER);
  output.sinteger = tagged;
  return output;
}

/* Assumes you've first checked the type of input Item. */
char *item2string (Item input)
{
  unsigned long tagged;

  assert_input_type(input,TYPESTRING);
  
  tagged = input.uinteger;
  tagged = tagged & ~TAG;
  return (char *) tagged;
}

/* Assumes you've first checked the type of input Item. */
long item2long (Item input)
{
  signed long tagged;
  
  assert_input_type(input,TYPEINT);
  tagged = input.sinteger;
  tagged = (tagged - TAGNUMBER)/INTFACTOR; 
  return tagged;
}

/* ================================================================ */
/* Dangerous functions for direct manipulation of code values       */
/* ================================================================ */

/* These functions provide limited direct access to symbol 
   codes.  They are for use in implementing ngram tables and 
   certain types of efficient data tables, e.g. storing auxiliary 
   information for most members of a large symbol set.   
   Use with caution.  */

/* Add symbol tag to 29-bit symbol code. */
Item code2symbol (unsigned long code)
{
  Item output;
  if (!symbols || code >= symbols->maxelement)
    error("code2symbol give bogus symbol code %ld\n", code);
  code = (code << SYMSHIFT) | TAGSYMBOL;
  output.uinteger = code;
  return output;
}

/* Returns a list containing the minimum and maximum integer codes
   currently in use for symbols.  All codes in this range (inclusive)
   should correspond to valid symbols.  The special symbols NIL,
   MISSING, and END are not included. */
Item symbolrange (void)
{
  unsigned long max;
  if (!symbols)
    return NIL;
  max = (symbols->maxelement)-1;
  return 
    list(2,long2item(symbols->firstcode), long2item(max));
}

/* Returns the code number associated with the symbol.  */
unsigned long symbol2code (Item symbol)
{
  unsigned long code;

  assert_input_type(symbol,TYPESYMBOL);
  code = symbol.uinteger;
  code = code >> SYMSHIFT;
  return code;
}

/* ================================================================ */
/* Support for ngrams                                               */
/* ================================================================ */

/* Ngrams are interned versions of lists (just as symbols are interned strings).
   For efficiency reasons, ngrams can contain only symbols.  If memory
   becomes less tight, this could be relaxed to also include symbols
   and integers.  However, storing pointer-type objects inside ngrams won't 
   do anything useful and will hide pointers from the garbage collector.

   Ngrams can be used to represent word n-grams.  They can also be
   used to represent "words" formed out of letters, each of which
   is a string.   For example, the letters might come from a 
   pronunciation alphabet (e.g. "ey") or they might be UTF-8 encodings
   of non-ASCII characters. 

   Ngrams are stored internally in packed form.  That is, we economize
   space rather than computation time.  The goal is to allow n-gram
   language models of reasonable size to be built using ngrams.
*/

Ngramtable *ngrams = 0; 

/* Display distribution of ngrams in the main bins ("houses") of the
   ngram table.   Temporary for debugging and testing. */
Item show_ngram_occ (void)
{
  show_ngram_occupancy(ngrams);
  return MISSING;
}

Item ngram2list (Item ngram)
{
  unsigned long code;

  assert_input_type(ngram,TYPENGRAM);

  if (!ngrams)
    ngrams = new_ngramtable();
  code = ngram.uinteger;
  code = code >> SYMSHIFT;
  return code2list(code, ngrams);
}

/* Is this list interned as an ngram yet? */
int listinterned (Item list)
{
  long code;

  assert_input_type(list,TYPELIST);

  if (!ngrams) return 0;

  code = intern_list(list, ngrams, 1);
  if (code < 0) return 0;
  else return 1;
}

Item list2ngram (Item list)
{
  Item output;
  unsigned long code;

  assert_input_type(list,TYPELIST);

  if (!ngrams)
    ngrams = new_ngramtable();

  code = intern_list(list, ngrams, 1);
  code = (code << SYMSHIFT) | TAGSYMBOL | TAGNGRAM;
  output.uinteger = code;
  return output;
}

/* Make a ngram out of the input items.
   First argument is the total number of items. */
Item ngram(unsigned int n, ...)
{
  int i;
  Item output = NIL;
  Item next;
  va_list ap;

  va_start(ap, n);
  for (i = 0; i < n; i++){
    next = va_arg(ap, Item);
    output = cons(next, output);
  }
  va_end(ap);

  output = reverse(output);
  return list2ngram(output);
}


/* Input should be an array of Items, terminated by END. */
Item seq2ngram(Item *seq)
{
  return list2ngram(seq2list(seq));
}

/* Output should be an array of Items, terminated by END. */
Item *ngram2seq (Item ngram)
{
  return(list2seq(ngram2list(ngram)));
}

/* Returns the code number associated with the ngram.  */
unsigned long ngram2code (Item ngram)
{
  unsigned long code;

  assert_input_type(ngram,TYPENGRAM);
  code = ngram.uinteger;
  code = code >> SYMSHIFT;
  return code;
}

/* Returns the code number associated with the ngram.  */
Item code2ngram (unsigned long code)
{
  Item output;

  code = (code << SYMSHIFT) | TAGSYMBOL | TAGNGRAM;
  output.uinteger = code;
  return output;
}

