/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/

/* Manipulating lists that contain a diversity of elements */

#ifndef schwalists
#define schwalists

#include "utils.h"
#include "symbol.h"


/* largest signed integer that will fit in 30 bits */
#define MAXINTEGER (256*256*256*32 - 1)    


/* Type tags.  For internal use only. 
   To prevent any non-MISSING object from being == 0, the tag 0 is 
   used for symbols and the symbol code 0 is kept reserved.  */
#define TAGSYMBOL 0          
#define TAGNUMBER 1   
#define TAGSTRING 2  
#define TAGCOMPLEX 3   
#define TAG 3         /* The two bits used in our tags, i.e. to be 
			 used for clearing these tags. */

#define INTFACTOR 4    /* Multiplicative factor corresponding to 
			  adding/removing 2-bit tags from integers. */

#define SYMTAG 7      /* Symbols and ngrams are distinguished via a
			 third bit (the 4 bit), which is set iff 
			 it's a ngram. */
#define TAGNGRAM 4       /* The extra bit that is set for ngrams.
			    So the full tag for an ngram is
			    TAGSYMBOL + TAGNGRAM */
#define SYMSHIFT 3       /* 3-bit shift used to add/strip symbol and 
			    ngram tags. */

/* Longer type tags, used to store the subtype of miscellanous
   complex objects. */
#define LONGTAG 63       /* The bits used for the misc complex object tags. */
#define SUBTYPESHIFT 6   /* The number of bits required to represent subtype 
			    for complex objects. */

/* The following tags must not use the first two (1 and 2) bits. */
#define TAGFLOAT       0
#define TAGFUNCTION   4
#define TAGSEQ   8
#define TAGCLOSURE 12

/* Types
   These are the values you should test against when examining the output 
   of item2type. I assume the compiler is bright enough to optimize away 
   the arithmetic with constants.*/
#define TYPEINT TAGNUMBER
#define TYPESTRING TAGSTRING
#define TYPESYMBOL TAGSYMBOL 
#define TYPENGRAM TAGSYMBOL + TAGNGRAM
#define TYPELIST   ((TAGCOMPLEX << SUBTYPESHIFT) + TAGCOMPLEX)        
#define TYPEFLOAT  (((TAGNUMBER + TAGFLOAT) << SUBTYPESHIFT) + TAGCOMPLEX)
#define TYPECLOSURE  (((TAGNUMBER + TAGCLOSURE) << SUBTYPESHIFT) + TAGCOMPLEX)
#define TYPEFUNCTION  (((TAGNUMBER + TAGFUNCTION) << SUBTYPESHIFT) + TAGCOMPLEX)
#define TYPESEQ  (((TAGNUMBER + TAGSEQ) << SUBTYPESHIFT) + TAGCOMPLEX)
#define TYPEBUNDLE  ((TAGSYMBOL << SUBTYPESHIFT) + TAGCOMPLEX)
#define TYPESPECIAL  ((TAGSTRING << SUBTYPESHIFT) + TAGCOMPLEX)

union item {
  unsigned long uinteger;   /* used for storing most values, listed first 
                               because used for initialization.*/
  long sinteger;            /* for storing actual integer values */
  float real;               /* for storing floating point values */
  void *ptr;               /*  for storing pointer values.   Also to force
                               proper alignment of pointers. */
};
typedef union item Item;


/* A pair of items.  This may represent a cons cell, a float, or a
   feature bundle.  The type tag of the cdr tells you which of the
   three it is.  As a consequence, dotted pairs are not supported.*/
struct pair {
  Item cdr;  
  Item car;
};
typedef struct pair Pair;

struct kvlist {
  Item key;
  Item value;
  struct kvlist *next;
};

typedef struct kvlist Kvlist;

/* This is the internal representation of hash tables.  This may
   change in the future.  Please use the iterator functions hashstart 
   and hashnext, rather than accessing this internal representation directly. 
   Exception:  read access to numoccupants is ok.  */
struct hashtable {
  Kvlist ***table;
  unsigned long tablesize;   
  unsigned long numoccupants;
  unsigned long bits;        /* tablesize = 2^bits */
};

typedef struct hashtable Hashtable;

/* A pointer to a position in a hash table, 
   used for iterating through the table. */
struct hashptr {
  Kvlist ***table;        /* The raw table of kvlists */
  unsigned long tablesize; /* Number of bins in table. */
  unsigned long bin;       /* Number of bin we are looking at */
  Kvlist *remaining;     /* Remaining unexamined contents of this bin. */
};

typedef struct hashptr Hashptr;

/* void print_occupancy(Hashtable *table, int groupsize); */


/* A closure, i.e. the result of evaluating a lambda expression. */
struct closure {
  Item formals;      /* The input variables, a list of symbols. */
  Item body;         /* A list of forms */
  Item bindings;  /* The definition environment (local bindings only), 
			a list of binding pairs. */
};

#include "ngram.h"
Item show_ngram_occ (void);


typedef struct closure Closure;

Item string2item (char *string);
Item long2item (long integer);
long item2long (Item input);
Item seq2item(Item *data);
Item *item2seq(Item input);
unsigned long seqlength(Item input);
void seqset(Item seq, long position, Item value);
Item seqget(Item seq, long position);
Item new_seq(long length);
void seqfill(Item seq, Item value);

char *item2string (Item input);
Item string2symbol (char *string);
int stringinterned (char *string);
char *symbol2string (Item symbol);
unsigned long symbol2code (Item symbol);
unsigned long ngram2code (Item symbol);
Item code2ngram (unsigned long code);
Item code2symbol (unsigned long code);
Item symbolrange (void);
Item float2item(float value);
float item2float(Item input);
Item string2number(char *input);

Item seq2list(Item *input);
Item *list2seq(Item input);
long seqlen(Item *seq);
Item *copyseq(Item *input);
Item *subseq (Item *start, Item *end);
Item *sortseq(Item *input);
int itemcmp(const Item one, const Item two);
int seqcmp(Item *one, Item *two);
int seqsimilar (Item *one, Item *two);


Item closure2item(Closure *table);
Closure *item2closure(Item input);

unsigned int item2type (Item input);
char *printitem (Item input);
char *printitemfull (Item input);
char *type2string(unsigned int type);
char *format(Item args);
char *format_infull(Item args);

Item function2item(void *function, unsigned int arity);
unsigned int item2arity(Item input);
void *item2ptr(Item input, Item label);
void *item2function(Item input);
Item ptr2item (void *ptr,Item label) ;
Item item2label(Item input);
void setspecialptr(Item special, void *newptr);
int is_special_type(Item object, Item label);


Item read_sexp (FILE *infile);
Item read_file (char *filename);

Item cons (Item car, Item cdr);
Item car (Item list);
Item cdr (Item list);
Item cadr (Item list);
Item reverse (Item list);
Item nreverse (Item list);
Item assq(Item object, Item list);
int eq(Item one, Item two);
int equal(Item one, Item two);

Item list(unsigned int n, ...);
Item copylist (Item list);
void set_cdr (Item list, Item value);
void set_car (Item list, Item value);
int memq(Item object, Item list);
Item delq(Item object, Item list);
int member(Item object, Item list);
int present(Item input);
int bundleisset(Item bundle, Item feature);
int null(Item input);
int at_end(Item input);
Item nconc (Item list1, Item list2);
Item append (Item list1, Item list2);
long length(Item list);
Item nth(long n, Item list);

Item new_bundle(Item head);
Item bundleset(Item bundle, Item feature, Item value);
Item bundleget(Item bundle, Item feature);
Item bundlehead(Item bundle);
Item bundlefeats(Item bundle);

Item ngram(unsigned int n, ...);
Item list2ngram (Item list);
int listinterned (Item list);
Item ngram2list (Item ngram);
Item seq2ngram(Item *seq);
Item *ngram2seq (Item ngram);

Hashtable *new_hashtable (void);
void hashset(Hashtable *table, Item key, Item value);
Item hashget(Hashtable *table, Item key);
Hashptr *hashstart(Hashtable *table);
Item hashnext(Hashptr *ptr);
Item *hashkeys(Hashtable *table);
void hashadd(Hashtable *table, Item key, Item value);
void hashincrement(Hashtable *table, Item key, int amount);

/* Macro used to check types of inputs to a function.
   Creates an error break if item (an Item) is not of the specified type 
   (an integer e.g. TYPESTRING). 
   This is a macro so that we have access to an informative function name. */
#define assert_input_type(myitem,mytype) \
if (item2type(myitem) != mytype) \
  error("input %s to function %s is not a %s\n", \
	printitem(myitem), __func__, type2string(mytype));

/* Similar, but error message is suitable for non-input intermediate
   values produced inside the function. */
#define assert_type(myitem,mytype) \
if (item2type(myitem) != mytype) \
  error("In function %s, this object %s is not a %s\n", \
	__func__, printitem(myitem), type2string(mytype));


/* Similar, but for objects of special type.
   Checks that object is indeed special type and that it has the specified label.  */
#define assert_special_input(myitem,label) \
  if (item2type(myitem) != TYPESPECIAL \
      || !eq(item2label(myitem), label)) \
  error("input %s to function %s is not a %s\n", \
	printitem(myitem), __func__, printitem(label));

/* Similar, but error message is suitable for non-input intermediate
   values produced inside the function. */
#define assert_special_type(myitem,label) \
  if (item2type(myitem) != TYPESPECIAL || !eq(item2label(myitem), label)) \
  error("In function %s, this object %s is not a %s\n", \
	__func__, printitem(myitem), printitem(label));

extern const Item NIL;
extern const Item MISSING; 
extern const Item ANY;
extern const Item END;



#endif

