/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/

/* Support for ngrams, which are interned lists of symbols. */

/* Margaret Fleck, 2005 */


/* The ngram facility is intended for several uses with very different
      requirements.  
      
      (1) language modelling, which typically requires 2-grams and
      3-grams over a large set of symbols (words, morphemes, chinese
      characters)

      (2) spellings and pronunciations (e.g. dictionaries), which
      typically contain longer (e.g. 6+ character) ngrams over a
      smaller (e.g. 30 word) vocabulary of symbols (e.g. characters,
      utf-8 characters, arpabet phoneme names)

      (3) type information for feature bundles, i.e. the head and list
      of feature names for the bundle.  These tend to be moderately
      long, over a diverse vocabulary of symbols, but we expect only a
      small number of distinct bundle types in any given application

   Utilities supporting ngrams (e.g. in lists.c) use a single global
   ngram table.  However, you can also invoke these functions to
   create (and destroy) additional ngram tables for your own nefarious
   purposes.

   Ngrams are stored in an ngram table which divided into houses.
   Each house contains a packed list of ngrams.  The ngrams in a house
   share a common 18-bit ID, computed from the symbol ID's in the
   ngram using a process of XOR and bit rotation.  This is randomized enough
   to be a moderately good hash code, but reversible enough that we
   can use the house ID to reduce the storage required for each ngram.

   The number of houses has to remain fixed as ngrams are added to the
   table, due to constraints of the encoding algorithm.  The number of
   houses (65536) is large enough to allow a wide space of symbol
   codes (21-bit) but small enough to guarantee even distribution of
   ngrams across houses.  To prevent excessive overhead for small sets
   of ngrams, houses are accessed via a hash table rather than by
   allocating a 65536-element array of houses.

   We use schwa's general-purpose Item-based hash table for ease of
   debugging, despite a small loss of efficiency.

   Each house contains an array of unsigned 32-bit integers.  Most of
   these contain packed ngram information:

        most significant 5 bits of car  +   ngram ID code of cdr (27 bits)

   By combining these 5 bits of the car with the 16-bit house ID, we obtain
   the full 21-bit symbol ID of the car.

   Position 0 in house 0 is reserved for the null ngram.

   The code for an ngram has the form
       position in house (13 bits) +  house ID (16 bits)
   Thus, an Item representing an ngram has the form
       position in house (13 bits) +  house ID (16 bits)  +  type tag (3 bits)

   Because ngrams are allocated reasonably uniformly across houses and because
   the house ID is in the least significant bits, the allocated ngram codes end
   up packed into mostly consecutive numbers starting at zero, once enough
   ngrams have been interned. 

   Each house also contains a sorted array of (2-byte) pointers to its entries.
   This allows entries to be found via binary search, important when entry lists 
   get long. 
*/

#include "lists.h"

/* ================================================================ */
/* Global definitions                                               */
/* ================================================================ */

/* A house is a packed array of ngrams, sharing a common 18-bit ID. */

/* Notice that the population and size fields are only short (2-byte) integers,
   which depends on the fact that each house contains at most 2^13 ngrams. */
struct house {
  unsigned long *array;        /* Array of ngrams. */
  unsigned short population;   /* How many ngrams are stored in array.  
				  This is also the position of the 
				  first unoccupied cell in array. */
  unsigned short *sorting;     /* Sorted list of positions in ngram array.  Same
				  length as array.  */
  unsigned short size;         /* Allocated size of array. */
};

typedef struct house House;


#define HOUSEBITS 16         /* Number of bits in house ID. */
#define MAXHOUSE (1024*64-1) /* Largest house, also mask for extracting house ID bits. */ 
#define MAXHOUSESIZE 1024*8     /* Limit on size of house arrays. */


/* ================================================================ */
/* Global constants                                                 */
/* ================================================================ */

#define STARTSIZE 2       /* Initial number of ngrams stored in each
			     house.  Code probably depends on this being >= 2. */

#define IDSHIFT 7             /* Amount of rotation in ID computation. */
#define MAXCAR (1024*2048 - 1)          /* Largest symbol code we can use in a ngram. */
#define MAXCDR (1024*1024*128 - 1)   /* Largest ngram code usable as cdr of new ngram. 
					Also the bit mask for retrieving cdr code. */
#define CDRBITS   27                 /* Length of code stored for cdr of ngram. */
#define EXPANDFACTOR 1.5             /* How fast to expand array of ngrams inside each house
					when it overflows.   Set relatively low to limit
					amount of space overhead. */


/* ================================================================ */
/* Core functions                                                   */
/* ================================================================ */


/* Combine the least significant HOUSEBITS bits of the two inputs, using XOR.
   The second input is first rotated by HOUSEBITS bits.   Notice that this
   operation is its own inverse.   That is, if 
      output = combine_rotated(one,two)
   then
      one = combine_rotated(output,two)
   This invertability is critical to the success of this storage method.
*/
static unsigned long combine_rotated(unsigned long one, unsigned long two)
{
  unsigned long topbits; 

  one = one&MAXHOUSE;
  two = two&MAXHOUSE;

  /* Rotate second input by IDSHIFT bits. */
  topbits = two >> (HOUSEBITS - IDSHIFT);  
  two = two << IDSHIFT;
  two = two&MAXHOUSE;
  two = two | topbits;

  /* XOR in first input value. */
  return (two ^ one);
}

/* Sets up and returns a new empty house. */
static House *new_house(void)
{
  House *output;

  output = safe_malloc(sizeof(House));
  output->population = 0;
  output->size = STARTSIZE;
  output->array = safe_malloc_atomic(STARTSIZE*sizeof(unsigned long));
  output->sorting = safe_malloc_atomic(STARTSIZE*sizeof(unsigned short));
  return output;
}


/* Find position (in sorting array) where the location of newentry (in entries array)
   has been stored.  
   If entry isn't present, returns an occupied storage location containing a value
   just before or just after the entry.  (Caller must test which.)
   Searches the positions between start and end, inclusive.
   Do not input an empty search interval. 
   Assumes input arrays contain no duplicate values. */
static long binary_search (unsigned long newentry, unsigned short *sorting, 
			   unsigned long *entries,
			   unsigned long start, unsigned long end)
{
  unsigned long middle, middleval;

  if (end < start){
    error("Null interval input to binary_search:  %d to %d\n", start, end);
    return 0;
  }
  if (end == start) 
    return start;
  middle = (start+end)/2;
  middleval = entries[sorting[middle]];
  if (newentry == middleval)
    return middle;
  else if (newentry < middleval){
    if (middle == start)
      return start;
    else
      return binary_search(newentry, sorting, entries, start, middle - 1);
  }
  else
    return binary_search(newentry, sorting, entries, middle + 1, end);
}


/* Find entry (carcode, cdrcode) in the house's ngram table.   
   Returns position in house array (always non-negative).
   If no matching entry is present, it either
      a) if autoadd> 0, adds the ngram and returns the new position
      b) otherwise, returns -1.  
*/
static long find_entry(unsigned long houseID, 
			       unsigned long carcode, unsigned long cdrcode, 
			       Ngramtable *table, int autoadd)
{
  House *house;
  Item houseitem;
  unsigned long newsize, newentry, topbits, start;
  long position, i, population;
  unsigned long *entries, oldvalue;
  unsigned short *mysort;

  if (houseID > MAXHOUSE)
    error("Putative house ID %ld greater than maximum bound\n", houseID);

  houseitem = hashget(table->houses,long2item(houseID));

  /* If house doesn't exist at all, create it. */
  if (!present(houseitem)){
    if (!autoadd) return -1;
    house = new_house();
    hashset(table->houses,long2item(houseID), 
	    ptr2item(house,string2symbol("HOUSE")));
  }
  else
    house = item2ptr(houseitem, string2symbol("HOUSE"));

  entries = house->array;
  mysort = house->sorting;

  /* Compute packed representation of our ngram. */
  topbits = carcode >> HOUSEBITS;
  newentry = (topbits << CDRBITS) + cdrcode;

  /* Get first real entry in the house.   Notice that position 0 in 
     house 0 is reserved for the empty list. */
  if (houseID == 0)
    start = 1;
  else
    start = 0;

  /* Check if this list is already in the house array. */
  population = house->population;
  if (population == start) 
    /* E.g. both zero.   Or both 1 in the case of house 0. */
    position = start;
  else {
    position = binary_search(newentry, mysort, entries, start, population-1);
    /*  Check location found by binary search to see if it contains our entry and, if not,
	whether its entry should precede or follow our new entry. */
    oldvalue = entries[mysort[position]];
    if (oldvalue == newentry)
      return mysort[position];
    else if (oldvalue < newentry)
      position++;
  }
  
  /* Ok. It's not present.  So we need to add entry to the house array. */
  if (!autoadd) return -1;

  /* The new entry will be put in entries[population].  
     We will also update the sorted list by setting mysort[position] = population.
     If position < population, we will first need to shift existing values rightwards. */

  /* If house is too small, enlarge it. */
  if (house->size == population){
    if (population >= MAXHOUSESIZE)
      error("Overflow in array for house %ld\n", houseID);
    newsize = EXPANDFACTOR*(house->size);
    if (newsize > MAXHOUSESIZE)
      newsize = MAXHOUSESIZE;
    entries = safe_realloc(entries,newsize*sizeof(unsigned long));
    mysort = safe_realloc(mysort,newsize*sizeof(unsigned short));
    house->array = entries;
    house->sorting = mysort;
    house->size = newsize;
  }

  /* Add new entry to the data array */
  house->population += 1;
  entries[population] = newentry;
  /* Add entry into sorted list. */
  if (position < population){
    for (i = population-1; i >= position; i--)
      mysort[i+1] = mysort[i];
    }
  mysort[position] = population;
  return population;
}


/* Returns 27-bit code number associated with this ngram. 
   If ngram has not yet been interned, it either
      a) if autoadd> 0, adds the ngram and returns the new position
      b) otherwise, returns -1.  
*/
long intern_list (Item list, Ngramtable *table, int autoadd)
{
  unsigned long cdrcode, houseid, carcode, position;
  long rv;

  assert_input_type(list,TYPELIST);

  if (null(list))
    return 0;

  /* Get code for cdr of list. */
  rv = intern_list(cdr(list), table, autoadd);
  if (rv < 0) return -1;
  cdrcode = rv;

  if (cdrcode > MAXCDR)
    error("cdr of list has ngram code too high for forming new ngram:  %s\n", 
	  printitem(list));

  if (item2type(car(list)) != TYPESYMBOL)
    error("Attempt to put non-symbol into ngram:  %s\n", printitem(car(list)));
  carcode = symbol2code(car(list));
  if (carcode > MAXCAR)
    error("Symbol %s has too high an ID (%ld) to use in an ngram\n", 
	  printitem(car(list)),
	  carcode);

  houseid = combine_rotated(carcode, cdrcode);
  position = find_entry(houseid, carcode, cdrcode, table, autoadd);

  if (position < 0) return -1;
  return (position<<HOUSEBITS) + houseid;
}


/* Retrieves the list corresponding to a 27-bit ngram code.  */
Item code2list (unsigned long code, Ngramtable *table)
{
  unsigned long houseid, cdrcode, packed, position, carcode;
  Item cdr, houseitem;
  House *house;

  if (code == 0)
    return NIL;
  
  houseid = code&MAXHOUSE;
  houseitem = hashget(table->houses,long2item(houseid));
  if (!present(houseitem))
    error("Input to code2list (code %lo, house %lo) isn't an allocated ngram code\n", code, houseid);
  house = item2ptr(houseitem, string2symbol("HOUSE"));

  position = code>>HOUSEBITS;
  if (position >= house->population)
    error("Ngram code %lo has position %lo past the end of array for house %lo\n", code, position, houseid);
  packed = house->array[position];

  cdrcode = MAXCDR&packed;   
  carcode = packed>>CDRBITS;
  carcode = carcode << HOUSEBITS;

  /* Need to reconstruct low-order HOUSEBITS bits of carcode. */
  carcode = carcode | combine_rotated(houseid,cdrcode);

  cdr = code2list(cdrcode, table); 
  return cons(code2symbol(carcode), cdr);
}


/* Makes a new empty ngramtable. */
Ngramtable *new_ngramtable ()
{
  Ngramtable *output;
  House *zerohouse;

  output = safe_malloc(sizeof(Ngramtable));

  /* The table contains pointers to a fixed number of "houses". */
  output->houses = new_hashtable();

  /* Allocate first cell of first house, so we can reserve it for the null ngram. */
  zerohouse = new_house();
  zerohouse->population = 1;
  hashset(output->houses,long2item(0), ptr2item(zerohouse,string2symbol("HOUSE")));

  return output;
}

/* ================================================================ */
/* For debugging and testing only                                   */
/* ================================================================ */

/* Display distribution of ngrams in the main bins ("houses") of the
   ngram table. */
void show_ngram_occupancy (Ngramtable *table)
{
  unsigned long i, total = 0, mycount;
  Item houseitem, *keys, *ptr;
  House *house;
  Hashtable *counts;

  if (!table){
    fprintf(stdout,"No ngrams currently interned.\n");
    return;
  }

  counts = new_hashtable();

  for (i = 0; i <= MAXHOUSE; i++){
    houseitem = hashget(table->houses,long2item(i));
    if (present(houseitem)){
      house = item2ptr(houseitem,string2symbol("HOUSE"));
      total += house->population;
      hashincrement(counts, long2item(house->population), 1);
    }
    else 
      hashincrement(counts, long2item(0), 1);
  }
  fprintf(stdout, "Total population %ld\n", total);

  /* Print out histogram of occupancy numbers */
  total = 0;
  keys = hashkeys(counts);
  sortseq(keys);
  fprintf(stdout, "\nHistogram of occupancy numbers: \n occupancy  numhouses  cum-percent\n");
  ptr = keys;
  while (!at_end(*ptr)){
    mycount = item2long(hashget(counts, *ptr));
    total += mycount;
    fprintf(stdout, "    %ld          %ld       %4.2f\n", item2long(*ptr), mycount, (total*100.0)/MAXHOUSE);
    ptr++;
  }

}

