/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/

/* Code for printing s-expressions.
   Margaret M. Fleck, 2005  */

/* The print routine incorporates a very primitive indentation algorithm. 
   It's not wonderful, but it beats the hell out of having nothing.  */

/* The print routines in this file produce string output, so they can
   be used from within C code.  Actually sending this string to an I/O 
   stream is done elsewhere, e.g. in the interpreter code in eval.c. 

*/


#include "schwa.h"

/* ================================================================ */
/* Printing global variables                                        */
/* ================================================================ */

/* These variables control the appearance of pretty-printed output. 
   They are globals rather than part of the interpreter's environment,
   because they are frequently called from C code that does not have
   access to the environment.    

   Unless the "full details" options are specified, printer will only
   recurse PRINDEPTH levels into structure and will only print the
   first PRINLENGTH elements of a list.  This prevents an infinite
   loop on circular structures.

   Feel free to reset these variables. */
unsigned int PRINDEPTH = 5;
unsigned int PRINLENGTH = 10; 
unsigned int LINEWIDTH = 120;
unsigned int INDENT = 3;      /* amount to indent each level of nesting */ 

/* ================================================================ */
/* Printing helper functions                                        */
/* ================================================================ */

static char *printitemx (Item input, int depth, int indenton, int indentamt, int fulldetails);

/* Prints each item in input list into a separate string, returning a 
   list of strings. Assumes you've already check that input is really a 
   list. */
static Item printlist_helper(Item input, int depth, int indenton, int indentamt,
		      int *length, int *count, int *maxlen, int fulldetails)
{
  Item output = NIL, printeditem;
  int itemlength;
  char *onestring;
  *count = 0;
  *length = 0;
  *maxlen = 0;
  while (!null(input)){
    if (!fulldetails && *count >= PRINLENGTH){
      printeditem = string2item("...");
      output = cons(printeditem,output);
      *count += 1;
      *length += 3;
      break;
    }
    onestring = printitemx(car(input), depth+1, indenton, indentamt, fulldetails);
    printeditem = string2item(onestring);
    *count += 1;
    itemlength = strlen(onestring);
    *length += itemlength;
    if (itemlength > *maxlen)
      *maxlen = itemlength;
    output = cons(printeditem,output);
    input = cdr(input);
  }
  
  return nreverse(output);
}

/* Copy strings from input list one-by-one into the output string.  */
static char *copyinstrings(Item list)
{
  int total = 0;
  char *onestring, *ptr, *outstring;
  Item llist;
  
  /* Get total string length */
  llist = list;
  while (!null(llist)){
    total += strlen(item2string(car(llist)));
    llist = cdr(llist);
  }

  outstring = safe_malloc(total+1);

  ptr = outstring;
  while (!null(list)){
    onestring = item2string(car(list));
    list = cdr(list);
    strcpy(ptr, onestring);
    ptr = ptr + strlen(onestring);
    }

  return outstring;
}

/* Add a spacer to the front of each item. 
   Normally prefixes a space, but prefixes newline+spacer when
   linewidth is getting too long or if always is set.  
   First item only gets a spacer if startingindent is set. */
static Item prefix_spacer(Item printeditems, char *spacer, int startingindent, int always)
{
  int current, total= 0, firstitem = 1;
  Item output = NIL;
  char *instring, *outstring;

  if (startingindent)
    firstitem = 0;

  while (!null(printeditems)){
    instring = item2string(car(printeditems));
    current = strlen(instring);
    if (firstitem){
      firstitem = 0;
      outstring = instring;
    }
    else if (always || total+current+1 > LINEWIDTH){
      outstring = gcprintf("\n%s%s", spacer, instring);
      total = 0;
    }
    else 
      outstring = gcprintf(" %s", instring);
     total += strlen(outstring);
    output = cons(string2item(outstring), output);
    printeditems = cdr(printeditems);
    }
  return nreverse(output);
}

/* Make a spacer with length spaces */
static char *makespacer(int length)
{
  char *output;
  int x;

  output = safe_malloc_atomic(1+length);
  for (x=0; x < length; x++)
    output[x] = ' ';
  output[length] = 0;

  return output;
}

int DIVIDEFRACT = 3;
int MANYITEMS = 5;

/* Print interior contents of input list.  
   Assumes you've already checked that input is really a list. */
static char *printlist (Item input, int depth, int indenton, int indentamt, int fulldetails)
{
  Item printeditems;
  int count, maxlen;
  int length;  /* Sum of lengths of printed representations of items in list. */
  char *output;

  if (null(input))
    return copystring("");

  /* Try printing all the items without allowing internal indentation. */
  printeditems = printlist_helper(input, depth, 0, indentamt, &length, &count, &maxlen, fulldetails);

  if (!indenton || (length+count <= LINEWIDTH-indentamt))
    /* everything will fit on one line */
    printeditems = prefix_spacer(printeditems, makespacer(indentamt), 0, 0);    
  else if (count >= MANYITEMS && maxlen < (LINEWIDTH-indentamt)/DIVIDEFRACT){
    /* list contains many short items */
    indentamt += INDENT;
    printeditems = prefix_spacer(printeditems, makespacer(indentamt), 0, 0);
  }
  else {
    /* list contains some long items */
    indentamt += INDENT;
    printeditems = printlist_helper(input, depth, 1, indentamt, &length, &count, &maxlen, fulldetails);
    printeditems = prefix_spacer(printeditems, makespacer(indentamt), 0, 1);
  } 


  /* Copy the strings one-by-one into output string. */
  output = copyinstrings(printeditems);

  return output;
}


/* Prints each feature in input bundle into a separate string, returning a list of strings. 
   Assumes you've already check that input is really a bundle. */
static Item printfeatures_helper(Item bundle, int depth, int indenton, int indentamt,
			  int *length, int *count, int *maxlen, int fulldetails)
{
  Item features, feature, value;
  Item output = NIL, printeditem;
  char *onestring, *featstring;
  int itemlength;

  *count = 0;
  *length = 0;
  *maxlen = 0;

  features = bundlefeats(bundle);
  while (!null(features)){
    if (!fulldetails && *count >= PRINLENGTH){
      printeditem = string2item("...");
      output = cons(printeditem,output);
      *count += 1;
      *length += 3;
      break;
    }
    feature = car(features);
    value = bundleget(bundle,feature);
    if (present(value)){
      featstring = printitemx(feature, depth+1, indenton, indentamt, fulldetails);
      if (eq(value, ANY))
	onestring = gcprintf("%s", featstring);
      else 
	onestring = gcprintf("%s:%s", featstring,
			     printitemx(value, depth+1, indenton, 
					indentamt+INDENT+strlen(featstring), fulldetails));
      printeditem = string2item(onestring);
      *count += 1;
      itemlength = strlen(onestring);
      *length += itemlength;
      if (itemlength > *maxlen)
	*maxlen = itemlength;
      output = cons(printeditem,output);
    }
    features = cdr(features);
  }
  return nreverse(output); 
}


/* If feature values are somewhat long, it's more readable to print them on separate lines even if
   they could in principle be packed onto one line. */
float FEATFACTOR = 0.3;    

/* Prints all the feature/value pairs in the bundle.
   Assumes you've checked that it's really a bundle. */
static char *printfeatures(Item bundle, int depth, int indenton, int indentamt, 
		    int fulldetails, int hadhead)
{
  Item printeditems;
  int count, maxlen;
  int length;  /* Sum of lengths of printed representations of items in list. */
  char *output;

  if (null(bundlefeats(bundle)))
    return copystring("");

  /* Try printing all the items without allowing internal indentation. */
  printeditems = printfeatures_helper(bundle, depth, 0, indentamt, &length, &count, &maxlen, fulldetails);

  if (!indenton || 
      ((length+count <= LINEWIDTH-indentamt)
       && (count == 1 || (length+count) <= (LINEWIDTH-indentamt)*FEATFACTOR)))
    /* Indenting is turned off, so we can't intend.
       Or it's better not to indent, either because it will all fit on one line
       or because we have a small number of long feat:value pairs which are more
       legible with line breaks. */
    printeditems = prefix_spacer(printeditems, makespacer(indentamt), 0, 0);    
  else if (count >= MANYITEMS && maxlen < (LINEWIDTH-indentamt)/DIVIDEFRACT){
    /* list contains many short items */
    indentamt += INDENT;
    printeditems = prefix_spacer(printeditems, makespacer(indentamt), 0, 0);
  }
  else {
    /* list contains some long items */
    indentamt += INDENT;
    printeditems = printfeatures_helper(bundle, depth, 1, indentamt, &length, &count, &maxlen, fulldetails);
    printeditems = prefix_spacer(printeditems, makespacer(indentamt), 0, 1);
  } 

  output = copyinstrings(printeditems);

  return output;
}

static char *specials = "\n\\\"}";

/* Protect special characters in string with backslashes, e.g. as part of converting
   the string to printable form. */
static char *slashify (char *input) 
{
  char *outstring, *inpos, *outpos;
  int numspecial = 0;

  /* count the special characters that will need to be slashified */
  inpos = input;
  while (*inpos){
    if (strchr(specials, *inpos))
      numspecial++;
    inpos++;
  }

  outstring = safe_malloc_atomic(strlen(input) + 1 + numspecial);
  inpos = input;
  outpos = outstring;
  while (*inpos){
    if (strchr(specials, *inpos)){
      *outpos = '\\';
      outpos++;
    }
    if (*inpos == '\n')
      *outpos = 'n';
    else
      *outpos = *inpos;
    outpos++;
    inpos++;
  }
  *outpos = 0;
  return outstring;
}

/* Prints an item that we've already verified to be a symbol. */
static char *print_symbol(Item input)
{
  int funnychars = 0;
  char *ptr, *string;

  if (at_end(input))
    return "#END";
  if (!present(input))
    return "#MISSING";
  if (eq(input,ANY))
    return "#ANY";

  /* Check for any funny characters inside input. */
  string = symbol2string(input);
  if (!string){
    fprintf(stderr, "Warning:  null string produced by print_symbol\n");
    return gcprintf("[??? %ld]", input.uinteger);
  }
  ptr = string;
  while (*ptr != 0){
    if (isspace(*ptr) || strchr("\\#%$()'{}[];\":", *ptr)){
      funnychars = 1;
      break;
    }
    ptr++;
  }
  if (isdigit(*string) || strchr("-.",*string))
    funnychars = 1;

  /* Use a special escape syntax if symbol contains any characters that
     might interfere with normal parsing. */
  if (funnychars)
    return gcprintf("{%s}", slashify(string));
  else
    return gcprintf("%s", string);
}

/* Helper function for printitem. 
     --  indenton is 1 if indenting is enabled (which propagates from
         outer constituents to inner ones) and 0 if it isn't. 
     --  indentamt keeps track, recursively, of how many spaces to indent.
     --  depth tracks the depth of nesting (lists and bundles) 
     --  fulldetails turns off thresholds on nesting and list length, so make
         sure whole structure is printed.   (Caller must ensure it's not circular.)
*/ 
static char *printitemx (Item input, int depth, int indenton, int indentamt, int fulldetails)
{
  int type;
  Item head;
  static int allsetup = 0;
  static Item quotetag;
  if (!allsetup){
    quotetag = string2symbol("quote");
    allsetup = 1;
  }

  type = item2type(input);
  switch (type){
  case TYPESTRING:
    return gcprintf("\"%s\"", slashify(item2string(input))); 
    break;
  case TYPEINT:
    return gcprintf("%ld", item2long(input)); 
    break;
  case TYPELIST:
    if (null(input))
      return "()";
    else if (!fulldetails && depth >= PRINDEPTH)
      return "#";
    else if (length(input) == 2 && eq(car(input), quotetag))
      return(gcprintf("'%s",
		      printitemx (car(cdr(input)), depth, indenton, 
				 indentamt, fulldetails))); 
    else
      return gcprintf("(%s)", 
		      printlist(input, depth, indenton, indentamt, fulldetails));
    break;
  case TYPEFLOAT:
    return gcprintf("%.6g", item2float(input));
    break;
  case TYPESYMBOL:
    return print_symbol(input);
    break;
  case TYPENGRAM:
    return gcprintf("$%s", printitemx(ngram2list(input), depth, indenton, indentamt, fulldetails));
    break;
  case TYPESEQ:
    return gcprintf("#%s", printitemx(seq2list(item2seq(input)), 
				      depth, indenton, indentamt, fulldetails));
    break;
  case TYPEBUNDLE:
    head = bundlehead(input);
    if (!fulldetails && depth >= PRINDEPTH)
      return "#";
    else if (!present(head))
      return gcprintf("[%s]", printfeatures(input,depth,indenton, indentamt, fulldetails, 0));
    else
      return gcprintf("[%%%s %s]", printitem(bundlehead(input)), 
		      printfeatures(input,depth,indenton, indentamt, fulldetails, 1));
    break;
  case TYPECLOSURE:
    return gcprintf("#[CLOSURE: %d]", input.uinteger);
    break;
  case TYPEFUNCTION: 
    return gcprintf("#[FUNCTION: %d]", input.uinteger);
    break;
  case TYPESPECIAL: 
    return gcprintf("#[%s: %d]", printitem(item2label(input)), input.uinteger);
    break;
  default:
    return gcprintf("#[??mystery type %d]", type);
    break;
  }
}

/* ================================================================ */
/* Printing top-level functions                                     */
/* ================================================================ */

/* These printfunctions are designed for objects of only modest size,
   because they create a single string containing the entire printed
   representation.   If you wish to print a very big object (e.g.
   a very long list), break it up (e.g. loop through it) explicitly. */

/* Prints the item into a newly-allocated string. 
   Prints entire structure, even if long/deep.   Will loop infinitely 
   if structure is circular. */
char *printitemfull (Item input)
{
  return printitemx(input, 0, 1, 0, 1);
}

/* Prints the item into a newly-allocated string.
   Imposes thresholds on depth of nesting and length of lists, to ensure
   output is human-friendly.   */
char *printitem (Item input)
{
  char *output;
  output = printitemx(input, 0, 1, 0, 0);
  return output;
}



/* Takes an integer representing a type and turns it into a human-friendly
   string, e.g. for writing error messages. */
char *type2string(unsigned int type)
{
  switch (type){
  case TYPESTRING:
    return "string";
    break;
  case TYPEINT:
    return "integer";
    break;
  case TYPELIST:
    return "list";
    break;
  case TYPEFLOAT:
    return "float";
    break;
  case TYPESYMBOL:
    return "symbol";
    break;
  case TYPEBUNDLE:
    return "bundle";
    break;
  case TYPEFUNCTION: 
    return "function";
    break;
  case TYPECLOSURE: 
    return "closure";
    break;
  case TYPESEQ: 
    return "seq";
    break;
  case TYPENGRAM: 
    return "ngram";
    break;
  case TYPESPECIAL: 
    return "special";
    break;
  default:
    return gcprintf("mysterytype");
    break;
  }
}

/* ================================================================ */
/* Formatted output                                                 */
/* ================================================================ */


/* Write formatted output into a string.
   Input should consist of a format string followed by some number of Items.
   Substitution locations for Items are marked in the format string by %s.
   The directive %x is used for inserting strings without enclosing quotes.
   Very long items are truncated if infull is zero.
   Use %% to print the % character itself. 
   For example, (format "The outputs are %s and %s\n", '(a b c)  myvar)  */
static char *format_helper(Item args, int infull)
{
  char *stringpattern, *ptr, *subst;
  Item pattern;
  Item outlist = NIL;

  if (null(args))
    error("No inputs given to format command\n");
  pattern = car(args);
  args = cdr(args);

  stringpattern = item2string(pattern);
  ptr = stringpattern;
  while (*ptr != 0){
    subst = strchr(ptr, '%');
    if (!subst || *(subst+1) == 0){
      outlist = cons(string2item(ptr), outlist);
      break;
    }
    if (subst > ptr)
      outlist = cons(string2item(substring(ptr,subst-1)), outlist);
    ptr = subst;

    /* ptr is now at a substitution location */
    if (*(ptr+1) == '%')
      outlist = cons(string2item("%"), outlist);
    else if (*(ptr+1) == 's'){
      if (null(args))
	error("Ran out of args during format:  %s\n", printitem(args));
      if (infull)
	outlist = cons(string2item(printitemfull(car(args))),outlist);
	else
	outlist = cons(string2item(printitem(car(args))),outlist);
      args = cdr(args);
    }
    else if (*(ptr+1) == 'x'){
      if (null(args))
	error("Ran out of args during format:  %s\n", printitem(args));
      if (item2type(car(args)) != TYPESTRING)
	error("Format directive %x got non-string %s\n", printitem(args));
      outlist = cons(car(args),outlist);
      args = cdr(args);
    }
    else
      error("Unrecognized format directive %%%c\n", *(ptr+1));
    ptr+= 2;
  }
  return copyinstrings(nreverse(outlist));
}

/* Write formatted output into a string.
   Input should consist of a format string followed by some number of Items.
   Substitution locations for Items are marked in the format string by %x or %s.
   Use %% to print the % character itself. 
   For example, (format "The outputs are %s and %s\n", '(a b c)  myvar)  */
char *format(Item args)
{
  return format_helper(args,0);
}

/* Similar, but prints full details of each Item, at the risk of looping
   on circular datastructures. */
char *format_infull(Item args)
{
  return format_helper(args,1);
}

