/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* Code for printing and reading s-expressions.
   Margaret M. Fleck, 2005  */

/* The print routine incorporates a very primitive indentation algorithm. 
   It's not wonderful, but it beats the hell out of having nothing.  */

/* The print routines in this file produce string output, so they can
   be used from within C code.  Actually sending this string to an I/O 
   stream is done elsewhere, e.g. in the interpreter code in eval.c. 

*/


#include "schwa.h"

/* ================================================================ */
/* Reader helper functions                                          */
/* ================================================================ */

/* Notice that this reader only handles straightforward types, such as
   symbols and lists.   It does not provide a way to input things like
   functions or other difficult/special types.  */


/* Skip any whitespace and/or comments intervening before the next real token.  */
static void skip_ws(FILE *infile)
{
  int nextchar;

  while ((nextchar = fgetc(infile)) != EOF && isspace(nextchar));

  if (nextchar == EOF)
    return;
  else if (nextchar == ';'){
    /* a comment, terminated by end of line */
    while ((nextchar = fgetc(infile)) != EOF && nextchar != '\n');
    skip_ws(infile);
    return;
  }
  else
    ungetc(nextchar,infile);
}



/* We've just read the first character of a multi-character token (e.g. string, number,
   symbol).  Read the rest and return it. */
static char *read_long_token(FILE *infile, char firstchar)
{
  int length = 10, pos, stringp = 0, slashcount = 0;
  char *output;
  int nextchar, delimiter = 0;

  output = safe_malloc_atomic(length+1);
  output[0] = firstchar;
  pos = 1;
  
  if (firstchar == '"'){
    stringp = 1;
    delimiter = '\"';
  }
  else if (firstchar == '{'){
    stringp = 1;
    delimiter = '}';
  }

  while ((nextchar = fgetc(infile)) != EOF){
    if (pos >= length){
      length = length*2;
      output = safe_realloc(output,length+1);
    }
    if (stringp){
      /* Strings are terminated by a double quote, funny symbols by a curly bracket. */
      output[pos] = nextchar;
      pos++;
      /* slashcount is 1 iff we've just seen an unfinished slash escape */
      if (nextchar == '\\')  
	slashcount = (slashcount+1)%2;   
      else if (nextchar == delimiter && slashcount == 0)
	break;
      else  /* any other character */
	slashcount = 0;
    }
    else {
      /* Other tokens are terminated by whitespace or one
	 of a small list of delimiter characters. */
      if (isspace(nextchar))
	break;
      else if (strchr("{}()[];\":", nextchar)){
	ungetc(nextchar,infile);
	break;
      }
      /* The new character extends the current token. */
      else {
	output[pos] = nextchar;
	pos++;
      }
    }
  }
  output[pos] = 0;
  return output;
}

/* Read next S-expression token from (open) file.
   Returns 0 at end of file. */
static char *read_stoken (FILE *infile)
{
  int nextchar, tempchar;
  char *output;

  /* skip whitespace */
  skip_ws(infile);

  nextchar = fgetc(infile);
  if (nextchar == EOF)
    return 0;
  if (strchr("()[]':", nextchar)){
    /* These punctuation characters always form one-character tokens,
       though some can occur embedded in longer tokens. */
    output = safe_malloc_atomic(2);
    output[0] = nextchar;
    output[1] = 0;
    return output;
  }

  /* The 2-character tokens #( and $( mark the start of seqs and ngrams. 
     #[ marks the start of an (unreadable) special object display. */
  if (nextchar == '#' || nextchar == '$'){
    tempchar = fgetc(infile);
    if (tempchar == '(')
      return gcprintf("%c(", nextchar);
    else if (nextchar == '#' && tempchar == '[')
      return gcprintf("%c[", nextchar);
    else
      ungetc(tempchar,infile);
  }

  /* Otherwise, we're at the start of a plain ordinary token like
     a symbol or number. */
  return read_long_token(infile,nextchar);
}


/* Unslashify string or special symbol token and strip its delimiters. */
static char *convert_string(char *input, char delimiter)
{
  char *outstring, *inpos, *outpos, outchar;
  int length;

  /* Remove leading and (if present) trailing delimiters. */
  outstring = input+1;
  length = strlen(outstring);
  if (outstring[length-1] == delimiter)
    outstring[length-1] = 0;

  /* Undo slashification.   Modifies outstring in place. */
  inpos = outstring;
  outpos = outstring;
  while (*inpos){
    if (*inpos == '\\' && *(inpos+1) != 0){
      if (*(inpos+1) == 'n')
	outchar = '\n';
      else 
	outchar = *(inpos+1);
      *outpos = outchar;
      outpos++;
      inpos += 2;
    }
    else {
      *outpos = *inpos;
      outpos++;
      inpos++;
    }
  }
  *outpos = 0;
  return outstring;
}

/* Convert a token starting with #. 
   Caller should have checked that the first character in input is #. */
static Item convert_hash(char *input)
{

  /* Strip the leading # */
  input = input+1;
  if (!strcmp(input,"MISSING"))
    return MISSING;
  else if (!strcmp(input,"NIL"))
    return NIL;
  else if (!strcmp(input,"ANY"))
    return ANY;
  else if (!strcmp(input,"END"))
    return END;
  else if (*input == '[')
    error("Printed representations for special objects (starting with '#[') can't be read back in.\n");
  else
    error("Bad #-based syntax:  #%s\n", input);
  return MISSING;  /* not actually reacheable. */
}

/* Convert atomic token into Item of appropriate type. 
   Assumes caller has already checked that input string is non-null
   and not a paren or bracket. */
static Item convert_atom(char *input)
{
  if (*input == '"')
    error("Why did convert_atom get called on string?  %s\n", input);
  /*     return string2item(substring(input+1,input+strlen(input)-2)); */
  else if (isalpha(*input) || strchr("+*/$%<>=@", *input))
    return string2symbol(input);
  else if (!strcmp(input,"-"))
    /* kludge for the subtraction operator */
    return string2symbol(input);
  else if (isdigit(*input) || *input == '.' || *input == '-')
    return(string2number(input));
  else if (*input == '#')
    return convert_hash(input);
  else
    error("Reader (convert_atom) hit unexpected token: %s\n", input);
  return MISSING;  /* this statement is actually unreadable */
}

static Item read_sexp_helper (FILE *infile, char *first);

/* We've just read opening paren of list.   Read items until closing paren. */
static Item read_list (FILE *infile)
{
  Item outlist = NIL, nextitem;
  char *nexttoken;
  while ((nexttoken = read_stoken(infile)) != 0){
    if (*nexttoken == ')')
      return nreverse(outlist);
    else{
      nextitem = read_sexp_helper(infile,nexttoken);
      outlist = cons(nextitem, outlist);
    }
  }
  error("Unexpected end of file (read_list)\n");
  return MISSING;  /* this statement is actually unreadable */
}

/* We've just read opening bracket of bundle.   Read items until closing bracket. */
static Item read_bundle (FILE *infile)
{
  Item output, feature, value;
  Item head = MISSING;
  int nextchar;
  char *nexttoken;

  /* read head of bundle, if any */
  nextchar = fgetc(infile);
  if (nextchar == '%'){
    head = read_sexp(infile);
  }
  else if (nextchar != EOF)
    ungetc(nextchar,infile);

  nexttoken = read_stoken(infile);
  output = new_bundle(head); 

  while (nexttoken != 0){
    if (*nexttoken == ']')
      return output;
    else{
      feature = convert_atom(nexttoken);
      nexttoken = read_stoken(infile);
      if (nexttoken == 0)
	break;
      if (strcmp(nexttoken, ":"))  /* feature with anonymous value */
	bundleset(output,feature,ANY);
      else {
	value = read_sexp(infile);
	bundleset(output,feature,value);
	nexttoken = read_stoken(infile);
      }
    }
  }
  error("Unexpected end of file (read_bundle)\n");
  return MISSING;  /* this statement is actually unreadable */
}

/* We've just read first token in sexpression.   Finish reading it. */
static Item read_sexp_helper (FILE *infile, char *first)
{

  if (*first == '(')
    return read_list(infile);
  else if (*first == '[')
    return read_bundle(infile); 
  else if (*first == '\'')
    return list(2, string2symbol("quote"), read_sexp(infile));
  else if (*first == '"')
    return string2item(convert_string(first, '\"'));
  else if (*first == '{')
    return string2symbol(convert_string(first, '}'));
  else if (!strcmp(first,"#("))
    return seq2item(list2seq(read_list(infile)));
  else if (!strcmp(first,"$("))
    return list2ngram(read_list(infile));
  else 
    return convert_atom(first);
}

/* ================================================================ */
/* Reader top-level functions                                       */
/* ================================================================ */

/* Read next S-expression from (open) file. 
   Returns END at end of file.
   END is also returned if the file contains the token #END.  Caveat emptor. 
      Writing this token into a file effectively truncates it.  Use
      the function at_end to test whether a read-type function has
      hit the end of a file. */
Item read_sexp (FILE *infile)
{
  char *next;

  next = read_stoken(infile);

  if (next == 0)
    return END;
  else 
    return read_sexp_helper(infile, next);
}


/* Read all s-expressions in file, returning a list of Items. 
   File should contain only s-expressions and/or comments starting with semicolon.
   The expressions are not evaluated. */
Item read_file (char *filename)
{
  FILE *fp;
  Item output = NIL, next;
  
  fp = fopen(filename,"r");
  if (!fp)
    error("Can't open file %s\n", filename);

  while (!at_end(next = read_sexp(fp)))
    output = cons(next, output);

  fclose(fp);

  return nreverse(output);
}
