/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/

/* Setting up the standard lisp interpreter environment,
   which you need to do prior to calling invoke_interpreter.

   The top-level function is standard_env, which returns
   such an environment.  

   The standard environment is deliberately somewhat sparse.
   You will want to extend it in ways appropriate to your needs.
   Use defun, defun_vararg, defun_sf, and setsymbol to add new 
   functions, special forms, and symbols.  Setsymbol and 
   evalsymbol can be used to rename functions, remove unwanted 
   definitions, etc.

   You can also define your own mutant environment by writing 
   an entirely new function similar to standard_env.  

   Browse this file for examples of how to do such things.
   
   In particular, notice the handy variables TRUE and FALSE
   to help when manipulating boolean expressions.   These
   are simply Item versions of 1 and 0.

   The functions in this file are declared static to avoid polluting
   your C namespace.   These functions aren't ones that you'd be
   likely to use in your C code.

   Margaret M. Fleck, 2005
*/

#include "schwa.h"

/* ================================================================ */
/* File I/O                                                         */
/* ================================================================ */

/* We implement file handles using the special complex type mechanism. 
   These objects will have label FILE_HANDLE.
   Therefore, don't use this label for your own special objects. 

   These objects are intended only for use in the interpreter, or by
   wrapper functions interfacing your C code to the interpreter.   C
   functions should use normal C file handles. */

/* openfiles is a list of all file handles that are currently open, in 
   case the user forgot to capture the value returned by a file open command.
   Make sure this initializer is the same as that for NIL. */
static Item openfiles = {TAGCOMPLEX};

/* Return list of all open files. */
static Item get_open_files(void)
{
  return openfiles;
}

/* Open a file for input, returning the file handle. */
static Item open_input_file(Item filename)
{
  FILE *fp;
  Item handle;

  fp = fopen(item2string(filename), "r");
  if (!fp)
    error("Can't open file %s\n", item2string(filename));
  handle = ptr2item(fp,string2symbol("FILE_HANDLE"));
  openfiles = cons(handle,openfiles);

  return handle;
}

/* Like open_input_file, except opens the file for output. */
static Item open_output_file(Item filename)
{
  FILE *fp;
  Item handle;

  fp = fopen(item2string(filename), "w");
  if (!fp)
    error("Can't open file %s\n", item2string(filename));
  handle = ptr2item(fp,string2symbol("FILE_HANDLE"));
  openfiles = cons(handle,openfiles);

  return handle;
}


/* Close the file connection associated with this handle. 
   Zeros the pointer inside the handle objects and removes it from the list
   of open files.  */
static Item close_file(Item handle)
{
  FILE *fp;
  Item newlist = NIL, ptr;

  fp = item2ptr(handle, string2symbol("FILE_HANDLE"));

  if (!fp)
    fprintf(stderr, "WARNING: attempt to close a closed file\n");
  else if (fp == stderr || fp == stdin || fp == stdout)
    fprintf(stderr, "Attempt to close standard input, output, or error.\n");
  else{
    fclose(fp);
    ptr = openfiles;
    while (!null(ptr)){
      if (!eq(car(ptr),handle))
	newlist = cons(car(ptr),newlist);
      ptr = cdr(ptr);
    }
    openfiles = nreverse(newlist);
    setspecialptr(handle,0);
  }

  return MISSING;
}

/* Read entire contents of a file, which should contain only s-expressions. */
static Item read_file_wp(Item filename)
{
  return read_file(item2string(filename));
}

/* The load command reads forms one-by-one from a file, evaluating
   each one.   Input forms and output values are echoed.   (To
   define a similar but silent function, use 0 as the third
   input to evalfile. */
static Item load_sf(Item args, Schwa_env *environment)
{
  if (length(args) != 1)
    error("Wrong number of args to load:  %s", printitem(args));
  evalfile(item2string(car(args)), environment, 1);
  return MISSING;
}

/* ================================================================ */
/* Directory information                                            */
/* ================================================================ */

/* Returns a list of filenames matching input specification */
static Item ls (char *specification)
{
  FILE *fp;
  Item output = NIL;
  char *next, *command;

  command = gcprintf("ls -d %s",specification);
  fp = popen(command, "r");
  if (!fp)
    error("Can't execute ls command %s\n", command);
  
  while((next=read_token(fp)))
    output = cons(string2item(next), output);

  pclose(fp);
  return nreverse(output);
}

static Item ls_wp(Item specification)
{

  return ls(item2string(specification));
}

/* ================================================================ */
/* Reading and writing                                              */
/* ================================================================ */

/* These functions all read/write to/from the current input and output
   channels (*evalin* and *evalout*).    By default, these are stdin and
   stdout.   To switch input/output to another (open) file handle,
   wrap your input/output commands inside the special forms 
   io_to and io_from. 

   The top-level functions are defined as special forms, because they
   require access to the environment to manipulate *evalin* and *evalout*.
*/

static FILE *get_default_input(Schwa_env *environment)
{
  static Item intag;
  static int setup = 0;  
  FILE *fp;
  Item ifp;

  if (!setup){
    setup = 1;
    intag = string2symbol("*evalin*");
  }

  ifp = evalsymbol(intag,environment);

  fp = item2ptr(ifp, string2symbol("FILE_HANDLE"));
  if (!fp)
    error("Attempt to read from a closed file handle\n");

  return fp;
}

static FILE *get_default_output(Schwa_env *environment)
{
  static Item outtag;
  static int setup = 0;  
  FILE *fp;
  Item ifp;

  if (!setup){
    setup = 1;
    outtag = string2symbol("*evalout*");
  }

  ifp = evalsymbol(outtag,environment);

  fp = item2ptr(ifp, string2symbol("FILE_HANDLE"));

  if (!fp)
    error("Attempt to write to a closed file handle\n");

  return fp;
}

/* Read the next s-expression from the current input channel. */
static Item read_sexpression(Item args, Schwa_env *environment)
{

  FILE *fp;
  fp = get_default_input(environment);

  return read_sexp(fp);
}

/* Write s-expressions into the open file to the current output channel. 
   Truncates structure if it is very deep, long, and/or circular. 
   That is, suitable for displaying data to the user. */
static Item write_sexp(Item args, Schwa_env *environment)
{
  FILE *fp;

  fp = get_default_output(environment);

  args = evalargs(args,environment);

  while (!null(args)){
    fprintf(fp,"%s ", printitem(car(args)));
    fflush(fp);
    args = cdr(args);
  }
  return MISSING;
}

/* Write s-expression into the open file to the current output channel. 
   Attempts to write entire structure, even if deep/long.
   That is, suitable for writing out data to a file. 
   Be aware that this will loop if the structure is circular. */
static Item write_sexp_full(Item args, Schwa_env *environment)
{
  FILE *fp;

  fp = get_default_output(environment);

  args = evalargs(args,environment);

  while (!null(args)){
    fprintf(fp,"%s ", printitemfull(car(args)));
    fflush(fp);
    args = cdr(args);
  }
  fflush(fp);
  return MISSING;
}

/* Write formatted output into a string. */
static Item sformat_wp(Item args, Schwa_env *environment)
{
  return string2item(format(evalargs(args, environment)));

}

/* Write formatted output into a string. */
static Item sformat_if_wp(Item args, Schwa_env *environment)
{
  return string2item(format_infull(evalargs(args, environment)));

}

/* Write formatted output to the current output channel. 
   Input should consist of a format string followed by some number of Items.
   Substitution locations for Items are marked in the format string by %s
   or %x.  Use %% to print the % character itself. 
   For example, (format "The outputs are %s and %s\n", '(a b c)  myvar)  */
static Item format_wp(Item args, Schwa_env *environment)
{
  FILE *fp;
  fp = get_default_output(environment);

  args = evalargs(args, environment);
  fprintf(fp, "%s", format(args));
  fflush(fp);
  return MISSING;
}

/* Similar, but prints full details of each Item, at the risk of looping
   on circular datastructures. */
static Item format_if_wp(Item args, Schwa_env *environment)
{
  FILE *fp;
  fp = get_default_output(environment);

  args = evalargs(args, environment);
  fprintf(fp, "%s", format_infull(args));
  fflush(fp);
  return MISSING;
}


static Item set_prindepth(Item newval)
{
  long i;

  i = item2long(newval);
  if (i <= 0)
    error("Printdepth must be positive\n");
  PRINDEPTH = i;
  return MISSING;
}

static Item set_prinlength(Item newval)
{
  long i;

  i = item2long(newval);
  if (i <= 0)
    error("Printlength must be positive\n");
  PRINLENGTH = i;
  return MISSING;
}

static Item set_linewidth(Item newval)
{
  long i;
  if (item2type(newval) != TYPEINT)
    error("Printlength must be an integer\n");
  i = item2long(newval);
  if (i <= 0)
    error("Linewidth must be positive\n");
  LINEWIDTH = i;
  return MISSING;
}


/* ================================================================ */
/* Special forms:   conditionals                                    */
/* ================================================================ */

/* Special forms have access to the environment and can decide how
   and when to evaluate their input arguments.  This greatly reduces
   the number of syntactic constructs that need to be hardwired into
   the interpreter. */

static Item if_sf(Item args, Schwa_env *environment)
{
  Item value;
  if (length(args) < 2)
    error("Too few args to if:  %s", printitem(args));
  if (length(args) > 3)
    error("Too many args to if:  %s", printitem(args));
  value = eval(car(args),environment);
  if (!eq(FALSE,value))
    return eval(nth(1,args),environment);
  else if (length(args) >= 3)
    return eval(nth(2,args),environment);
  else 
    return MISSING;
}

static Item cond_sf(Item args, Schwa_env *environment)
{
  Item value, clause;

  while (!null(args)){
    clause = car(args);
    args = cdr(args);
    if (null(clause))
      error("Null clause in cond\n");
    value = eval(car(clause),environment);
    if (!eq(FALSE,value))
      return(evalsequence(cdr(clause),environment));
  }
    return MISSING;
}

static Item case_sf(Item args, Schwa_env *environment)
{
  Item value, clause, testitem;

  if (null(args))
    error("case requires at least one argument\n");
  testitem = eval(car(args), environment);
  args = cdr(args);

  while (!null(args)){
    clause = car(args);
    args = cdr(args);
    if (null(clause))
      error("Null clause in case statement\n");
    value = car(clause);
    if (eq(value, ANY) || memq(testitem, value))
      return(evalsequence(cdr(clause),environment));
  }
  return MISSING;
}


/* when is similar to if.   However, it is only single-branch (if but
   not else) and a series of statements can be placed after the test */
static Item when_sf(Item args, Schwa_env *environment)
{
  Item value;
  if (length(args) < 1)
    error("No args to when:  %s", printitem(args));
  value = eval(car(args),environment);
  if (!eq(FALSE,value))
    return(evalsequence(cdr(args),environment));
  else 
    return MISSING;
}

/* ================================================================ */
/* Special forms:  loops                                            */
/* ================================================================ */


static Item while_sf(Item args, Schwa_env *environment)
{
  if (null(args))
    error("Too few args to while:  %s", printitem(args));
  while (!eq(FALSE,eval(car(args), environment)))
    evalsequence(cdr(args), environment);
  return MISSING;
}


/* A for loop.
  (for (i a b) (....))
      assigns i to values a through (and including) b.
   If a > b, the loop counts downwards.
   Example:
      (for (i 5 7) (format "%s\n" i)) 
      prints the values 5, 6, and 7. */
static Item for_sf(Item args, Schwa_env *environment)
{
  Item increment, symbol, expr, bound, oldbindings;
  int i, max, min;

  if (null(args))
    error("No args in for loop %s\n", printitem(args));
  increment = car(args);
  if (length(increment) != 3 
      || (item2type(car(increment))) != TYPESYMBOL)
    error("Malformed increment expression in for:  %s\n", 
	  printitem(increment));
  symbol = car(increment);

  bound = eval(nth(1,increment),environment);
  min = item2long(bound);
  bound = eval(nth(2,increment),environment);
  max = item2long(bound);
  expr = cdr(args);

  oldbindings = environment->localbindings;
  environment->localbindings 
    = cons(list(2,symbol, long2item(min)),oldbindings);

  if (min <= max){
    for (i=min; i <= max; i++){
      setsymbol(symbol, long2item(i),environment);
      evalsequence(expr, environment);
    }
  }
  else{
    for (i=min; i >= max; i--){
      setsymbol(symbol, long2item(i),environment);
      evalsequence(expr, environment);
    }
  }

  environment->localbindings = oldbindings;

  return MISSING;
}


/* Iterate over a list. */
static Item dolist_sf(Item args, Schwa_env *environment)
{
  Item increment, symbol, expr;
  Item ptr, oldbindings;

  if (null(args))
    error("No args for dolist:  %s\n", printitem(args));
  increment = car(args);
  if (length(increment) != 2 
      || (item2type(car(increment))) != TYPESYMBOL)
    error("Malformed increment expression in dolist:  %s\n", 
	  printitem(increment));
  symbol = car(increment);
  ptr = eval(nth(1,increment), environment);
  if (item2type(ptr) != TYPELIST)
    error("Iteration domain in dolist is not a list:  %s\n", 
	  printitem(cadr(increment)));
  expr = cdr(args);

  oldbindings = environment->localbindings;
  environment->localbindings = cons(list(2,symbol,MISSING), oldbindings);

  while (!null(ptr)){
    setsymbol(symbol, car(ptr), environment);
    evalsequence(expr, environment);
    ptr = cdr(ptr);
  }

  environment->localbindings = oldbindings;   /* restore old binding list */

  return MISSING;
}

/* Iterate over a seq. */
static Item doseq_sf(Item args, Schwa_env *environment)
{
  Item increment, symbol, expr;
  Item theseq, oldbindings;
  Item *ptr;

  if (null(args))
    error("No args for doseq:  %s\n", printitem(args));
  increment = car(args);
  if (length(increment) != 2 
      || (item2type(car(increment))) != TYPESYMBOL)
    error("Malformed increment expression in dolist:  %s\n", 
	  printitem(increment));
  symbol = car(increment);
  theseq = eval(nth(1,increment), environment);
  if (item2type(theseq) != TYPESEQ)
    error("Iteration domain in doseq is not a seq:  %s\n", 
	  printitem(cadr(increment)));
  ptr = item2seq(theseq);
  expr = cdr(args);

  oldbindings = environment->localbindings;
  environment->localbindings = cons(list(2,symbol,MISSING), oldbindings);

  while (!at_end(*ptr)){
    setsymbol(symbol, *ptr, environment);
    evalsequence(expr, environment);
    ptr++;
  }

  environment->localbindings = oldbindings;   /* restore old binding list */

  return MISSING;
}

/* Iterate over a hashtable.   The iteration variable is
   bound, one by one, to each (key value) pair. */
static Item dohash_sf(Item args, Schwa_env *environment)
{
  Item increment, symbol, expr, table, kvpair, oldbindings;
  Hashptr *ptr;
  Hashtable *rawtable;

  if (null(args))
    error("No args for dohash:  %s\n", printitem(args));
  increment = car(args);
  if (length(increment) != 2 
      || (item2type(car(increment))) != TYPESYMBOL)
    error("Malformed increment expression in dohash:  %s\n", 
	  printitem(increment));

  symbol = car(increment);
  table = eval(cadr(increment), environment);
  expr = cdr(args);
  
  oldbindings = environment->localbindings;
  environment->localbindings = cons(list(2,symbol,MISSING), oldbindings); 

  rawtable = (Hashtable *)item2ptr(table, string2symbol("HASHTABLE"));
  ptr = hashstart(rawtable);
  while (!at_end(kvpair=hashnext(ptr))){
    setsymbol(symbol, kvpair, environment);
    evalsequence(expr, environment);
  }

  environment->localbindings = oldbindings;

  return MISSING;
}

/* Iterate over an earray.   The iteration variable is
   bound, one by one, to each (key value) pair. */
static Item doearray_sf(Item args, Schwa_env *environment)
{
  Item increment, symbol, expr, table, kvpair, oldbindings, isym;
  Earrayptr *iter;
  unsigned long position;
  Item *ptr;

  if (null(args))
    error("No args for dohash:  %s\n", printitem(args));
  increment = car(args);
  if (length(increment) != 2 
      || (item2type(car(increment))) != TYPESYMBOL)
    error("Malformed increment expression in doearray:  %s\n", 
	  printitem(increment));

  symbol = car(increment);
  table = eval(cadr(increment), environment);
  expr = cdr(args);
  
  oldbindings = environment->localbindings;
  environment->localbindings = cons(list(2,symbol,MISSING), oldbindings); 
  isym = string2symbol("EARRAY");
  table = eval(cadr(increment), environment);
  iter = earraystart(item2ptr(table, isym));
  while ((ptr = earraynext(iter, &position)) != 0){
    kvpair = list(2, long2item(position), *ptr);
    setsymbol(symbol, kvpair, environment);
    evalsequence(expr, environment);
  }

  environment->localbindings = oldbindings;

  return MISSING;
}

/* ================================================================ */
/* Special forms:   I/O                                             */
/* ================================================================ */

/* A special form is required to manipulate the default input and
   output channels, because this requires access to the environment. 
   Temporary bindings for these channels have dynamic scope, so that
   the bindings will be propagated if our function invokes another
   function. */

/* Change the input channel while evaluating a list of expressions.  */
static Item iofrom_sf(Item args, Schwa_env *environment)
{
  Item value, handle, oldbindings;
  static Item intag;
  static int setup = 0;

  if (!setup){
    setup = 1;
    intag = string2symbol("*evalin*");
  }

  if (null(args))
    error("No args for evaliofrom\n");
  handle = eval(car(args),environment);
  
  if (!is_special_type(handle,string2symbol("FILE_HANDLE")))
    error("Non-file first argument to evaliofrom:  %s\n", printitem(car(args)));

  /* New value for *evalin* is set up as a temporary dynamic binding. */
  oldbindings = environment->dynamicbindings;
  environment->dynamicbindings = cons(list(2,intag,handle), oldbindings); 

  value = evalsequence(cdr(args), environment); 

  /* Remove temporary binding. */
  environment->dynamicbindings = oldbindings;

  return value;
}

/* Change the output channel while evaluating a list of expressions.  */
static Item ioto_sf(Item args, Schwa_env *environment)
{
  Item value, handle, oldbindings;
  static Item outtag;
  static int setup = 0;

  if (!setup){
    setup = 1;
    outtag = string2symbol("*evalout*");
  }


  if (null(args))
    error("No args for evalioto\n");
  handle = eval(car(args),environment);
  if (!is_special_type(handle,string2symbol("FILE_HANDLE")))
    error("Non-file first argument to evalioto:  %s\n", printitem(car(args)));

  /* New value for *evalout* is set up as a temporary dynamic binding. */
  oldbindings = environment->dynamicbindings;
  environment->dynamicbindings = cons(list(2,outtag,handle), oldbindings); 

  value = evalsequence(cdr(args), environment); 

  /* Remove temporary binding. */
  environment->dynamicbindings = oldbindings;

  return value;
}



/* ================================================================ */
/* Control structure etc                                            */
/* ================================================================ */

/* Invokes the error function, as in C.
   First arg should be the format string.
*/
static Item error_wp(Item args)
{
  char *message;
  message = format(args);
  error(message);
  return MISSING;
}

/* Exit the interpreter session.  Unlike error, this is intended to
   be a planned exit.  If invoked from a shell script, will terminate
   the script but this will still be followed by an interactive
   shell if leave_live has been set. */
static Item exit_sf(Item arglist, Schwa_env *environment)
{
  /* Make it so interpreter won't auto-restart */
  environment->errorrestart = 0;
  /* Generate an error break, but without error message. */
  if (errorjump)
    siglongjmp(errorenv, 1);
  else 
    error("Internal error: no jump point for evalexit\n");
  return MISSING;
}

/* Evaluate expression in the current environment */
static Item eval_sf (Item arglist, Schwa_env *environment)
{
  if (length(arglist) != 1)
    error("Wrong number of inputs to eval:  %s\n", printitem(arglist));
  return eval(eval(car(arglist), environment), environment);
}

/* Evaluate apply function to args in the current environment */
static Item apply_sf (Item arglist, Schwa_env *environment)
{
  Item head, args;
  if (length(arglist) != 2)
    error("Wrong number of inputs to apply:  %s\n", printitem(arglist));
  head = eval(car(arglist), environment);
  args = eval(cadr(arglist), environment);
  return apply(head, args, environment);
}

static Item setprompt_sf(Item arglist, Schwa_env *environment)
{
  Item val;
  if (length(arglist) != 1)
    error("setprompt requires exactly one input\n");
  val = eval(car(arglist), environment);
  environment->prompt = item2string(val);
  return MISSING;
}

/* ================================================================ */
/* Misc operations                                                  */
/* ================================================================ */

static Item eq_wp(Item a, Item b)
{
  return long2item(eq(a,b));
}

static Item equal_wp(Item a, Item b)
{
  return long2item(equal(a,b));
}

static Item null_wp(Item a)
{
  return long2item(null(a));
}

static Item at_end_wp(Item a)
{
  return long2item(at_end(a));
}

static Item present_wp(Item a)
{
  return long2item(present(a));
}


static Item memq_wp(Item a, Item b)
{
  return long2item(memq(a, b));
}

static Item member_wp(Item a, Item b)
{
  return long2item(member(a, b));
}

/* Returns a symbol representing the type of the object */
static Item item2type_wp(Item input)
{
  return string2symbol(type2string(item2type(input)));
}

/* Builds a new bundle in one convenient step.
   First arg should be the head (or #MISSING if you don't want a head).  
   Each later arg should be a key or a (key value) pair.  */
static Item bundle_wp(Item args)
{
  Item output, next;
  if (null(args))
    error("No inputs to bundle creation function\n");
  output = new_bundle(car(args));
  args = cdr(args);
  while (!null(args)){
    next = car(args);
    switch (item2type(next)){
    case TYPESYMBOL:
      bundleset(output, next, ANY);
      break;
    case TYPELIST:
      if (length(next) != 2)
	error("Malformed key/value pair %s\n", printitem(next));
      bundleset(output, car(next), cadr(next));
      break;
    default:
      error("Malformed key/value pair %s\n", printitem(next));
    }
    args = cdr(args);
  }
  return output;
}

static Item bundleset_wp(Item args)
{
  Item value;

  if (length(args) > 3 || length(args) < 2)
    error("bundleset requires 2 or 3 args\n");
  if (length(args) == 3)
    value = nth(2,args);
  else
    value = ANY;
  return bundleset(car(args), cadr(args), value);
}


/* ================================================================ */
/* Hashtables                                                       */
/* ================================================================ */

Item new_hash_wp(void)
{
  return ptr2item(new_hashtable(), string2symbol("HASHTABLE"));
}     

static Item hashget_wp(Item table, Item key) 
{
  return hashget(item2ptr(table, string2symbol("HASHTABLE")), key);
}
  
static Item hashset_wp(Item table, Item key, Item value)
{
  hashset(item2ptr(table,string2symbol("HASHTABLE")), key, value);
  return MISSING;
}

/* Number of items in hashtable */
static Item hashlen(Item table)
{
  Hashtable *rawtable;
  rawtable = item2ptr(table,string2symbol("HASHTABLE"));

  return long2item(rawtable->numoccupants);
}

static Item hashkeys_wp(Item table)
{
  return seq2item(hashkeys(item2ptr(table,string2symbol("HASHTABLE"))));
}

static Item hashadd_wp(Item table, Item key, Item value)
{
  hashadd(item2ptr(table,string2symbol("HASHTABLE")), key, value);
  return MISSING;
}

static Item hashincrement_wp(Item table, Item key, Item amount)
{
  hashincrement(item2ptr(table,string2symbol("HASHTABLE")), 
		key,
		item2long(amount));
  return MISSING;
}


/* ================================================================ */
/* List, seq, ngram operations                                           */
/* ================================================================ */

/* The list function.
   Input arglist should be temporary list structure created by
   the reader, so no good reason to copy it.   */
static Item list_wp(Item arglist)
{
  return arglist;
}

static Item nth_wp(Item n,Item list) 
{
  return(nth(item2long(n),list));
}

/* input can be a list, string, or seq */
static Item length_wp(Item input)
{
  int type;

  type = item2type(input);

  switch (type){
  case TYPELIST:
    return long2item(length(input));
    break;
  case TYPESEQ:
    return long2item(seqlength(input));
    break;
  case TYPESTRING:
    return long2item(strlen(item2string(input)));
  default:
    error("length doesn't yet handle inputs of type %s\n", type2string(type));
    break;
  }
  return MISSING;
}


/* Create a seq of the specified length */
static Item newseq_wp(Item length)
{
  return new_seq(item2long(length));
}

/* Return the nth item in a seq.  */
static Item seqget_wp(Item seq, Item n)
{
  return seqget(seq, item2long(n));
}

/* Fill all positions in seq with a specific value. */
static Item seqfill_wp(Item seq, Item value)
{
  seqfill(seq, value);
  return MISSING;
}

/* Set the nth item in a seq */
static Item seqset_wp(Item seq, Item n, Item value)
{
  seqset(seq, item2long(n), value);
  return MISSING;
}

static Item sortseq_wp(Item input)
{
  return seq2item(sortseq(item2seq(input)));
}

static Item seq_wp(Item arglist)
{
  return seq2item(list2seq(arglist));
}

static Item ngram_wp(Item arglist)
{
  return list2ngram(arglist);
}



/* ================================================================ */
/* Type conversions                                                 */
/* ================================================================ */

static Item symbol2string_wp(Item input)
{
  return string2item(symbol2string(input));
}

static Item string2symbol_wp(Item input)
{
  return string2symbol(item2string(input));
}


static Item stringinterned_wp(Item input)
{
  return long2item(stringinterned(item2string(input)));
}


static Item string2number_wp(Item input)
{
  return string2number(item2string(input));
}


static Item list2ngram_wp(Item list)
{
  return list2ngram(list);
}

static Item listinterned_wp(Item list)
{
  return long2item(listinterned(list));
}

static Item ngram2list_wp(Item ngram)
{
  return ngram2list(ngram);
}

static Item seq2list_wp(Item seq)
{
  return seq2list(item2seq(seq));
}

static Item list2seq_wp(Item list)
{
  return seq2item(list2seq(list));
}


static Item seq2ngram_wp(Item seq)
{
  return seq2ngram(item2seq(seq));
}

static Item ngram2seq_wp(Item ngram)
{
  return seq2item(ngram2seq(ngram));
}


/* ================================================================ */
/* Arithmetic operations                                            */
/* ================================================================ */

static Item plus_wp(Item arglist)
{
  Item next;
  long output = 0;
  float foutput = 0;
  int isfloat = 0;

  while (!null(arglist)){
    next = car(arglist);
    if (item2type(next) == TYPEINT){
      if (!isfloat)
	output += item2long(next);
      else
	foutput += item2long(next);
    }
    else if (item2type(next) == TYPEFLOAT){
      if (!isfloat){
	foutput = output;
	isfloat = 1;
      }
      foutput += item2float(next);
    }
    else
      error("Non-number input to +:  %s\n", printitem(arglist));
    arglist = cdr(arglist);
  }
  if (isfloat)
    return float2item(foutput);
  else
    return long2item(output);
}


static Item times_wp(Item arglist)
{
  Item next;
  long output = 1;
  float foutput = 1;
  int isfloat = 0;

  while (!null(arglist)){
    next = car(arglist);
    if (item2type(next) == TYPEINT){
      if (!isfloat)
	output *= item2long(next);
      else
	foutput *= item2long(next);
    }
    else if (item2type(next) == TYPEFLOAT){
      if (!isfloat){
	foutput = output;
	isfloat = 1;
      }
      foutput *= item2float(next);
    }
    else
      error("Non-number input to *:  %s\n", printitem(arglist));
    arglist = cdr(arglist);
  }
  if (isfloat)
    return float2item(foutput);
  else
    return long2item(output);
}


static Item minus_wp(Item a, Item b)
{
  long output = 0;
  float foutput = 0;
  int isfloat = 0;

  /* Process first input */
  if (item2type(a) == TYPEINT)
    output = item2long(a);
  else if (item2type(a) == TYPEFLOAT){
    isfloat = 1;
    foutput = item2float(a);
  }
  else
    error("Non-number input to -:  %s\n", printitem(a));
  
  /* Process second input */
  if (item2type(b) == TYPEINT){
    if (isfloat)
      foutput -= item2long(b);
    else
      output -= item2long(b);
  }
  else if (item2type(b) == TYPEFLOAT){
    if (!isfloat)
      foutput = output;
    isfloat = 1;
    foutput -= item2float(b);
  }
  else
    error("Non-number input to -:  %s\n", printitem(b));

  if (isfloat)
    return float2item(foutput);
  else
    return long2item(output);
}

static Item divide_wp(Item a, Item b)
{
  long output = 0;
  float foutput = 0;
  int isfloat = 0;

  /* Process first input */
  if (item2type(a) == TYPEINT)
    output = item2long(a);
  else if (item2type(a) == TYPEFLOAT){
    isfloat = 1;
    foutput = item2float(a);
  }
  else
    error("Non-number input to -:  %s\n", printitem(a));
  
  /* Process second input */
  if (item2type(b) == TYPEINT){
    if (isfloat)
      foutput /= item2long(b);
    else
      output /= item2long(b);
  }
  else if (item2type(b) == TYPEFLOAT){
    if (!isfloat)
      foutput = output;
    isfloat = 1;
    foutput /= item2float(b);
  }
  else
    error("Non-number input to -:  %s\n", printitem(b));

  if (isfloat)
    return float2item(foutput);
  else
    return long2item(output);
}


/* Computes (n mod m).   Both inputs must be non-negative
   integers, with m non-zero. */
static Item mod_wp (Item n, Item m)
{
  long vn, vm;

  assert_input_type(n,TYPEINT);
  assert_input_type(m,TYPEINT);

  vn = item2long(n);
  vm = item2long(m);

  if (vn < 0)
    error("First input to mod must be non-negative\n");
  if (vm <= 0)
    error("Second input to mod must be positive\n");
  return long2item(vn%vm);
}

int primitive_gt(Item n, Item m)
{
  int tn, tm;
  tn = item2type(n);
  tm = item2type(m);

  if (tn == TYPEINT && tm == TYPEINT)
    return (item2long(n) > item2long(m));
  else if (tn == TYPEFLOAT && tm == TYPEFLOAT)
    return (item2float(n) > item2float(m));
  else if (tn == TYPEINT && tm == TYPEFLOAT)
    return (item2long(n) > item2float(m));
  else if (tn == TYPEFLOAT && tm == TYPEINT)
    return (item2float(n) > item2long(m));
  else
    error("Non-number input to numerical comparison:  %s or %s\n", printitem(n), printitem(m));
  
  return 0;  /* In fact, unreacheable. */
}

static Item gt_wp(Item n, Item m)
{

  return long2item(primitive_gt(n,m));
}

static Item le_wp(Item n, Item m)
{
  return long2item(!primitive_gt(n,m));
}


static Item ge_wp(Item n, Item m)
{
  return le_wp(m,n);
}

static Item lt_wp(Item n, Item m)
{
  return gt_wp(m,n);
}

static Item numequal_wp(Item n, Item m)
{
  int tn, tm;
  tn = item2type(n);
  tm = item2type(m);

  if (tn == TYPEINT && tm == TYPEINT){
    if (item2long(n) == item2long(m))
      return TRUE;
    else 
      return FALSE;
  }
  else
    error("Non-integer input to numerical equality:  %s or %s\n", printitem(n), printitem(m));
  return MISSING;
}


/* Round input number to nearest integer */
static Item round_wp (Item input)
{
  int type;
  long rounded;
  

  type = item2type(input);
  if (type == TYPEINT)
    return input;
  else if (type == TYPEFLOAT){
    rounded = rint(item2float(input));
    return long2item(rounded);
  }
  else
    error("Input %s to round isn't a number\n", printitem(input));
  return MISSING;
}

/* ================================================================ */
/* Boolean operators                                                */
/* ================================================================ */

static Item not_wp(Item n)
{
  if (eq(n,FALSE))
    return TRUE;
  else
    return FALSE;
}

static Item and_sf(Item arglist, Schwa_env *environment)
{
  Item next;

  while (!null(arglist)){
    next = eval(car(arglist), environment);
    if (eq(next,FALSE))
      return(FALSE);
    arglist = cdr(arglist);
  }
  return TRUE;
}

static Item or_sf(Item arglist, Schwa_env *environment)
{
  Item next;

  while (!null(arglist)){
    next = eval(car(arglist), environment);
    if (!eq(next,FALSE))
      return(TRUE);
    arglist = cdr(arglist);
  }
  return FALSE;
}

/* ================================================================ */
/* Earrays                                                          */
/* ================================================================ */

/* The following wrappers allow you to perform simple operations on
   Earrays whose values are Items.   Earrays are intended primarily 
   for optimizing storage and, therefore, will typically be manipulated
   from C.   The underlying C data structure allows you to store
   an arbitrary C object (e.g. a char value) in an Earray and loop
   efficiently through very large Earrays. */

static Item new_earray_wp (void)
{
  Earray *ptr;
  ptr = new_earray(sizeof(Item));
  return ptr2item(ptr, string2symbol("EARRAY"));
}

static Item earray_ref_wp(Item input, Item position)
{
  Item *ptr;
  Item isym;
  long pos;

  isym = string2symbol("EARRAY");
  pos = item2long(position);

  if (pos < 0)
    error("Earray_ref:   earray positions cannot be negative %ld\n", pos);

  ptr = earray_ref((Earray *)item2ptr(input, isym), pos);
  return *ptr;

}

static Item earray_set_wp(Item input, Item position, Item value)
{
  Item *ptr;
  Item isym;
  long pos;

  isym = string2symbol("EARRAY");
  pos = item2long(position);
  if (pos < 0)
    error("Earray_ref:   earray positions cannot be negative %ld\n", pos);

  ptr = earray_ref((Earray *)item2ptr(input, isym), pos);
  *ptr = value;
  return value;
}

/* ================================================================ */
/* Top-level function                                               */
/* ================================================================ */

/* Allocate and return a blank environment. */
Schwa_env *null_env(char *prompt)
{
  Schwa_env *output;

  output = (Schwa_env *)safe_malloc(sizeof(Schwa_env));

  output->globalbindings = new_hashtable();
  output->localbindings = NIL;
  output->dynamicbindings = NIL;
  output->leavelive = 0;
  output->force_echo = 0;
  output->prompt = prompt;

  return output;
}


/* Returns an environment preloaded with definitions of standard built-in 
   functions, variables, and special forms.  

   This environment can be customized to your needs by 
     -- adding new functions (defun, defun_vararg), special forms
        (defun_sf), and symbols (setsymbol).
     -- removing unwanted definitions (rebind the name to MISSING
        using setsymbol)
     -- renaming predefined functions, special forms, and symbols (use
        evalsymbol to retrieve a predefined item, then setsymbol to
	bind it to a new name). 
*/
Schwa_env *standard_env()
{
  Schwa_env *environment;
  environment = null_env(">> ");

  defun("car", car, 1, environment);
  defun("cdr", cdr, 1, environment);
  defun("cadr", cadr, 1, environment);
  defun("cons", cons, 2, environment);
  defun("nth", nth_wp, 2, environment);
  defun("assq", assq, 2, environment);
  defun("length", length_wp, 1, environment);
  defun("new_hash", new_hash_wp, 0, environment);
  defun("hashget", hashget_wp, 2, environment);
  defun("hashset", hashset_wp, 3, environment);
  defun("hashadd", hashadd_wp, 3, environment);
  defun("hashincrement", hashincrement_wp, 3, environment);
  defun("hashlen", hashlen, 1, environment);
  defun("hashkeys", hashkeys_wp, 1, environment);
  defun("sortseq", sortseq_wp, 1, environment);
  defun_vararg("list", list_wp, environment);
  defun_vararg("+", plus_wp, environment);
  defun_vararg("*", times_wp, environment);
  defun("-", minus_wp, 2, environment);
  defun("/", divide_wp, 2, environment);
  defun("mod", mod_wp, 2, environment);
  defun("round", round_wp, 1, environment);
  defun("not", not_wp, 1, environment);
  defun_sf("and", and_sf, environment);
  defun_sf("or", or_sf, environment);
  defun(">", gt_wp, 2, environment);
  defun(">=", ge_wp, 2, environment);
  defun("<", lt_wp, 2, environment);
  defun("<=", le_wp, 2, environment);
  defun("==", numequal_wp, 2, environment);
  defun("eq", eq_wp, 2, environment);
  defun("equal", equal_wp, 2, environment);
  defun("null", null_wp, 1, environment);
  defun("reverse", reverse, 1, environment);
  defun("append", append, 2, environment);
  defun("memq", memq_wp, 2, environment);
  defun("delq", delq, 2, environment);
  defun("member", member_wp, 2, environment);
  defun_vararg("bundle", bundle_wp, environment);
  defun("bundleget", bundleget, 2, environment);
  defun_vararg("bundleset", bundleset_wp, environment);
  defun("bundlehead", bundlehead, 1, environment);
  defun("bundlefeats", bundlefeats, 1, environment);
  defun("set_prindepth", set_prindepth, 1, environment);
  defun("set_prinlength", set_prinlength, 1, environment);
  defun("set_linewidth", set_linewidth, 1, environment);
  defun_sf("set_prompt", setprompt_sf, environment);
  defun("get_open_files",  get_open_files, 0, environment);
  defun("open_input_file", open_input_file, 1, environment);
  defun("open_output_file", open_output_file, 1, environment);
  defun("close_file", close_file, 1, environment);
  defun_sf("write", write_sexp, environment);
  defun_sf("write_infull", write_sexp_full, environment);
  defun_sf("format", format_wp, environment);
  defun_sf("format_infull", format_if_wp, environment);
  defun_sf("sformat", sformat_wp, environment);
  defun_sf("sformat_infull", sformat_if_wp, environment);
  defun_sf("read", read_sexpression, environment);
  defun_sf("load", load_sf, environment);
  defun("read_file", read_file_wp, 1, environment);
  defun("at_end", at_end_wp, 1, environment);
  defun("present", present_wp, 1, environment);
  defun_vararg("error", error_wp, environment);
  defun_sf("exit", exit_sf, environment);
  defun("string2symbol", string2symbol_wp, 1, environment);
  defun("stringinterned", stringinterned_wp, 1, environment);
  defun("string2number", string2number_wp, 1, environment);
  defun("symbol2string", symbol2string_wp, 1, environment);
  defun_vararg("ngram", ngram_wp, environment);
  defun("list2ngram", list2ngram_wp, 1, environment);
  defun("listinterned", listinterned_wp, 1, environment);
  defun("ngram2list", ngram2list_wp, 1, environment);
  defun_vararg("seq", seq_wp, environment);
  defun("list2seq", list2seq_wp, 1, environment);
  defun("seq2list", seq2list_wp, 1, environment);
  defun("seqget", seqget_wp, 2, environment);
  defun("seqset", seqset_wp, 3, environment);
  defun("seqfill", seqfill_wp, 2, environment);
  defun("new_seq", newseq_wp, 1, environment);
  defun("ngram2seq", ngram2seq_wp, 1, environment);
  defun("seq2ngram", seq2ngram_wp, 1, environment);
  defun("ls", ls_wp, 1, environment);
  defun("item2type", item2type_wp, 1, environment);
  defun("new_earray", new_earray_wp, 0, environment);
  defun("earray_ref", earray_ref_wp, 2, environment);
  defun("earray_set", earray_set_wp, 3, environment);
  defun_sf("dohash", dohash_sf, environment);
  defun_sf("doearray", doearray_sf, environment);
  defun_sf("if", if_sf, environment);
  defun_sf("when", when_sf, environment);
  defun_sf("cond", cond_sf, environment);
  defun_sf("case", case_sf, environment);
  defun_sf("while", while_sf, environment);
  defun_sf("for", for_sf, environment);
  defun_sf("dolist", dolist_sf, environment);
  defun_sf("doseq", doseq_sf, environment);
  defun_sf("io_from", iofrom_sf, environment);
  defun_sf("io_to", ioto_sf, environment);
  defun_sf("eval", eval_sf, environment);
  defun_sf("apply", apply_sf, environment);
  setsymbol(string2symbol("*evalin*"), 
	    ptr2item(stdin,string2symbol("FILE_HANDLE")),
	    environment);
  setsymbol(string2symbol("*evalout*"), 
	    ptr2item(stdout, string2symbol("FILE_HANDLE")),
	    environment);
  setsymbol(string2symbol("*stderr*"), 
	    ptr2item(stderr,string2symbol("FILE_HANDLE")),
	    environment);
  setsymbol(string2symbol("*stdin*"), 
	    ptr2item(stdin, string2symbol("FILE_HANDLE")),
	    environment);
  setsymbol(string2symbol("*stdout*"), 
	    ptr2item(stdout,string2symbol("FILE_HANDLE")),
	    environment);
  return environment;
}

