/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/

/* Symbol tables.

   These are hash tables which map strings to/from integer codes.
   They support additions, but not deletions.  

   The integer codes are guaranteed to be allocated consecutively starting 
   near zero.   This allows applications to create packed arrays mapping symbol
   ID's directly to values.  It also allows the ngram implementation (ngram.c)
   to impose a reasonable limit on the maximum symbol code. 

   Hashing is case-sensitive.

   The table keeps its own private copies of strings, so you don't have
   to be careful about how you handle strings passed to/from its functions.

   The forward table stores pointers to strings rather than packing strings 
   directly into the forward table.   The latter would be more compact but
   would sabotage the goal of allocating ID numbers consecutively.

   Hashtable bins contain short linked lists, not (e.g.) somewhat longer 
   arrays of difference-encoded values.   The latter would be more compact
   but would make symbol access slower.

   Utilities supporting symbols (e.g. in lists.c) use a single global
   symbol table.  However, you can also invoke these functions to
   create (and destroy) additional symbol tables for your own
   nefarious purposes.

   The table arrays are segmented into blocks, so that they can be allocated
   incrementally in small pieces.   The garbage collector will not be happy
   if we keep allocating and deallocating really large arrays. 

   Margaret M. Fleck 2004-2005
 */

#include "utils.h"
#include "symbol.h"

/* Ratio of hashtable size to forward map size.   Or, equivalently, the
   average number of hash cells per stored string, at the point when the
   forward table has been filled to capacity. */
#define OCCUPANCY 0.3 
/* Starting length for expandable tables. */
#define STARTSIZE 100
/* Factor for expanding table size when it fills up. */
#define EXPANDFACTOR 2
/* Constants used for dividing big tables into reasonable-sized blocks. 
   Stay well clear of GC's chunk size of 4K */
#define BLOCKSIZE 512
#define BLOCKSHIFT 9
#define BLOCKBITS 511

/* Map string into hash/backward map, assuming it's already in the forward map
   and not yet in the hash table.  Used for rebuilding table after enlarging arrays. */
static void rehash_string (char *string, unsigned long code, Symboltable *table)
{
  unsigned long bin;
  Codelist *ptr, *bincontents;
  Codelist **block;

  if (string == 0)  /* can remove after testing */
       error("Null string in rehash_string\n"); 
  /* Find the appropriate hash table bin. */
  bin = hashstring(string, table->hashtablesize);
  if ((bin>>BLOCKSHIFT) >= table->hashblocklen) /* can remove after testing */
    error("Issue in rehash\n");
  block = table->hashtable[bin>>BLOCKSHIFT];
  if (block == 0) /* can remove after testing */
    error("Null block %ld in rehash\n", bin>>BLOCKSHIFT);
  bincontents = block[bin&BLOCKBITS];

  /* Now, add code to the hash table bin. */
  ptr = safe_malloc(sizeof(Codelist));
  ptr->value = code;
  ptr->next = bincontents;
  block[bin&BLOCKBITS] = ptr;
}


/* Expand size of hash table. */
static void enlarge_hash_table(Symboltable *table, unsigned long newsize)
{
  unsigned long i, j, blocknum, oldlength, newlength, numblocks;
  Codelist **block;
  char **fblock;

  numblocks = 1+ ((newsize-1)>>BLOCKSHIFT);
  table->hashtablesize = newsize;
  /* Check if top-level table is too short */
  if (numblocks > table->hashblocklen){
    oldlength = table->hashblocklen;
    newlength = oldlength*2;
    table->hashblocklen = newlength;
    table->hashtable = 
      (Codelist ***)safe_realloc(table->hashtable, 
				 newlength*sizeof(Codelist **));
    for (i = oldlength; i < newlength; i++)
      table->hashtable[i] = 0;
  }

  /* Now, allocate any additional blocks required. */
  for (i = 0; i < numblocks; i++){
    if (table->hashtable[i] == 0){
      block = (Codelist **)safe_malloc(BLOCKSIZE*sizeof(Codelist *));
      table->hashtable[i] = block;
      for (j = 0; j < BLOCKSIZE; j++)
	block[j] = 0;
    }
  }

  /* Rehash all values in table. */
  for (i = table->firstcode; i < table->maxelement; i++){
    blocknum = i>>BLOCKSHIFT;
    fblock = table->forwardtable[blocknum];
    if (fblock == 0)   /* can remove after testing */
      error("Zero block in enlarge_hash_table\n");
    rehash_string(fblock[i&BLOCKBITS], i, table);
    }
}

/* Returns the string associated with a particular hash code. 
   Returns 0 if this code has not been allocated.  */
char *code2string (unsigned long code, Symboltable *table)
{
  unsigned int blocknum;
  char **block;

  if (code >= table->maxelement
      || code < table->firstcode)
    return 0;
  else{
    blocknum = code >> BLOCKSHIFT;
    block = table->forwardtable[blocknum];
    return copystring(block[code&BLOCKBITS]);
  }
}


/* Old table doesn't have any more spaces for new blocks. */
static void enlarge_forward_table (Symboltable *table)
{
  unsigned long oldlength, i, newlength;
  
  oldlength = table->forwardsize;
  newlength = oldlength*EXPANDFACTOR;
  table->forwardsize = newlength;
  table->forwardtable = 
    (char ***)safe_realloc(table->forwardtable, newlength*sizeof(char **));
  for (i = oldlength; i < newlength; i++)
    table->forwardtable[i] = 0;
}

/* Returns the code (always non-negative) assigned to the string.  
   If no code has been assigned yet, it either
      a) if autoadd> 0, adds the string and returns the new code
      b) otherwise, returns -1.  
*/
long intern_string (char *string, Symboltable *table, int autoadd)
{
  unsigned long bin, code, blocknum, hashtablesize;
  Codelist *ptr, *bincontents;
  char *candidate;
  char **block;
  Codelist **hashblock;

  /* Find the appropriate hash table bin. */
  bin = hashstring(string, table->hashtablesize);
  hashblock = table->hashtable[bin>>BLOCKSHIFT];
  bincontents = hashblock[bin&BLOCKBITS];

  /* Search bin's contents to see if our string is already there. */
  ptr = bincontents;
  while (ptr != 0){
    code = ptr->value;
    blocknum = code >> BLOCKSHIFT;
    block = table->forwardtable[blocknum];
    candidate = block[code&BLOCKBITS];
    if (!strcmp(candidate, string))
      return code;
    ptr = ptr->next;
  }

  /* Ok, string's not already in the hash table. */
  if (!autoadd) return -1;

  /* User wants us to add it, so give it the next free space 
     in the forward mapping table.  */
  code = table->maxelement;
  if (code > MAXCODE)
    die("Too many symbols, exceeded 29-bit range.\n");
  blocknum = code >> BLOCKSHIFT;
  if (blocknum >= table->forwardsize)   
    /* no more spaces for blocks in forward table */
    enlarge_forward_table(table);
  block = table->forwardtable[blocknum];
  /* copy input string before storing, in case user decides to 
     modify his copy later, and to insure that our string lives 
     on the heap (not on the stack) */
  if (block == 0){  
    /* need to allocate a new block.  Take this opportunity to 
       check hash occupancy level. */
    block = safe_malloc(BLOCKSIZE*sizeof(char*));
    table->forwardtable[blocknum] = block;
    hashtablesize = findhashprime(BLOCKSIZE*blocknum*OCCUPANCY);
    if (hashtablesize > table->hashtablesize){
      enlarge_hash_table(table, hashtablesize);
      bin = hashstring(string, table->hashtablesize);
      hashblock = table->hashtable[bin>>BLOCKSHIFT];
      bincontents = hashblock[bin&BLOCKBITS];
    }
  }
  block[code&BLOCKBITS] = copystring(string);
  table->maxelement += 1;


  /* Now, add code to the hash table bin. */
  ptr = safe_malloc(sizeof(Codelist));
  ptr->value = code;
  ptr->next = bincontents;
  hashblock[bin&BLOCKBITS] = ptr;
  
  return code;
}



/* Makes a new empty symboltable, reserving <numreserve> special codes at
   start of table.   Tables are allocated in multiples of BLOCKSIZE.
   This is slightly inefficient for really small symbol tables, but
   we don't expect these to occur in practice.   (See hash.c for code
   that works better on small tables.)  */
Symboltable *new_symboltable (unsigned int numreserve)
{
  Symboltable *output;
  long forwardsize = STARTSIZE, hashtablesize, hashblocklen = STARTSIZE;
  int i;
  Codelist **block;

  output = safe_malloc(sizeof(Symboltable));

  /* Forward table will initially have a single block, so hashtable can also
     fit into a single block. */
  output->forwardsize = forwardsize;
  output->firstcode = numreserve;
  output->maxelement = numreserve;
  hashtablesize = findhashprime(BLOCKSIZE*OCCUPANCY);
  if (hashtablesize > BLOCKSIZE)  /* can remove after testing */
    error("Internal error in symbol table, occupancy factor > 1?");
  output->hashtablesize = hashtablesize;

  /* Create forward table and allocate first block of cells in it. */
  output->forwardtable = (char ***)safe_malloc(forwardsize*sizeof(char **));
  for (i = 0; i < forwardsize; i++)
    output->forwardtable[i] = 0;
  output->forwardtable[0] = (char **)safe_malloc(BLOCKSIZE*sizeof(char *));

  /* Create and zero hash table. */
  output->hashblocklen = hashblocklen;
  output->hashtable = 
    (Codelist ***)safe_malloc(hashblocklen*sizeof(Codelist **));
  for (i = 0; i < hashblocklen; i++)
    output->hashtable[i] = 0;
  block = (Codelist **)safe_malloc(BLOCKSIZE*sizeof(Codelist *));
  output->hashtable[0] = block;
  for (i = 0; i < BLOCKSIZE; i++)
    block[i] = 0;

  return output;
}


