/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* Generally useful utility functions, which don't depend on Schwa specifics
   (e.g. list.c). */
/* Margaret M. Fleck, 2004-2005 */

/* Don't complain that these functions could be given lots more features.  
   They aren't intended to solve all the world's problems, just simplify
   assembly of routine code and test scaffolds. */


#include "utils.h"

/* ================================================================ */
/* Error handling                                                   */
/* ================================================================ */

/* Print an error message and then die.  Intended for non-recoverable errors,
   e.g. ran out of memory, mysteriously broken behavior.  */
void die(const char *format, ...)
{
  va_list ap;
  fprintf(stderr, "\nFATAL ERROR:");
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  exit(1);
}

/* errorjump should be set non-zero if programmer has defined a jump point for errors to return back to. 
   errorenv is the buffer for the long jumps involved. */
int errorjump = 0;
sigjmp_buf errorenv;


/* Error is like die, but for errors that user might want to try
   recovering from, e.g.  syntax error in input data files, bugs in
   commands typed at user interface.  By default, die causes the
   program to stop.  To instead create a recovery point, put something
   like the following where you want errors to return to:

       errorjump = 1; 
       rv = sigsetjmp(errorenv, 1);
       if (rv == 0)
          [code to run, in which the error might occur]
       else {
          fprintf(stderr, "Restarting after error.\n");
          [code to run after the error has happened]
          }

   For routine type checking, it's usually easier to use the macro
   assert_type and its relatives, defined in schwa.h */
void error(const char *format, ...)
{
  va_list ap;
  fprintf(stderr, "\nERROR:");
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  
  if (errorjump)
    siglongjmp(errorenv, 1);
  else 
    exit(1);
}

/* ================================================================ */
/* Storage allocation                                               */
/* ================================================================ */


/* Calls GC_malloc and checks that it has succeeded.   The garbage
   collector automatically fills this memory with zeros. */
void *safe_malloc(size_t nbytes)
{
  void *output;

  output = GC_malloc(nbytes);
  if (output == 0)
    die("Garbage collector ran out of memory.\n");
  return output;
}

/* Same, but for allocating objects that can't contain pointers, 
   e.g. strings of characters.   Notice that this memory is NOT
   automatically initialized to zero. */
void *safe_malloc_atomic(size_t nbytes)
{
  void *output;

  output = GC_malloc_atomic(nbytes);
  if (output == 0)
    die("Garbage collector ran out of memory.\n");
  return output;
}

/* Same, but for changing size of an object. */
void *safe_realloc(void *old, size_t newsize)
{
  void *output;

  output = GC_realloc(old, newsize);
  if (output == 0)
    die("Garbage collector ran out of memory.\n");
  return output;
}


/* ================================================================ */
/* String manipulation                                              */
/* ================================================================ */

#define PRINTGUESS 10  /* constant for make_messages */

/* Print into a dynamically allocated string. */
char *gcprintf(const char *format, ...) 
{
  va_list ap;
  int rv, maxlength = PRINTGUESS;
  char *output;

  output = safe_malloc_atomic (maxlength);
  va_start(ap, format);
  rv = vsnprintf (output, maxlength, format, ap);
  va_end(ap);
  if (rv < 0)
    error("gcprintf: odd  behavior from vsprintf");
  else if (rv >= maxlength){
    /* vsnprintf ran out of room.   Its return value tells us how much space it
       would have needed to print the whole thing.   Retry the print command with
       a string of the correct length. */
    maxlength = rv+1; 
    output = safe_realloc (output, maxlength);  
    va_start(ap, format);
    rv = vsnprintf (output, maxlength, format, ap);
    va_end(ap);
    if (rv < 0 || rv >= maxlength)
      error("gcprintf: odd  behavior from vsprintf");
  }

  return output;
}


/* Makes a copy of the input string.  Output array is newly-allocated
   by the garbage collector. */
char *copystring (char *input)
{
  char *output;
  unsigned long length;

  length = strlen(input);
  output = safe_malloc_atomic(length+1); 

  strcpy(output,input);
  return output;
}

/* Makes a copy of the string between the two input positions,
   including the data at both positions.  Returns the empty string if
   end precedes start. */
char *substring (char *start, char *end)
{
  char *output;
  unsigned long length, i;

  if (end < start)
    return copystring("");

  length = end - start + 1;
  output = safe_malloc_atomic(length+1);
  for (i = 0; i < length; i++)
    output[i] = *(start+i);
  output[length] = 0;
  return output;
}

/* ================================================================ */
/* Reading tokens                                                   */
/* ================================================================ */

/* read_token and sread_token provide very simple whitespace
   tokenization.  This isn't the ultimate in tokenizers, but
   could serve as the first step for simple tasks. */

#define TOKENGUESS 10 /* Initial guess at length of token */

/* Get next whitespace-delimited token off the input stream.  Ensures
   that token contains no whitespace but doesn't attempt to remove
   punctuation (etc).  Returns a newly-allocated string containing the
   token, or 0 at the end of the file.  Input file pointer is left
   right after end of token.  Returns 0 at end of string. */ 
char *read_token (FILE *infile)
{
  char *output;
  int nextchar, loc = 0, foundword = 0, length = TOKENGUESS;
  output = safe_malloc_atomic(TOKENGUESS);

  while ((nextchar = fgetc(infile)) != EOF){
    if (isspace(nextchar) && foundword)   
      /* found whitespace after start of word */
      break;
    if (!isspace(nextchar)){
      foundword = 1;
      if (loc >= length-1){   /* Ran out of space in output string, 
				 remembering that we'll need to leave 
				 space for terminating null character. */
	length = length*2;
	output = GC_realloc(output, length);
      }
      output[loc] = nextchar;
      loc++;
    }}
  if (loc == 0) return 0;
  output[loc] = 0;
  return output;
}

/* Similar to read_token, but takes input from a string.
   Advances input string pointer to first whitespace character 
   after token.  */
char *sread_token(char **string)
{
  char *pos, *start;

  pos = *string;

  /* Find first non-whitespace character in string. */
  while (isspace(*pos))
    pos++;
  if (*pos == 0)
    return 0;
  start = pos;  

  /* Find end of token. */
  while (*pos != 0 && !isspace(*pos))
    pos++;
  *string = pos;
  return substring(start, pos-1);
}



/* ================================================================ */
/* File I/O                                                         */
/* ================================================================ */

#define LINEGUESS 20 /* Initial guess at length of line */

/* Read a line of input, returning a newly allocated string.  Gobbles
   up the newline character but doesn't include it in the returned
   string.  Returns an empty string if input contains two consecutive
   newlines.  Returns 0 at end of file. */
char *getline(FILE *infile)
{
  char *output;
  int length = LINEGUESS, nextchar, pos;

  output = safe_malloc_atomic(length + 1);
  pos = 0;

  while ((nextchar = fgetc(infile)) != EOF && nextchar != '\n'){
    if (pos >= length){
      length = length*2;
      output = safe_realloc(output, length + 1);
    }
    output[pos] = nextchar;
    pos++;
  }
  if (nextchar == EOF && pos == 0)
    return 0;
  output[pos] = 0;
  return output;
}

/* ================================================================ */
/* Hash functions                                                   */
/* ================================================================ */

#define HASHSHIFT 5
#define WORDWIDTH 32  /* number of bits in a word */

/* So-called modified CRC hash function, which I got from Geoff Kuenning. 
   This is similar to an algorithm described in Aho, Sethi, and Ullman, 
   "Compilers: Principles, Techniques, and Tools." 
   Table sizes should be prime numbers.   */
unsigned long hashstring(const char *key, unsigned long tablesize)
{
  unsigned long output = 0, topbits;
  while (*key != 0){
    /* Rotate output by HASHSHIFT bits. */
    topbits = output >> (WORDWIDTH - HASHSHIFT);  
    output = output << HASHSHIFT;
    output = output | topbits;
    /* XOR in the next character of key */
    output = output ^ *key;
    key++;
  }
  return output%tablesize;
}


/* A floating point number that Knuth recommends for scrambling key values.
   This is approximately (sqrt(5) - 1)/2.  */
#define MAGIC 0.6180339887

/* Hashes an integer, using the "multiplication method" from "Algorithms"
   by Cormen, Leiserson, and Rivest.   Table sizes are powers of two for
   efficiency.  Table size should be 2^shift. */
unsigned long hashint (unsigned long input, unsigned int shift)
{
  double mixer;
  unsigned long multiplier;
  
  mixer = MAGIC*pow(2,WORDWIDTH);
  multiplier = floor(mixer);

  return (input*multiplier) >> (WORDWIDTH - shift); 
}

/* Selected primes, increasing in size by roughly a factor of two. */
static unsigned long primes[] = 
  {
    29, 53, 113, 251, 503, 1019, 
    2029, 4079, 8161, 16369, 32717, 65519, 131041, 262111, 524261, 1048517,
    2097133, 4194277, 8388581, 16777139, 33554383,
    0
  };


/* Find the next prime table size that is no smaller than target.
   Or, if that's not possible, the largest prime in our table. */
unsigned long findhashprime (unsigned long target)
{
  unsigned long *ptr = primes;
  unsigned long output;

  output = *ptr;
  while (output < target){
    ptr++;
    if (*ptr == 0)
      break;
    output = *ptr;
  }
  return output;
}
