/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* Code for reading CMU's pronouncing dictionary. 
   Margaret M. Fleck, 2005 */

#include "cmudict.h"

/* This code reads the CMU dictionary into a hashtable.  Each written
   word has one entry, which contains one or more pronunciations.
   It is hashed under the written form of the word (in lowercase).

   Each entry is a bundle, containing fields
      word:  the symbol corresponding to the lowercase version of the word
      pronunciations:   a list of pronunciations

   Each pronunciation is a seq of phonemes, each of which is a symbol
   just like the ones used in CMU's file format.

   Stress markers immediately precede the corresponding vowels (or other 
   syllable nuclei) and look like '1 or '2.  (Unstressed syllables aren't
   marked. 

   This representation is somewhat non-compact, as each phoneme is represented
   by an Item (4 bytes) when it could be converted to a character (1 byte).

*/

/* ================================================================ */
/* Symbolic constants                                               */
/* ================================================================ */

/* Schwa symbols aren't constant from C's point of view.   This can
   impair the efficiency and/or readability of code that manipulates
   feature bundles.   The following code illustrates one workaround. */

static int setup = 0;  /* Have the constants been initialized yet? */
static Item S_ENTRY;
static Item S_WORD;
static Item S_PRONUNCIATIONS;

/* Set up values for the symbolic constants.  */
static void symsetup(void) 
{
  S_ENTRY = string2symbol("entry");
  S_WORD = string2symbol("word");
  S_PRONUNCIATIONS = string2symbol("pronunciations");
}


/* ================================================================ */
/* CMU's pronouncing dictionary                                     */
/* ================================================================ */

/* Strip the variant pronunciation number (if any)
   and convert word to lowercase. */
static char *normalize_CMU_word(char *word)
{
  int length;
  char *ptr;

  length = strlen(word);
  word = copystring(word);

  /* Remove any variant pronunciation number from end of word. */
  ptr = word+length-1;
  if (length >= 4 && *ptr == ')' && *(ptr-2) == '(')
    *(ptr-2) = 0;

  return string2lower(word);
}

/* Each line contains the written form of the word, followed by 
   one pronunciation.  Each phone is represented by a whitespace
   delimited string of 1-2 ASCII characters.  Variant pronunciations
   are indicated by numbers at the end of the word, e.g. 
   GOTHENBERG(2).  Words are written in all-caps. */
static Item parse_cmudict_line(char *line)
{
  char *token, *stress, *last;
  Item pronunciation = NIL, entry, phone; 

  if (!setup) symsetup();

  /* Read the written form of the word at start of line. */
  token = sread_token(&line);
  if (!token)
    return MISSING;
  token = normalize_CMU_word(token);
  entry = new_bundle(S_ENTRY);
  bundleset(entry, S_WORD, string2symbol(token));    

  /* Read pronunciation(s).  Each token is a string representing a phoneme.
     An integer at the end of vowels marks stress level. */
  while ((token = sread_token(&line)) != 0){
    last = token + strlen(token) - 1;
    if (isdigit(*last)){   /* peel off stress marking, if present. */
      if (*last != '0'){
	stress = gcprintf("'%c", *last);
	pronunciation = cons(string2symbol(stress), pronunciation);   
      }
      *last = 0;
    }
    /*     phone = convert_CMU_phone(token); */
    phone = string2symbol(string2lower(token));
    pronunciation = cons(phone, pronunciation);   
  }

  /* The output will be a list containing this single pronunciation, to which
     we may later add other pronunciations from adjacent lines of the file. */
  bundleset(entry, S_PRONUNCIATIONS, list(1,seq2item(list2seq(nreverse(pronunciation)))));   

  return entry;
}



/* Read the CMU pronouncing dictionary, or a file in similar format. 
   Variant pronunciations of words are listed on separate lines.
   However, these lines are not necessarily adjacent
   because related words ending in (say) 's can intervene. */
Hashtable *read_cmudict_file(char *filename)
{
  FILE *fp;
  Item entry, pronunciations, newpronun, word, previous;
  char *rawline;
  Hashtable *output;

  if (!setup) symsetup();
  output = new_hashtable();

  fp = fopen(filename,"r");
  if (!fp) 
    error("Can't open file %s\n", filename);
  
  rawline = getline(fp);

  while (rawline != 0){
    if (*rawline != 0 && *rawline != '#'){    /* comment lines start with # */
      entry = parse_cmudict_line(rawline);
      if (present(entry)){
	word = bundleget(entry,S_WORD);
	newpronun = car(bundleget(entry,S_PRONUNCIATIONS));

	/* Check for prior entries (other variant pronunciations) 
	   for this word. */
	previous = hashget(output, word);
	if (present(previous)){
	  pronunciations = cons(newpronun, bundleget(previous,S_PRONUNCIATIONS));
	  bundleset(previous,S_PRONUNCIATIONS, pronunciations);
	}
	else
	  hashset(output, word, entry);
      }
    }
    rawline = getline(fp);
  }

  fclose(fp);
  return output;
}


