/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* 
   Miscellaneous utility functions 
   
   These functions are generally useful, but not so universal that they
   belong in the Schwa core.  


   Margaret M. Fleck, 2005 */

#include "misc.h"

/* Remove sub-list structure (e.g. tokenizer bracketing) */
Item flatten_list (Item input)
{
  Item output = NIL, next, flat;

  while (!null(input)){
    next = car(input);
    input = cdr(input);
    switch (item2type(next)){
    case TYPELIST:
      flat = flatten_list(next);
      while (!null(flat)){
	output = cons(car(flat), output);
	flat = cdr(flat);
      }
      break;
    default:
      output = cons(next, output);
    }
  }
  return nreverse(output);
}


/* Read past any whitespace. */
void skip_whitespace(FILE *infile)
{
  int nextchar;

  while ((nextchar = fgetc(infile)) != EOF && isspace(nextchar));
  if (nextchar != EOF)
    ungetc(nextchar, infile);
}


/* Convert word to lowercase. */
char *string2lower(char *word)
{
  char *ptr;

  word = copystring(word);
  ptr = word;
  while (*ptr){
    *ptr = tolower(*ptr);
    ptr++;
  }
  return word;
}

/* Convert word to uppercase. */
char *string2upper(char *word)
{
  char *ptr;

  word = copystring(word);
  ptr = word;
  while (*ptr){
    *ptr = toupper(*ptr);
    ptr++;
  }
  return word;
}


/* Union together two lists using member (not memq). 
   Assumes that neither input list contains duplicates. */
Item unionlists (Item one, Item two)
{
  Item output;
  output = two;
  while (!null(one)){
    if (!member(car(one), output))
      output = cons(car(one),output);
    one = cdr(one);
  }
  return output;
}


/* This function allows us to get input from files (e.g. big corpus
     data files) without knowing in advance whether they are
     compressed or not.  File data can also be run through a filtering
     program (filterprgm), e.g. to transform it into plaintext.  (Set
     filterprgm to 0 if filtering isn't needed.)  Less efficient than
     a direct call to fopen if you do know that a file won't be
     compressed.  The returned file pointer can be read as normal, but
     must be closed with pclose rather than fclose.  The value 0 is
     returned if the file can't be opened.  */
 FILE *open_datafile (char *filename, char *filterprgm)
{
  char *dot, *catname = "cat", *command;

  dot = strrchr(filename,'.');

  if (dot){
    if (!strcmp(dot+1,"gz") || !(strcmp(dot+1, "Z")))     
      /* file is compressed with compress or gzip*/
      catname = "zcat";
    else if (!strcmp(dot+1,"bz") || !(strcmp(dot+1, "bz2")))   
      /* file is compressed with bzip2*/
      catname = "bzcat";
  }

  if (filterprgm)
    command = gcprintf("%s %s | %s", catname, filename, filterprgm);
  else
    command = gcprintf("%s %s", catname, filename);
  /*   fprintf(stderr, "Command is %s\n", command); */
  return popen(command, "r");
}
  
/* Helper for show_utterances.   Displays a single word (which may be a list
   of tokens).   Returns number of characters printed. */
int displayword (FILE *fp, Item word)
{
  char *outstring = 0;
  int length = 0;

  if (item2type(word) == TYPELIST){
    while (!null(word)){
      length += displayword(fp,car(word));
      word = cdr(word);
    }
    return length;
  }

  if (item2type(word) == TYPESYMBOL)
    outstring = symbol2string(word);
  else if (item2type(word) == TYPESTRING)
    outstring = item2string(word);
  else 
    error("Bad input to displayword:  %s\n", printitem(word));

  fprintf(fp, "%s ", outstring);
  return 1+strlen(outstring);
}

#define LINETARGET 80   /* Target length for each output line. */

/* Display a nice, plaintext rendition of a list of utterances.
   Input should be a list of utterances, each of which is a list of "words".
   Each word may be a string, symbol, or a list of strings and/or symbols
   (e.g. resulting from a word that has been divided by a tokenizer). */
void show_utterances (FILE *fp, Item input)
{
  Item utt, word;
  int linelen = 0;

  while (!null(input)){
    fprintf(fp, "\n");
    utt = car(input);
    input = cdr(input);
    if (linelen > 0)
      fprintf(fp, "\n");
    linelen = 0;
    while (!null(utt)){
      if (linelen > LINETARGET){
	fprintf(fp, "\n");
	linelen = 0;
      }
      word = car(utt);
      utt = cdr(utt);
      linelen += displayword(fp, word);
      }
  }
  fprintf(fp, "\n");
}
