/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/

/* The entity translation code was adapted (with significant rewriting)
   from code in HTML Tidy. The copyright notice for Tidy reads as follows:

  Copyright (c) 1998-2004 World Wide Web Consortium
  (Massachusetts Institute of Technology, European Research 
  Consortium for Informatics and Mathematics, Keio University).
  All Rights Reserved.

  CVS Info :

    $Author: terry_teague $ 
    $Date: 2004/02/29 03:59:25 $ 
    $Revision: 1.9 $ 

  Contributing Author(s):

     Dave Raggett <dsr@w3.org>

  The contributing author(s) would like to thank all those who
  helped with testing, bug fixes and suggestions for improvements. 
  This wouldn't have been possible without your help.

  COPYRIGHT NOTICE:
 
  This software and documentation is provided "as is," and
  the copyright holders and contributing author(s) make no
  representations or warranties, express or implied, including
  but not limited to, warranties of merchantability or fitness
  for any particular purpose or that the use of the software or
  documentation will not infringe any third party patents,
  copyrights, trademarks or other rights. 

  The copyright holders and contributing author(s) will not be held
  liable for any direct, indirect, special or consequential damages
  arising out of any use of the software or documentation, even if
  advised of the possibility of such damage.

  Permission is hereby granted to use, copy, modify, and distribute
  this source code, or portions hereof, documentation and executables,
  for any purpose, without fee, subject to the following restrictions:

  1. The origin of this source code must not be misrepresented.
  2. Altered versions must be plainly marked as such and must
     not be misrepresented as being the original source.
  3. This Copyright notice may not be removed or altered from any
     source or altered source distribution.
 
  The copyright holders and contributing author(s) specifically
  permit, without fee, and encourage the use of this source code
  as a component for supporting the Hypertext Markup Language in
  commercial products. If you use this source code in a product,
  acknowledgment is not required but would be appreciated.
*/


/* SGML parsing utilities

   Margaret M. Fleck, 2004-2005

   Routines for parsing HTML and the simpler forms of SGML and XML
   (e.g. those used in linguistic corpora).  Primarily aimed at
   converting markup to plain text or extracting simple types of 
   information from it, e.g. finding links in web pages.

   This code only parses individual markup tags, plus some very basic further
   processing.  The caller is expected to handle higher-level analysis.

   Since a primary application is to analyze data from sources
   (e.g. the web) that often produce non-standard markup, parsing
   attempts to be non-judgmental.  That is, the code tries to accept
   anything that looks generally like HTML/XML/SGML, interpreting it
   in a plausible way.  Only the most conspicuous types of
   malformation are reported.

   Top-level functions are next_sgml_item, read_sgml_file, read_sgml_block,
   extract_sgml_block, and sgml2txt.

   The option preservecase (shared by much of this code) controls
   whether the reader preserves case distinctions in markup tag names.
   Set it non-zero (preserve case distinctions) for parsing XML.  Set
   it to zero (convert all tag names to lowercase) when reading HTML.

Known limitations:

   Assumes that javascript will either be enclosed in a comment or CDATA block,
   or else not contain the < character.

   Discards comments.  Discards internal contents of !DOCTYPE declarations.   

   Whitespace is ignored, which is correct for HTML but not for XML.
   
   Of the HTML entity codes, only handles the Latin-1 ones.  Doesn't
   even attempt to handle XML entity definitions.   
*/



#include "sgml.h"

/* ================================================================ */
/* Global constant symbols                                          */
/* ================================================================ */

static int setup = 0;
static Item S_END;
static Item S_EMPTY;
static Item S_UNPARSED;
static Item S_DOCTYPE;

/* Set up various useful symbols. */
static void symsetup (void)
{
  setup = 1;
  S_END = string2symbol("#end#");
  S_EMPTY = string2symbol("#empty#");
  S_UNPARSED = string2symbol("#unparsed#");
  S_DOCTYPE = string2symbol("!DOCTYPE");
}


/* ================================================================ */
/* Reading input from a file                                        */
/* ================================================================ */

/* Reads characters until it hits the end of the file or finds the string "-->" */
static void skip_comment(FILE *infile)
{
  int state = 0;
  int nextchar;
  
  while (state != 3 && (nextchar = fgetc(infile))){
    switch (state) {
    case 0:     /* initial state */
      if (nextchar == '-') state = 1;
      break;
    case 1:    /* we've seen at least one hyphen */
      if (nextchar == '-') state = 2;
      break;
    case 2:   /* we've seen at least two hyphens in a row */
      if (nextchar == '>') state = 3;
      else if (nextchar != '-') state = 0;
      break;
    default:
      die("Mystery error inside skip-comment\n");
      break;
      /* state 3 is the terminating state that breaks the loop */
    }
  }
}

/* Reads and returns the next sgml token in the file.
   A token is either a whitespace-separated word of text or a markup tag
   enclosed in '<>'.   
   Skips silently past comments.
   Returns 0 at end of file. */
static char *next_sgml_token(FILE *infile)
{
  int nextchar, buflength = 10, pos = 0;
  int tagtoken = 0;  /* Is this token a markup tag? */
  int isdoctype = 0;  /* Is this an XML doctype declaration? */
  int internaldecls = 0;  /* Are we inside the internal part of a doctype declaration? */
  int iscdata = 0;  /* Is this a CDATA block? */
  char *buffer;

  buffer = (char *)safe_malloc_atomic(buflength);

  skip_whitespace(infile);

  while ((nextchar = fgetc(infile)) != EOF){
    buffer[pos] = nextchar;
    pos++;
    if (pos == 9 && !strncmp(buffer, "<!DOCTYPE", 9))
      isdoctype = 1;
    if (pos == 9 && !strncmp(buffer, "<![CDATA[", 9))
      iscdata = 1;
    if (pos >= buflength){
      buflength *= 2;
      buffer = (char *)safe_realloc(buffer, buflength);
    }
    if (iscdata){  
      /* A CDATA block can contain absolutely anything except its magic terminating sequence. 
	 So we have to dispose of this case first. */
      if (!strncmp(buffer+pos-3, "]]>", 3))
	break;
    }
    else if (pos >= 4 && !strncasecmp(buffer+pos-4, "<!--", 4)){
      /* Found a comment.  Skip past it. */
      skip_comment(infile);
      skip_whitespace(infile);
      /* There should be no characters preceding the token marker, because the '<' should
	 have caused any preceding token to be terminated.  (See case below.) */
      if (pos != 4)
	fprintf(stderr, "warning: mystery error in next_sgml_token");
      pos = 0;
      tagtoken = 0;
    }
    else if (isdoctype && nextchar == '[')  
      internaldecls = 1;
    else if (tagtoken && nextchar == '>' && !internaldecls)
      break;
    else if (isdoctype && nextchar == ']')
      internaldecls = 0;
    else if (nextchar == '<' && !internaldecls){
      if (pos == 1)    /* no characters preceding the '<' in buffer */
	tagtoken = 1;
      else {           /* a partial token preceded this '<';  terminate it */
	ungetc(nextchar, infile);
	pos--;
	break;
      }
    }
    else if (!tagtoken && isspace(nextchar)){
      pos--;
      break;
    }
  }
  if (pos == 0)  /* should only happen at EOF */
    return 0;
  buffer[pos] = 0;
  return buffer;
}


/* ================================================================ */
/* Translating SGML entity codes                                    */
/* ================================================================ */

/* The entity-translation code in this section borrows heavily from the code in
   Dave Raggett's tidy program (whose copyright permits free re-use). 
   See copyright notice at the top of this file. */

struct _entity
{
    char *name;
    unsigned char code;
};
typedef struct _entity entity;

static entity entities[] =
{
    /*  Markup pre-defined character entities  */
    { "quot",    34 },
    { "amp",     38 }, 
    { "AMP",     38 },  /* used in English Gigaword */
    { "apos",    39 },
    { "lt",      60 }, 
    { "gt",      62 },

    /* Latin-1 character entities  */ 
    { "nbsp",        160 }, 
    { "iexcl",       161 },
    { "cent",        162 },
    { "pound",       163 },
    { "curren",      164 },
    { "yen",         165 },
    { "brvbar",      166 },
    { "sect",        167 },
    { "uml",         168 },
    { "copy",        169 },
    { "ordf",        170 },
    { "laquo",       171 },
    { "not",         172 },
    { "shy",         173 },
    { "reg",         174 },
    { "macr",        175 },
    { "deg",         176 },
    { "plusmn",      177 },
    { "sup2",        178 },
    { "sup3",        179 },
    { "acute",       180 },
    { "micro",       181 },
    { "para",        182 },
    { "middot",      183 },
    { "cedil",       184 },
    { "sup1",        185 },
    { "ordm",        186 },
    { "raquo",       187 },
    { "frac14",      188 },
    { "frac12",      189 },
    { "frac34",      190 },
    { "iquest",      191 },
    { "Agrave",      192 },
    { "Aacute",      193 },
    { "Acirc",       194 },
    { "Atilde",      195 },
    { "Auml",        196 },
    { "Aring",       197 },
    { "AElig",       198 },
    { "Ccedil",      199 },
    { "Egrave",      200 },
    { "Eacute",      201 },
    { "Ecirc",       202 },
    { "Euml",        203 },
    { "Igrave",      204 },
    { "Iacute",      205 },
    { "Icirc",       206 },
    { "Iuml",        207 },
    { "ETH",         208 },
    { "Ntilde",      209 },
    { "Ograve",      210 },
    { "Oacute",      211 },
    { "Ocirc",       212 },
    { "Otilde",      213 },
    { "Ouml",        214 },
    { "times",       215 },
    { "Oslash",      216 },
    { "Ugrave",      217 },
    { "Uacute",      218 },
    { "Ucirc",       219 },
    { "Uuml",        220 },
    { "Yacute",      221 },
    { "THORN",       222 },
    { "szlig",       223 },
    { "agrave",      224 },
    { "aacute",      225 },
    { "acirc",       226 },
    { "atilde",      227 },
    { "auml",        228 },
    { "aring",       229 },
    { "aelig",       230 },
    { "ccedil",      231 },
    { "egrave",      232 },
    { "eacute",      233 },
    { "ecirc",       234 },
    { "euml",        235 },
    { "igrave",      236 },
    { "iacute",      237 },
    { "icirc",       238 },
    { "iuml",        239 },
    { "eth",         240 },
    { "ntilde",      241 },
    { "ograve",      242 },
    { "oacute",      243 },
    { "ocirc",       244 },
    { "otilde",      245 },
    { "ouml",        246 },
    { "divide",      247 },
    { "oslash",      248 },
    { "ugrave",      249 },
    { "uacute",      250 },
    { "ucirc",       251 },
    { "uuml",        252 },
    { "yacute",      253 },
    { "thorn",       254 },
    { "yuml",        255 },


    { 0,            0 }
};


/* Look up symbolic entity name in table.   
   Returns -1 if name is not known to us.
   Assumes caller has checked that input string is not empty. 
   The comments in tidy's code claim that checking the first
   character is faster than hashing for <= 500 items in table. */
static int lookup_entity(char *input)
{
  char firstchar;
  entity *ptr;  /* pointer into table mapping entity names to latin-1 character codes */

  firstchar = *input;
  for (ptr = entities; ptr->name != 0; ptr++){
    if (*(ptr->name) == firstchar
	&& !strcmp(input, ptr->name))
      return ptr->code;
  }
  return -1;
}

#define MAXCODELENGTH 10 /* Maximum length of any character code, to impose sanity
			    in case someone leaves off the final semicolon in their markup. */

/* Translate an individual entity code.  
   Input should contain the code name or number, with the leading & and trailing ;
   stripped off.   Caller should have checked that input is not the empty string.  
   Returns -1 for unrecognized or non-Latin-1 characters.  Otherwise, output is in [0,255]. */
static int trans_code (char *input){
  unsigned int numcode;

  if (input[0] == '#'){  /* numerical entity code */
    if (input[1] == 'x' || input[1] == 'X')   /* hexadecimal number */
      sscanf(input+2, "%x", &numcode );
    else
      sscanf(input+1, "%u", &numcode );
    if (numcode > 255)
      return -1;
    else
      return (unsigned char) numcode;
  }
  else {   /* symbolic entity name */
    return lookup_entity(input);
  }
}

/* Translate any SGML entity codes into normal characters.
   Anything unrecognized or outside Latin-1 is passed through unchanged. */
static char *trans_entity_codes(char *input)
{
  char *output, *codeend, *outptr;
  char buffer[MAXCODELENGTH+1];
  int length, newchar;

  output = safe_malloc_atomic(strlen(input)+1);
  outptr = output;
  while (*input){
    if (*input == '&'){
      codeend = strchr(input,';');
      if (codeend 
	  && (length = codeend-input-1) <= MAXCODELENGTH
	  && length > 0){
	strncpy(buffer,input+1,length);
	buffer[length] = 0;
	newchar = trans_code(buffer); 
	if (newchar >= 0){
	  *outptr = (unsigned char)newchar;
	  outptr++;
	  input = codeend+1;
	  continue;
	}
      }
    }
    *outptr = *input;
    input++;
    outptr++;
  }
  *outptr = 0;
  return output;
}


/* ================================================================ */
/* Parsing markup tokens                                            */
/* ================================================================ */

/* Extract the next sgml attribute value.
   Returns the name.   Advances pointer to next location
   after name. 
   Returns 0 if no next name, e.g. next non-whitespace character isn't alphabetic. 
   Advances past leading whitespace in either case. */
static char *next_sgml_value(char **input)
{
  char *start, *end, *name, quotechar;
  int length;

  start = *input;
  if (!start) 
    return 0;

  /* flush any preceding whitespace */
  while (isspace(*start))
    start++;

  end = start;

  if (*end == '"' || *end == '\''){    /* attribute value is enclosed in quotation marks */
    quotechar = *end;
    end++;
    while (*end && *end != quotechar)
      end++;
    /* trim off quotation marks */
    start++;
    if (!*end)  /* file ended before we saw terminating double quote */
      *input = end;
    else
      *input = end+1;
  }
  else {    /* otherwise, the attribute value is terminated by whitespace */
    while (*end && !isspace(*end)){
      end++;
    }
    *input = end;
  }

  length = end - start;
  if (start == end)
    return 0;
  name = safe_malloc_atomic(length+1);
  strncpy(name,start,length);
  name[length] = 0;

  return name;
}


/* Extract the next sgml head or attribute name.
   Returns the name.   Advances pointer to next location
   after name. 
   Returns 0 if no next name, e.g. next non-whitespace character isn't alphabetic. 
   Advances past leading whitespace in either case. 
*/
static char *next_sgml_name(char **input, int preservecase)
{
  char *start, *end, *name;
  int length;

  start = *input;
  if (!start)
    return 0;

  /* flush any preceding whitespace */
  while (isspace(*start))
    start++;

  end = start;
  while (*end && (isalnum(*end) || strchr("-?:_.", *end))){
    end++;
  }
  *input = end;
  if (start == end)
    return 0;
  length = end - start;

  name = safe_malloc_atomic(length+1);
  strncpy(name,start,length);
  name[length] = 0;

  /* Downcase the output string, if user specified that */
  if (!preservecase){
    start = name;
    while (*start){
      *start = tolower(*start);
      start++;
    }}

  /* temporarily sanity checking, while we're still debugging this code. */
  /*  if (*end && !isspace(*end) && !(strchr("=>/", *end)))
    fprintf(stderr, "next_sgml_name: Unexpected character %c after sgml keyword name %s\n", 
    *end, name); */
  return name;
}

/* Parse mark-up tag into its component parts.
   Assumes that caller has already verified it's a markup tag. 
   May alter the input string. 
*/
static Item parse_tag(char *stringtag, int preservecase)
{
  char *ptr, *head, *attribute, *value, *end;
  Item output;
  int endtag = 0;

  /* skip over initial < */
  ptr = stringtag+1;

  /* Handle special cases of CDATA and DOCTYPE blocks */
  if (!strncmp(ptr,"![CDATA[", 8)){
    /* Remove cdata ending tag, if any */
    end = stringtag;
    while (*end)  
      end++;
    if (!strcmp(end-3,"]]>"))
      *(end-3) = 0;
    return string2item(ptr+8);
  }
  if (!strncmp(ptr,"!DOCTYPE",8))
    return new_bundle(S_DOCTYPE);


  /* handle endtags */ 
  if (*ptr == '/'){
    endtag = 1;
    ptr++;
  }

  /* Find head. */
  head = next_sgml_name(&ptr, preservecase);

  /* Create output bundle object */
  if (head)
    output = new_bundle(string2symbol(head));
  else
    output = new_bundle(MISSING);
  if (endtag)
    bundleset(output, S_END, ANY);


  /* Parse end of tag, and strip off the end-of-tag material so it can't confuse
     extraction of attribute/value pairs. */
  if (*ptr){
    end = ptr;
    while (*end)   /* move to terminating null character */
      end++;
    end--;         /* We're now on last character in string */
    if (*end == '>' && end >= ptr)
      end--;
    if (*end == '?' && end >= ptr)
      end--;
    else if (*end == '/' && end >= ptr){
      bundleset(output, S_EMPTY, ANY);
      end--;
      }
    *(end+1) = 0;
  }

  /* find all the attribute/value pairs in the tag */
  while ((attribute = next_sgml_name(&ptr, preservecase))){
    while (isspace(*ptr))
      ptr++;
    value = 0;
    if (*ptr == '='){
      ptr++;
      value = next_sgml_value(&ptr);
    }
    if (value)
      bundleset(output, string2symbol(attribute), string2item(trans_entity_codes(value)));
    else
      bundleset(output, string2symbol(attribute), ANY); 
  }

  if (*ptr)   /* Attribute parsing has jammed but we aren't at the end of the tag. */
    bundleset(output, S_UNPARSED, string2item(ptr));

  return output;
}

/* ================================================================ */
/* Helpers for top-level functions                                  */
/* ================================================================ */

/* Helper function for next_sgml_item.   Does the low-level tokenization
   but not the skiplist work. */
static Item next_sgml_item_helper (FILE *infile, int preservecase)
{
  Item output;
  char *stringtoken;

  stringtoken = next_sgml_token(infile);
  if (!stringtoken)
    output = END;
  else if (stringtoken[0] == 0)
    die("next_sgml_token mysteriously returned empty string");
  else if (stringtoken[0] == '<')
    output = parse_tag(stringtoken, preservecase);
  else
    /*     output = string2item(stringtoken); */
    output = string2item(trans_entity_codes(stringtoken)); 
  return output;
}


/* We've just seen the inital tag of a block with this label.
   Read through the file until we see the matching end tag, dropping the
   contents on the floor. */
static void sgml_flush_block (FILE *infile, Item label, int preservecase)
{
  Item next;
  if (!setup) symsetup();
  int nesting = 1;

  next = next_sgml_item_helper(infile, preservecase);
  while (!at_end(next)){
    if (item2type(next) == TYPEBUNDLE && eq(bundlehead(next),label)){
      if (bundleisset(next, S_END))
	nesting--;
      else if (!bundleisset(next, S_EMPTY))
	nesting++;
      if (nesting == 0) 
	return;
    }
    next = next_sgml_item_helper(infile, preservecase);
  }
}

/* ================================================================ */
/* Top-level functions                                              */
/* ================================================================ */

/* The top-level functions allow the caller to specify a "skiplist".   This
   is a list of markup tag names (symbols).   When the reader encounter an opening
   tag matching any of these symbols, it ignores input up to and including the
   matching close tag.   For example, it's often a good idea to skip style and script
   blocks in HTML, because their internal contents may confuse later processing.

   The skiplist should not include HTML tags for which a matching end tag would
   not be expected, e.g. img.  
*/

/* next_sgml_item reads next sgml token and returns parsed version of it. 
   Output is one of:
      a bundle, representing an HTML/SGML/XML markup tag
      a non-empty string, representing a plain old piece of text (including CDATA blocks)
      the symbol END, when you reach the end of the file.
   For markup tags, the head of the bundle is the name of the tag.   The
      bundle features are the tag's attribute/value pairs.  An attribute has
      value ANY if no explicit value was set in the HTML.   The
      special features #end# and #empty# indicate end tags and empty tags.
   Entity codes are translated into appropriate characters (Latin-1 only).
   When tag parsing has failed (probably malformed input), the bundle head
      may be MISSING and/or the special feature #unparsed# may contain 
      internal tag contents that couldn't be parsed. */
Item next_sgml_item (FILE *infile, int preservecase, Item skiplist)
{
  Item next;
  if (!setup) symsetup();

  next = next_sgml_item_helper(infile, preservecase);
  while (!at_end(next) && item2type(next) == TYPEBUNDLE 
	 && memq(bundlehead(next),skiplist)){
    if (!bundleisset(next, S_END) && !bundleisset(next, S_EMPTY))
      sgml_flush_block(infile, bundlehead(next), preservecase);
    next = next_sgml_item_helper(infile, preservecase);
  }
  return next;
}

/* Finds and returns the next section of input file which starts and ends
   with tags of type blocklabel.   Returns MISSING if no such block exists.

   This can be used to retrieve the body of an HTML document.  More
   importantly, it can also be used to chunk up reading of a big SGML
   file containing many small documents, e.g. each story in English
   Gigaword is delimited by DOC tags.  Such chunking avoids reading
   the whole file into memory but also avoids having to write all your
   code in streaming style.
   
*/
Item read_sgml_block(FILE *infile, Item blocklabel, int preservecase, Item skiplist)
{
  Item output = NIL, next;
  int inblock = 0;   /* Are we in the specified type of block?  If so, nested how deep? */

  if (!setup) symsetup();

  /* Read until we find a start (or empty) tag matching blocklabel */
  while (!at_end(next = next_sgml_item(infile, preservecase, skiplist))){
    if (item2type(next) == TYPEBUNDLE && eq(bundlehead(next), blocklabel) && !bundleisset(next,S_END)){
      if (!bundleisset(next,S_EMPTY))
	inblock = 1;
      output = list(1,next);
      break;
    }
  }

  /* Now, if we found a (non-empty) block, read the rest of the block until an end tag. */
  if (inblock){
    while (!at_end(next = next_sgml_item(infile, preservecase, skiplist))){
      output = cons(next,output);

      if (item2type(next) == TYPEBUNDLE && eq(bundlehead(next), blocklabel)){
	if (bundleisset(next,S_END))
	  inblock--; 
	else if (!bundleisset(next,S_EMPTY))   /* A nested block. */
	  inblock++;
      }
      
      if (inblock <= 0)   /* should never be < 0 */
	break;
    }}

  if (null(output))
    return MISSING;
  else 
    return nreverse(output);
}

/* Also extracts a block delimited by tags of the specified label, but the input
   is a list of tokens that have already been read from a file. */
Item extract_sgml_block(Item input, Item blocklabel)
{
  Item output = NIL, next;
  int inblock = 0;   /* Are we in the specified type of block?  If so, nested how deep? */

  if (!setup) symsetup();

  /* Read until we find a start (or empty) tag matching blocklabel */
  while (!null(input)){
    next = car(input);
    input = cdr(input);
    if (item2type(next) == TYPEBUNDLE && eq(bundlehead(next), blocklabel) && !bundleisset(next,S_END)){
      if (!bundleisset(next,S_EMPTY))
	inblock = 1;
      output = list(1,next);
      break;
    }
  }

  /* Now, if we found a (non-empty) block, read the rest of the block until an end tag. */
  if (inblock) {
    while (!null(input)){
      next = car(input);
      input = cdr(input);
      output = cons(next,output);
      if (item2type(next) == TYPEBUNDLE && eq(bundlehead(next), blocklabel)){
	if (bundleisset(next,S_END))
	  inblock--; 
	else if (!bundleisset(next,S_EMPTY))   /* A nested block. */
	  inblock++;
      }
      
      if (inblock <= 0)   /* should never be < 0 */
	break;
    }}

  if (null(output))
    return MISSING;
  else 
    return nreverse(output);
}

/* Read whole sgml file, returning a list of tokens. */
Item read_sgml_file (char *filename, int preservecase, Item skiplist)
{
  Item next, output = NIL;
  FILE *fp;
  
  fp = open_datafile(filename,0);
  if (!fp) 
    error("Can't open file %s\n", filename);

  if (!setup) symsetup();
  
  while (!at_end(next=next_sgml_item(fp,preservecase, skiplist)))
    output = cons(next,output);

  pclose(fp);
  return nreverse(output);
}

/* Strips tags from input list of HTML/SGML/XML tokens, returning bare
   text.  Output is a list of utterances, each of which is a list of
   strings.  A new utterance is started whenever a tag in breaklist
   occurs.  The output of this function might, for example, be fed to
   your favorite tokenizer and then to a function for counting bigram
   frequencies. */
Item sgml2txt (Item input, Item breaklist)
{
  Item output = NIL, utterance = NIL, next;

  while (!null(input)){
    next = car(input);
    input = cdr(input);
    if (item2type(next) != TYPEBUNDLE)
      utterance = cons(next,utterance);
    else if (memq(bundlehead(next), breaklist)){
      if (!null(utterance)){
	output = cons(nreverse(utterance), output);
	utterance = NIL;
      }}}
  if (!null(utterance))
    output = cons(nreverse(utterance), output);
  
  return nreverse(output);
}

