
/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* Reader for files from the Mississippi State Switchboard transcriptions. 

   Margaret Fleck, 2004-2005

   This code can read both the word-by-word and the phrase-by-phrase
   version of these transcripts.  However, full information on
   location and timing of non-speech sounds is only found in the
   word-by-word version.

   These transcripts are line-by-line formatted.   Each line has the form
      <turn ID> <starting time> <ending time> <word or words>

*/


#include "swbd.h"

/* There are three types of parsed words:  real words, sounds, and aside 
       markup tags.
   Real words have some or all of the following features:
      -- original:  the literal form of the word as found in the transcript 
         files, useful if you want to look up its pronunciation in the 
         Switchboard dictionary.  
      -- normalized:  version of word in standard orthography, useful for 
         comparing this data to other types of data, including treebank's 
         tagged/parsed versions of switchboard.  This should always
	 be present. Included only if different from original
      -- full: the full version of the word, included if the normalized form
         represents a truncated version of the word.
      -- intended:  what the speaker apparently intended to say.   Sometimes 
         the original transcription, used in treebank, corrects words to 
	 what they should have been.
      -- laughter:  simultaneous laughter during spoken word
   The normalized and intended versions may have the form -[xxx]yyy or 
      yyy[xxx]-, which shows both the full intended word and the part (yyy) 
      that was actually uttered.  Both versions may need to be considered 
      when matching against the parsed/tagged versions of Switchboard.
   The normalization code drops on the floor
      -- the indication (via curly brackets) that some words are new coinages
      -- the suffix _1 indicating a deviant pronunciation
      -- capitalization information (found only on proper nouns)
*/


/* Do final, minor cleanup.   Specifically, remove any curly brackets from 
   around word and remove "_1". */
char *clean_word (char *input)
{
  int length;
  length = strlen(input);
  if (*input == '{')
    input = substring(input+1,input+length-2);
  if (length >= 2 && *(input+length-1) == '1' && *(input+length-2) == '_')
    input = substring(input,input+length-3);
  return string2lower(input);
}

/* If input represents a truncated word, remove the material in brackets 
   to produce a more typical representation of the truncated word. */
char *truncated_form (char *input)
{
  char *output, *ptr;
  char nextchar;
  int length, i;
  int inbrackets = 0;

  length = strlen(input);
  /* Representations of truncated words have a hyphen at start or end. */
  if ((*input != '-') &&  (*(input+length-1) != '-'))
    return input;

  /* Ok, this is a truncated word. */
  output = safe_malloc_atomic(1+length);
  ptr = output;
  for (i=0; i < length; i++){      
    nextchar = input[i];
    if (nextchar == ']'){
      inbrackets = 0;
      continue;
    }
    if (nextchar == '['){
      inbrackets = 1;
    }
    if (inbrackets)
      continue;
    *ptr = nextchar;
    ptr++;
  }
  *ptr = 0;
  return output;
}



/* If input represents a truncated word, return the full form of the word. */
char *full_form (char *input)
{
  char *output, *ptr;
  char nextchar;
  int length, i;

  length = strlen(input);
  /* Representations of truncated words have a hyphen at start or end. */
  if ((*input != '-') &&  (*(input+length-1) != '-'))
    return input;

  /* Ok, this is a truncated word. */
  output = safe_malloc_atomic(1+length);
  ptr = output;
  for (i=0; i < length; i++){      
    nextchar = input[i];
    if (nextchar == '[' || nextchar == ']')     /* remove brackets */
      continue; 
    if (nextchar == '-' && (i == 0 || i == (length-1)))    
      /* remove initial and final hyphens, don't remove any internal ones */
      continue;
    *ptr = nextchar;
    ptr++;
  }
  *ptr = 0;
  return output;
}

/* Parse a "real word" into a bundle containing its various features and 
   normalized versions.
   The caller should have already verified that this is a word rather than
   a sound or a markup tag.   */
Item parse_swbd_word (char *input)
{
  static int allsetup = 0;
  Item output;
  static Item wordtag, origtag, laughtag, normtag, intendedtag, fulltag;
  int length;
  char *primaryword, *intended, *slashloc, *normalized, *full, *lastpos;

  if (!allsetup){
    origtag = string2symbol("original");
    wordtag = string2symbol("word");
    laughtag = string2symbol("laughter");
    normtag = string2symbol("normalized");    
    fulltag = string2symbol("full");
    intendedtag = string2symbol("intended");
    allsetup = 1;
  }

  output = new_bundle(wordtag);

  if (!strncasecmp(input,"[laughter-", 10)){
    /* A spoken word with simultaneous laughter. */
    bundleset(output,laughtag,ANY);
    length = strlen(input);
    lastpos = input+length-1;
    /* Some words have laughter and are also fragments.   All examples are initial fragments
       and the final fragment hyphen is typically inside the closing bracket of the laughter.
       Except in one case, patched by the following two lines, where the hyphen is outside
       the bracket.  */
    if (*lastpos == '-')     
      *(lastpos-1) = '-';
    primaryword = substring(input+10,lastpos-1);
  }
  else if (*input == '['){
    /* A pair, containing the word as actually said and what the transcriber
       thinks they intended, separated by a slash. */
    length = strlen(input);
    slashloc = strchr(input,'/');
    if (!slashloc){
      primaryword = substring(input+1,input+length-2);
    }
    else {
      primaryword = substring(input+1,slashloc-1);
      intended = substring(slashloc+1,input+length-2);
      bundleset(output,intendedtag,string2symbol(clean_word(intended)));
    }}
 else {
   /* A single word with at most minor adornments. */
   primaryword = input;
  }
  normalized = clean_word(primaryword);
  full = full_form(normalized);
  normalized = truncated_form(normalized);
  bundleset(output,origtag,string2symbol(input));
  if (strcmp(normalized,input))
    bundleset(output,normtag,string2symbol(normalized));
  if (strcmp(full, normalized))
    bundleset(output,fulltag,string2symbol(full));
  return output;
}

/* Parse a switchboard word or non-word token. */
Item parse_token(char *input)
{
  static int allsetup = 0;
  static Item noisetag, asidetag, starttag, endtag, typetag;
  static Item laughtag,soundtag,vnoisetag,silencetag;
  Item output;

  if (!allsetup){
    noisetag = string2symbol("noise");
    vnoisetag = string2symbol("vocalized-noise");
    asidetag = string2symbol("aside");
    typetag = string2symbol("type");
    starttag = string2symbol("start");
    endtag = string2symbol("end");
    laughtag = string2symbol("laughter");
    silencetag = string2symbol("silence");
    soundtag = string2symbol("sound");
    allsetup = 1;
  }

  if (*input == '<'){   /* An aside tag. */
    output = new_bundle(asidetag);
    if (!strcasecmp(input,"<b_aside>"))
      bundleset(output,typetag,starttag);
    else if (!strcasecmp(input,"<e_aside>"))
      bundleset(output,typetag,endtag);
    else
      fprintf(stderr, "Warning: mysterious markup tag %s\n", input);
  }
  else if (*input == '['){
    if (!strcasecmp(input,"[laughter]")){
      output = new_bundle(soundtag);
      bundleset(output,typetag,laughtag);
    }
    else if (!strcasecmp(input,"[noise]")){
      output = new_bundle(soundtag);
      bundleset(output,typetag,noisetag);
    }
    else if (!strcasecmp(input,"[vocalized-noise]")){
      output = new_bundle(soundtag);
      bundleset(output,typetag,vnoisetag);
    }
    else if (!strcasecmp(input,"[silence]")){
      output = new_bundle(silencetag);
    }
    else {
      output = parse_swbd_word(input);
    }
  }
  else {
      output = parse_swbd_word(input);
  }

  return output;
}


/* Extract the turn number from the turn ID.*/
Item extract_turn_number(char *turnID)
{
  char *hyphen;

  /* Find last hyphen in ID */
  hyphen = strrchr(turnID, '-');
  if (!hyphen || *(hyphen+1) == 0)
    error("This is not a legal switchboard turn ID:  %s\n", turnID);
  return long2item(atoi(hyphen+1));
}


/* Read and parse next line of switchboard transcription file.
   Returns a list of words, non-word noises, and markup tokens.  
   Each of these is a bundle containing a string and sometimes timing information. 
   Returns END at end of file. */
Item next_swbd_line (FILE *infile)
{
  char *rawline, *token, *turnID;
  Item wordlist = NIL, word;
  Item line;
  float start = -1, end = -1, t;
  int rv;
  static int allsetup = 0;
  static Item starttag, endtag, turntag, linetag, wordtag;

  if (!allsetup){
    starttag = string2symbol("starttime");
    endtag = string2symbol("endtime");
    turntag = string2symbol("ID");
    linetag = string2symbol("line");
    wordtag = string2symbol("words");
    allsetup = 1;
  }

  rawline = getline(infile);

  if (rawline == 0)
    return END;

  line = new_bundle(linetag);

  /* Read past the ID code at start of line. */
  turnID = sread_token(&rawline);
  
  /*   bundleset(line, turntag, string2symbol(turnID));*/
  bundleset(line, turntag, extract_turn_number(turnID));

  /* Read the two timing numbers */
  token = sread_token(&rawline);
  rv = sscanf(token, "%f", &t);
  if (rv >= 1)
    start = t;
  token = sread_token(&rawline);
  rv = sscanf(token, "%f", &t);
  if (rv >= 1)
    end = t;

  /* Read the word or words associated with these timing numbers. */
  while ((token = sread_token(&rawline)) != 0){
    word = parse_token(token);
    wordlist = cons(word, wordlist);
  }

  if (null(wordlist))  /* This can only happen if the input file is somehow damaged. */
    return NIL;
  
  /* Reverse wordlist and also add the timing information to first and last words. */
  if (end >= 0)
    bundleset(car(wordlist), endtag, float2item(end));
  wordlist = reverse(wordlist);
  if (start >= 0)
    bundleset(car(wordlist), starttag, float2item(start));

  bundleset(line, wordtag, wordlist);

  return line;
}


/* Read an entire transcript file, returning results structured into turns. */
Item read_swbd_file(char *filename)
{
  FILE *fp;
  Item turnID = MISSING, newturnID, output = NIL, turn = MISSING;
  Item nextline, wordtag, turntag, turnIDtag, wordlist = NIL;

  wordtag = string2symbol("words");
  turnIDtag = string2symbol("ID");
  turntag = string2symbol("utt");

  fp = open_datafile(filename, 0);
  if (!fp)
    error("Can't open file %s\n", filename);
  
  while (!at_end(nextline = next_swbd_line(fp))){
    newturnID = bundleget(nextline,turnIDtag);

    /* At start of a new turn, flush old turn object onto output and make a new one. */
    if (!eq(turnID,newturnID)){              
      if (present(turn)){   /* turn = MISSING when we first start reading file */
	bundleset(turn,wordtag,wordlist);
	output = cons(turn,output);
	wordlist = NIL;
      }
      turnID = newturnID;
      turn = new_bundle(turntag);
      bundleset(turn, turnIDtag, turnID);
    }
    wordlist = nconc(wordlist,bundleget(nextline,wordtag));
  }

  if (present(turn)){
    bundleset(turn,wordtag,wordlist);
    output = cons(turn,output);
  }

  pclose(fp);
  return reverse(output);
}


/* Convert structured Switchboard representation (as returned by read_swbd_file)
   into a very plain text output.   This consists of a list of turns, each of
   which is a list of words (symbols).   Each word is in normalized form.
   Noises and silences are suppressed. */
Item swbd2plain(Item turns)
{
  Item words, next, normword; 
  static Item wordtag, origtag, normtag, wordstag;
  static int setup = 0;
  Item outturn, output = NIL;

  if (!setup){
    setup = 1;
    wordtag = string2symbol("word");
    wordstag = string2symbol("words");
    normtag = string2symbol("normalized");
    origtag = string2symbol("original");
  }

  while (!null(turns)){
    outturn = NIL;
    words = bundleget(car(turns),wordstag);
    turns = cdr(turns);
    while (!null(words)){
      next = car(words);
      words = cdr(words);
      if (eq(bundlehead(next), wordtag)){   /* exclude noises, silences */
	normword = bundleget(next,normtag);
	if (!present(normword))
	  normword = bundleget(next,origtag);
	if (!present(normword))
	  fprintf(stderr, "Word %s has no normalized form\n", printitem(next));
	else
	  outturn = cons(normword,outturn);
      }}
    if (!null(outturn))
      output = cons(reverse(outturn),output);
  }
  return reverse(output);
}
