/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/


/* Code to divide incoming text into meaningful tokens. 

   Margaret M. Fleck, 2005

   This code tries to do the parts of tokenization that are
   more-or-less mechanical and independent of the language, discourse
   topic, and character encoding system (e.g. Latin-1 vs. UTF-8).
   A hook is provided for subsequent language- and task-dependent 
   fix-up rules, and some examples are included at the end. 

   The top-level functions are tokenize_next_line, tokenize_list, and
   tokenize_word.  These functions are intended for analyzing plain
   text files (tokenize_next_line) or list representations of plain
   text (tokenize_list).   Data containing markup tags (or the like) 
   should be pre-processed to remove them.

   Each whitespace-delimited word is divided at changes between character
   type (alphabetic, digits, punctuation).   Apostrophe (frequently used as
   alphabetic cross-linguistically) and all non-ASCII characters are considered 
   alphabetic.  Each punctuation character is returned as a separate token, 
   except that consecutive sequences of the same character are kept as a unit.

   For corpora which use non-HTML entity codes and languages which use
   a numerical or punctuation character as alphabetic (e.g. Tzotzil),
   it may help to pre-process the input so as to map the offending
   characters or entities to something innocuous (e.g. an alphabetic or
   non-ASCII character).

   When a whitespace-delimited word is subdivided, the subtokens are grouped
   together into a list.   Your fix-up rules operate on this list, to
   identify and split/rejoin tokens that weren't correctly handled by
   the default rules. 

   The default rules tend to break up complex numerical expressions and
   hyphenated words.  Some applications may need to re-assemble these.
   However, beware that reassembly may cause an explosion in the set of 
   distinct tokens for some large datasets, especially those that number
   (or assign unique identifiers to) each article or page of text.
  
   Processing is line-by-line so that the caller can identify blank lines.
   Blank lines frequently indicate a break (e.g. paragraph, utterance). */


#include "token.h"

/* Explicit listing of all ASCII punctuation characters.
   Apostrophe is explicitly excluded. */
static char *punctuation = "~!@#$%^&*()_-+={}[]|`\\:;\"?/>.<,";
static char *digits = "0123456789";


/* Break up the run of punctuation in the string *start.
   Each character is returned separately, in a list, except that 
   consecutive copies of the same character are kept as a unit.
   Output tokens are returned in reverse order. */
static Item break_up_punc(char *start)
{
  Item output = NIL;
  char *pos, *pos2, *subtoken;

  pos = start;
  while (*pos){
    pos2 = pos+1;
    while (*pos2 == *pos)
      pos2++;
    subtoken = substring(pos,pos2-1);
    output = cons(string2item(subtoken), output);
    pos = pos2;
  }
  return output;
}


#define MAXRUN 3

/* Break up the run of digits in the string *start.
   Runs of digits are divided into subsequences of length <= MAXRUN.
   This is to avoid explosion in the number of distinct tokens
   on datasets containing many numbers (e.g. page numbers).
   Output tokens are returned in reverse order.  */
static Item break_up_digits(char *start)
{
  Item output = NIL;
  char *pos, *subtoken;

  pos = start;
  while (*pos){
    if (pos-start+1 >= MAXRUN || *(pos+1) == 0){
      subtoken = substring(start,pos);
      output = cons(string2item(subtoken), output);
      start = pos+1;
    }
    pos++;
  }
  return output;
}


/* Turn string tokens into symbol tokens and remove
   layer of list structure around words that weren't divided. */
Item symbolize_tokens(Item tokenlist)
{
  Item output = NIL;

  while (!null(tokenlist)){
    output = cons(string2symbol(item2string(car(tokenlist))),output);
    tokenlist = cdr(tokenlist);
  }

  if (null(output))
    return NIL;
  else if (null(cdr(output)))
    return car(output);
  else 
    return nreverse(output);
}

/* Divides whitespace-delimited word into character, punctuation, and
   alphabetic-or-other substrings.   Further subdivides punctuation strings
   into individual tokens, except that runs of the same character are kept
   together.  Output is a symbol or list of symbols. 
   fixup (if not 0) is a postprocessing function:  see below for details 
   and examples. */
Item tokenize_word (char *input, FUNC1 fixup)
{
  Item output = NIL;
  char *pos, *startpos, class, oldclass, *divided;

  if (*input == 0)
    error("tokenize_word received empty input word\n");

  oldclass = 27;   /* Random disused integer. */
  startpos = input;
  pos = input;
  while (*pos != 0){
    if (strchr(punctuation, *pos))
      class = 0;
    else if (strchr(digits, *pos))
      class = 1;
    else 
      class = 2;
    if (pos != input && class != oldclass){
      divided = substring(startpos,pos-1);
      if (oldclass == 0)
	output = append(break_up_punc(divided),output);
      else if (oldclass == 1)
	output = append(break_up_digits(divided),output);
      else
	output = cons(string2item(divided),output);
      startpos = pos;
    }
    oldclass = class;
    pos++;
  }

  if (*startpos != 0){
    divided = substring(startpos,pos-1);
    if (oldclass == 0)
      output = append(break_up_punc(divided),output);
    else if (oldclass == 1)
      output = append(break_up_digits(divided),output);
    else
      output = cons(string2item(divided),output);
  } 

  if (fixup)
    return symbolize_tokens(fixup(nreverse(output)));
  else
    return symbolize_tokens(nreverse(output));
}


/* Return tokenized version of list of words.   Each word should
   be a string, containing one plain-text whitespace-delimited word, 
   with no whitespace inside the word.  */
Item tokenize_list (Item wordlist, FUNC1 fixup)
{
  Item output = NIL;
  while (!null(wordlist)){
    output = cons(tokenize_word(item2string(car(wordlist)),fixup), output);
    wordlist = cdr(wordlist);
  }
  return nreverse(output);
}

/* Tokenize a list of utterances, each of which is a list of words appropriate
   for giving to tokenize_list. */
Item tokenize_utterances (Item uttlist, FUNC1 fixup)
{
  Item output = NIL;
  while (!null(uttlist)){
    output = cons(tokenize_list(car(uttlist),fixup), output);
    uttlist = cdr(uttlist);
  }
  return nreverse(output);
}


/* Return tokenized version of next line in file. 
   Returns END at end of file.  
   fixup (if not 0) is a postprocessing function:  see below for details 
   and examples. */
Item tokenize_next_line (FILE *fp, FUNC1 fixup)
{
  char *nextline, *wstoken;
  Item output = NIL, tokens;

  nextline = getline(fp);
  if (nextline == 0)
    return END;
  while ((wstoken = sread_token(&nextline)) != 0){
    tokens = tokenize_word(wstoken,fixup);
    output = cons(tokens,output);
  }
  return nreverse(output);
}

/* ================================================================ */
/* Sampler of fix-up rules                                          */
/* ================================================================ */

/* The second input to tokenize_word or tokenize_next_line
   is a fixup function.   If not 0, this function should
   take as input, and return as output, a list of string
   tokens.  This list is the decomposition of a single
   whitespace delimited word.  

   The goal of a fixup function is to improve the default 
   tokenization, e.g. using knowledge of the specific language
   or specific task at hand.  

   The following functions provide some examples of fix-up rules.
*/


/* Duplicated single quotes at ends of words are almost
   certainly double quotes, not apostrophes.  */
Item tokfix_dquotes (Item tokenlist)
{
  Item output = NIL, partial;
  char *next, *endquotes;
  int length; 

  while (!null(tokenlist)){
    next = item2string(car(tokenlist));
    partial = NIL;
    if (!strncmp(next,"''", 2) || !strncmp(next, "``", 2)){
      partial = cons(string2item(substring(next,next+1)), partial);
      next = next+2;
    }
    length = strlen(next);
    endquotes = 0;
    if (length >= 2 
	&& (!strncmp(next+length-2, "''", 2) || !strncmp(next+length-2, "``", 2))){
      endquotes = substring(next+length-2, next+length-1);
      next[length-2] = 0;
    }
    partial = cons(string2item(next), partial);
    if (endquotes)
      partial = cons(string2item(endquotes), partial);
    output = append(partial,output);
    tokenlist = cdr(tokenlist);
  }
  return nreverse(output);
}

/* Describes the case structure of token.
   Returns "Cap" (initial cap), "CAP" (all caps),
   "CaP" (mixed case), or 0 (all lower). 
   Apostrophe's and non-ASCII characters are consider to have indeterminate case. */
char *describe_case (char *token)
{
  int initialcap = 0;                   /* Case of first letter */
  int foundupper = 0, foundlower = 0;   /* Case of subsequent letters */

  /* Walk past any initial apostrophe's */
  while (*token == '\'' || *token == '`')
    token++;

  if (*token == 0)
    return 0;

  if (isupper(*token))
    initialcap = 1;
  /* Characters with indeterminate case are treated as non-capitals here. */
  token++;

  while (*token){
    if (isupper(*token))
      foundupper = 1;
    else if (islower(*token))
      foundlower = 1;
    /* otherwise, it's a character with indeterminate case */
    token++;
  }
  
  if (initialcap && !foundupper)
    return "<Cap>";
  else if (initialcap && !foundlower)
    return "<CAP>";
  else if (!initialcap && !foundupper)
    return 0;
  else
    return "<CaP>";
}

/* Normalizes words to lowercase. 
   Words that were originally capitalized are preceded by <Cap>
     or <CAP>
   Words with irregular capitalization patterns are preceded by
     <CaP> and are not converted to lowercase.  They should receive
     further specialized processing. */
Item tokfix_case (Item tokenlist)
{
  Item output = NIL;
  char *mycase, *token;

  while (!null(tokenlist)){
    token = item2string(car(tokenlist));
    mycase = describe_case(token);
    if (!mycase)
      output = cons(string2item(token),output);
    else if (!strcmp(mycase, "<CaP>")){
      output = cons(string2item(mycase), output);
      output = cons(string2item(token),output);
    }
    else {
      output = cons(string2item(mycase), output);
      output = cons(string2item(string2lower(token)),output);
    }
    tokenlist = cdr(tokenlist);
  }
  return nreverse(output);
}

/* A sample fix-up function for English text. */
Item tokfix_english (Item tokenlist)
{
  tokenlist = tokfix_dquotes(tokenlist);
  tokenlist = tokfix_case(tokenlist);
  return tokenlist;
}
