/*
University of Illinois/NCSA
Open Source License

Copyright © 2005, The Board of Trustees of the University of Illinois.
All rights reserved.
Developed by:

  Department of Computer Science
  University of Illinois, Urbana-Champaign
  http://www.cs.uiuc.edu

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

  * Redistributions  of  source  code  must retain the above copyright
    notice, this list of conditions and the following disclaimers.
  * Redistributions  in binary form must reproduce the above copyright
    notice, this  list of conditions and the following disclaimers in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the names of the Department of Computer Science,
    University of Illinois, Urbana-Champaign, nor the names of its
    contributors may be used to endorse or promote products derived 
    from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

*/

/*   Wrapper functions for code in the Schwa tools directory. 
   
   This file can be used in two ways:

      -- Call the top-level function, define_schwa_tools to add 
         the tools functions to your interpreter enviroment.

      -- Copy/modify (parts of ) this file to add your own customized 
         interface to some/all of the tools code.

   Margaret M. Fleck, 2005
*/

/* ================================================================ */
/* What is a wrapper?                                               */
/* ================================================================ */

/* To link a function into the interpreter, its inputs and output
   must all be Items.   If your C function has different input
   and/or output types, a wrapper function is needed to do the
   type conversions.  

   The type conversion functions automatically include type checking.
   Wrapper functions can include other forms of sanity-checking, e.g.
   checking the range of input numbers.

   The interpreter API defined by your wrapper functions will often be
   similar to the C API.  However, it does not have to be a literal
   copy.  You may wish to hide certain options from the script writer,
   omit functions that don't seem useful in scripting, and/or add
   extra utility functions.  The API can be made more Scheme-like,
   e.g. use special forms to create new looping constructs or support
   variable numbers of arguments.

   The wrapper functions are typically not intended to be called from
   C code.   Therefore, they are kept separate from the C library files
   (which might be used without the interpreter) and declared static
   (to avoid polluting the C namespace).

*/

#include "tools_wrappers.h"

/* ================================================================ */
/* Wrappers for cmudict functions                                   */
/* ================================================================ */

/* Top-level function from cmudict.c */
static Item read_cmudict_wp(Item filename)
{
  return ptr2item(read_cmudict_file(item2string(filename)),
		  string2symbol("HASHTABLE"));
}

/* ================================================================ */
/* Wrappers for code from sgml.c                                    */
/* ================================================================ */

static Item read_sgml_file_wp (Item filename, Item preservecase, Item skiplist)
{
  return read_sgml_file(item2string(filename), item2long(preservecase), skiplist);
}

static Item read_sgml_block_wp (Item filehandle, Item label, Item preservecase, Item skiplist)
{
  FILE *fp = 0;

  if (is_special_type(filehandle, string2symbol("FILE_HANDLE")))
    fp = item2ptr(filehandle,string2symbol("FILE_HANDLE"));
  else if (is_special_type(filehandle, string2symbol("PIPE")))
    fp = item2ptr(filehandle,string2symbol("PIPE"));
  else
    error("Input %s to read_sgml_block isn't a file handle or pipe.\n", printitem(filehandle));

  return read_sgml_block(fp, label, item2long(preservecase), skiplist);
}


/* ================================================================ */
/* Wrappers for code from misc.c                                    */
/* ================================================================ */

static Item show_utterances_wp(Item filehandle, Item input)
{
  show_utterances(item2ptr(filehandle, string2symbol("FILE_HANDLE")), input);
  return MISSING;
}

/* Open a file which might be compressed.  */
static Item open_datafile_wp(Item filename)
{
  FILE *fp;

  fp = open_datafile (item2string(filename),0);
  if (!fp)
    error("Can't open file %s\n", item2string(filename));
  return ptr2item(fp,string2symbol("PIPE"));
}

static Item close_datafile_wp(Item filehandle)
{
  pclose(item2ptr(filehandle,string2symbol("PIPE")));
  return MISSING;
}

static Item string2lower_wp (Item input)
{
  return string2item(string2lower(item2string(input)));
}

static Item string2upper_wp (Item input)
{
  return string2item(string2upper(item2string(input)));
}


/* ================================================================ */
/* Wrappers for code from token.c                                   */
/* ================================================================ */

/* These functions "dumb down" the capabilities of the underlying C
   API by calling a specific, somewhat English-biased fixup-function.
   You'll want to build your own new wrapper functions if you are
   experimenting seriously with tokenization.

   A more general interpreted API is certainly possible, but I've
   found that people have strong, divergent opinions about the
   design of such things. 
*/

static Item tokenize_next_line_wp (Item filehandle)
{
  FILE *fp;
  fp = item2ptr(filehandle, string2symbol("FILE_HANDLE"));
  return tokenize_next_line(fp, tokfix_english);
}

static Item tokenize_utterances_wp (Item input)
{
  return tokenize_utterances(input, tokfix_english);
}

static Item tokenize_list_wp (Item input)
{
  return tokenize_list(input, tokfix_english);
}

static Item tokenize_word_wp (Item input)
{
  return tokenize_word(item2string(input), tokfix_english);
}

/* ================================================================ */
/* Wrappers for functions from swbd.c                               */
/* ================================================================ */

Item swbd_file_wp(Item filename)
{
  return read_swbd_file(item2string(filename));  
}

Item swbd_read_line(Item filehandle)
{
  FILE *fp;
  fp = item2ptr(filehandle,string2symbol("FILE_HANDLE"));
  return next_swbd_line(fp);
}


/* ================================================================ */
/* Top-level function                                               */
/* ================================================================ */

/* Adds all the tools functions to the input environment */
void define_schwa_tools(Schwa_env *environment)
{
  defun("read_sgml_file", read_sgml_file_wp, 3, environment);
  defun("extract_sgml_block", extract_sgml_block, 2, environment);
  defun("sgml2txt", sgml2txt, 2, environment);
  defun("read_sgml_block", read_sgml_block_wp, 4, environment);

  defun("show_utterances", show_utterances_wp, 2, environment);
  defun("open_datafile", open_datafile_wp, 1, environment);
  defun("close_datafile", close_datafile_wp, 1, environment);
  defun("union", unionlists, 2, environment);
  defun("string2upper", string2upper_wp, 1, environment);
  defun("string2lower", string2lower_wp, 1, environment);
  defun("flatten_list", flatten_list, 1, environment);

  defun("tokenize_word", tokenize_word_wp, 1, environment);
  defun("tokenize_list", tokenize_list_wp, 1, environment);
  defun("tokenize_utterances", tokenize_utterances_wp, 1, environment);
  defun("tokenize_next_line", tokenize_next_line_wp, 1, environment);
  defun("tokenize_utterances", tokenize_utterances_wp, 1, environment);

  defun("read_cmudict", read_cmudict_wp, 1, environment);

  defun("swbd_read_line", swbd_read_line, 1, environment);
  defun("read_swbd_file", swbd_file_wp, 1, environment);
  defun("swbd2plain", swbd2plain, 1, environment);

}
